/*
 * This file is generated by gdbus-codegen, do not modify it.
 *
 * The license of this code is the same as for the D-Bus interface description
 * it was derived from. Note that it links to GLib, so must comply with the
 * LGPL linking clauses.
 */

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include "atspi-component.h"

#include <string.h>

/* ------------------------------------------------------------------------ */
/* Definitions for org.a11y.atspi.Component */

const GDBusArgInfo atspi_component_interface__contains_method_in_args__x_arg =
{
  -1,  /* ref count */
  (gchar *) "x",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_component_interface__contains_method_in_args__y_arg =
{
  -1,  /* ref count */
  (gchar *) "y",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_component_interface__contains_method_in_args__coord_type_arg =
{
  -1,  /* ref count */
  (gchar *) "coord_type",
  (gchar *) "u",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_component_interface__contains_method_in_args[] =
{
  &atspi_component_interface__contains_method_in_args__x_arg,
  &atspi_component_interface__contains_method_in_args__y_arg,
  &atspi_component_interface__contains_method_in_args__coord_type_arg,
  NULL,
};

const GDBusArgInfo atspi_component_interface__contains_method_out_args__unnamed_arg3_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg3",
  (gchar *) "b",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_component_interface__contains_method_out_args[] =
{
  &atspi_component_interface__contains_method_out_args__unnamed_arg3_arg,
  NULL,
};

const GDBusMethodInfo atspi_component_interface__contains_method =
{
  -1,  /* ref count */
  (gchar *) "Contains",
  (GDBusArgInfo **) atspi_component_interface__contains_method_in_args,
  (GDBusArgInfo **) atspi_component_interface__contains_method_out_args,
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_component_interface__get_accessible_at_point_method_in_args__x_arg =
{
  -1,  /* ref count */
  (gchar *) "x",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_component_interface__get_accessible_at_point_method_in_args__y_arg =
{
  -1,  /* ref count */
  (gchar *) "y",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_component_interface__get_accessible_at_point_method_in_args__coord_type_arg =
{
  -1,  /* ref count */
  (gchar *) "coord_type",
  (gchar *) "u",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_component_interface__get_accessible_at_point_method_in_args[] =
{
  &atspi_component_interface__get_accessible_at_point_method_in_args__x_arg,
  &atspi_component_interface__get_accessible_at_point_method_in_args__y_arg,
  &atspi_component_interface__get_accessible_at_point_method_in_args__coord_type_arg,
  NULL,
};

const GDBusArgInfo atspi_component_interface__get_accessible_at_point_method_out_args__unnamed_arg3_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg3",
  (gchar *) "(so)",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_component_interface__get_accessible_at_point_method_out_args[] =
{
  &atspi_component_interface__get_accessible_at_point_method_out_args__unnamed_arg3_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_component_interface__get_accessible_at_point_method_annotations__org_qtproject__qt_dbus__qt_type_name__out0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.Out0",
  (gchar *) "QSpiObjectReference",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_component_interface__get_accessible_at_point_method_annotations[] =
{
  &atspi_component_interface__get_accessible_at_point_method_annotations__org_qtproject__qt_dbus__qt_type_name__out0_annotation,
  NULL,
};

const GDBusMethodInfo atspi_component_interface__get_accessible_at_point_method =
{
  -1,  /* ref count */
  (gchar *) "GetAccessibleAtPoint",
  (GDBusArgInfo **) atspi_component_interface__get_accessible_at_point_method_in_args,
  (GDBusArgInfo **) atspi_component_interface__get_accessible_at_point_method_out_args,
  (GDBusAnnotationInfo **) atspi_component_interface__get_accessible_at_point_method_annotations,
};

const GDBusArgInfo atspi_component_interface__get_extents_method_in_args__coord_type_arg =
{
  -1,  /* ref count */
  (gchar *) "coord_type",
  (gchar *) "u",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_component_interface__get_extents_method_in_args[] =
{
  &atspi_component_interface__get_extents_method_in_args__coord_type_arg,
  NULL,
};

const GDBusArgInfo atspi_component_interface__get_extents_method_out_args__unnamed_arg1_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg1",
  (gchar *) "(iiii)",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_component_interface__get_extents_method_out_args[] =
{
  &atspi_component_interface__get_extents_method_out_args__unnamed_arg1_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_component_interface__get_extents_method_annotations__org_qtproject__qt_dbus__qt_type_name__out0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.Out0",
  (gchar *) "QSpiRect",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_component_interface__get_extents_method_annotations[] =
{
  &atspi_component_interface__get_extents_method_annotations__org_qtproject__qt_dbus__qt_type_name__out0_annotation,
  NULL,
};

const GDBusMethodInfo atspi_component_interface__get_extents_method =
{
  -1,  /* ref count */
  (gchar *) "GetExtents",
  (GDBusArgInfo **) atspi_component_interface__get_extents_method_in_args,
  (GDBusArgInfo **) atspi_component_interface__get_extents_method_out_args,
  (GDBusAnnotationInfo **) atspi_component_interface__get_extents_method_annotations,
};

const GDBusArgInfo atspi_component_interface__get_position_method_in_args__coord_type_arg =
{
  -1,  /* ref count */
  (gchar *) "coord_type",
  (gchar *) "u",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_component_interface__get_position_method_in_args[] =
{
  &atspi_component_interface__get_position_method_in_args__coord_type_arg,
  NULL,
};

const GDBusArgInfo atspi_component_interface__get_position_method_out_args__x_arg =
{
  -1,  /* ref count */
  (gchar *) "x",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_component_interface__get_position_method_out_args__y_arg =
{
  -1,  /* ref count */
  (gchar *) "y",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_component_interface__get_position_method_out_args[] =
{
  &atspi_component_interface__get_position_method_out_args__x_arg,
  &atspi_component_interface__get_position_method_out_args__y_arg,
  NULL,
};

const GDBusMethodInfo atspi_component_interface__get_position_method =
{
  -1,  /* ref count */
  (gchar *) "GetPosition",
  (GDBusArgInfo **) atspi_component_interface__get_position_method_in_args,
  (GDBusArgInfo **) atspi_component_interface__get_position_method_out_args,
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_component_interface__get_size_method_out_args__width_arg =
{
  -1,  /* ref count */
  (gchar *) "width",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_component_interface__get_size_method_out_args__height_arg =
{
  -1,  /* ref count */
  (gchar *) "height",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_component_interface__get_size_method_out_args[] =
{
  &atspi_component_interface__get_size_method_out_args__width_arg,
  &atspi_component_interface__get_size_method_out_args__height_arg,
  NULL,
};

const GDBusMethodInfo atspi_component_interface__get_size_method =
{
  -1,  /* ref count */
  (gchar *) "GetSize",
  NULL,  /* no in args */
  (GDBusArgInfo **) atspi_component_interface__get_size_method_out_args,
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_component_interface__get_layer_method_out_args__unnamed_arg0_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg0",
  (gchar *) "u",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_component_interface__get_layer_method_out_args[] =
{
  &atspi_component_interface__get_layer_method_out_args__unnamed_arg0_arg,
  NULL,
};

const GDBusMethodInfo atspi_component_interface__get_layer_method =
{
  -1,  /* ref count */
  (gchar *) "GetLayer",
  NULL,  /* no in args */
  (GDBusArgInfo **) atspi_component_interface__get_layer_method_out_args,
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_component_interface__get_mdizorder_method_out_args__unnamed_arg0_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg0",
  (gchar *) "n",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_component_interface__get_mdizorder_method_out_args[] =
{
  &atspi_component_interface__get_mdizorder_method_out_args__unnamed_arg0_arg,
  NULL,
};

const GDBusMethodInfo atspi_component_interface__get_mdizorder_method =
{
  -1,  /* ref count */
  (gchar *) "GetMDIZOrder",
  NULL,  /* no in args */
  (GDBusArgInfo **) atspi_component_interface__get_mdizorder_method_out_args,
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_component_interface__grab_focus_method_out_args__unnamed_arg0_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg0",
  (gchar *) "b",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_component_interface__grab_focus_method_out_args[] =
{
  &atspi_component_interface__grab_focus_method_out_args__unnamed_arg0_arg,
  NULL,
};

const GDBusMethodInfo atspi_component_interface__grab_focus_method =
{
  -1,  /* ref count */
  (gchar *) "GrabFocus",
  NULL,  /* no in args */
  (GDBusArgInfo **) atspi_component_interface__grab_focus_method_out_args,
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_component_interface__get_alpha_method_out_args__unnamed_arg0_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg0",
  (gchar *) "d",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_component_interface__get_alpha_method_out_args[] =
{
  &atspi_component_interface__get_alpha_method_out_args__unnamed_arg0_arg,
  NULL,
};

const GDBusMethodInfo atspi_component_interface__get_alpha_method =
{
  -1,  /* ref count */
  (gchar *) "GetAlpha",
  NULL,  /* no in args */
  (GDBusArgInfo **) atspi_component_interface__get_alpha_method_out_args,
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_component_interface__set_extents_method_in_args__x_arg =
{
  -1,  /* ref count */
  (gchar *) "x",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_component_interface__set_extents_method_in_args__y_arg =
{
  -1,  /* ref count */
  (gchar *) "y",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_component_interface__set_extents_method_in_args__width_arg =
{
  -1,  /* ref count */
  (gchar *) "width",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_component_interface__set_extents_method_in_args__height_arg =
{
  -1,  /* ref count */
  (gchar *) "height",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_component_interface__set_extents_method_in_args__coord_type_arg =
{
  -1,  /* ref count */
  (gchar *) "coord_type",
  (gchar *) "u",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_component_interface__set_extents_method_in_args[] =
{
  &atspi_component_interface__set_extents_method_in_args__x_arg,
  &atspi_component_interface__set_extents_method_in_args__y_arg,
  &atspi_component_interface__set_extents_method_in_args__width_arg,
  &atspi_component_interface__set_extents_method_in_args__height_arg,
  &atspi_component_interface__set_extents_method_in_args__coord_type_arg,
  NULL,
};

const GDBusArgInfo atspi_component_interface__set_extents_method_out_args__unnamed_arg5_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg5",
  (gchar *) "b",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_component_interface__set_extents_method_out_args[] =
{
  &atspi_component_interface__set_extents_method_out_args__unnamed_arg5_arg,
  NULL,
};

const GDBusMethodInfo atspi_component_interface__set_extents_method =
{
  -1,  /* ref count */
  (gchar *) "SetExtents",
  (GDBusArgInfo **) atspi_component_interface__set_extents_method_in_args,
  (GDBusArgInfo **) atspi_component_interface__set_extents_method_out_args,
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_component_interface__set_position_method_in_args__x_arg =
{
  -1,  /* ref count */
  (gchar *) "x",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_component_interface__set_position_method_in_args__y_arg =
{
  -1,  /* ref count */
  (gchar *) "y",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_component_interface__set_position_method_in_args__coord_type_arg =
{
  -1,  /* ref count */
  (gchar *) "coord_type",
  (gchar *) "u",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_component_interface__set_position_method_in_args[] =
{
  &atspi_component_interface__set_position_method_in_args__x_arg,
  &atspi_component_interface__set_position_method_in_args__y_arg,
  &atspi_component_interface__set_position_method_in_args__coord_type_arg,
  NULL,
};

const GDBusArgInfo atspi_component_interface__set_position_method_out_args__unnamed_arg3_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg3",
  (gchar *) "b",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_component_interface__set_position_method_out_args[] =
{
  &atspi_component_interface__set_position_method_out_args__unnamed_arg3_arg,
  NULL,
};

const GDBusMethodInfo atspi_component_interface__set_position_method =
{
  -1,  /* ref count */
  (gchar *) "SetPosition",
  (GDBusArgInfo **) atspi_component_interface__set_position_method_in_args,
  (GDBusArgInfo **) atspi_component_interface__set_position_method_out_args,
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_component_interface__set_size_method_in_args__width_arg =
{
  -1,  /* ref count */
  (gchar *) "width",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_component_interface__set_size_method_in_args__height_arg =
{
  -1,  /* ref count */
  (gchar *) "height",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_component_interface__set_size_method_in_args[] =
{
  &atspi_component_interface__set_size_method_in_args__width_arg,
  &atspi_component_interface__set_size_method_in_args__height_arg,
  NULL,
};

const GDBusArgInfo atspi_component_interface__set_size_method_out_args__unnamed_arg2_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg2",
  (gchar *) "b",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_component_interface__set_size_method_out_args[] =
{
  &atspi_component_interface__set_size_method_out_args__unnamed_arg2_arg,
  NULL,
};

const GDBusMethodInfo atspi_component_interface__set_size_method =
{
  -1,  /* ref count */
  (gchar *) "SetSize",
  (GDBusArgInfo **) atspi_component_interface__set_size_method_in_args,
  (GDBusArgInfo **) atspi_component_interface__set_size_method_out_args,
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_component_interface__scroll_to_method_in_args__type_arg =
{
  -1,  /* ref count */
  (gchar *) "type",
  (gchar *) "u",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_component_interface__scroll_to_method_in_args[] =
{
  &atspi_component_interface__scroll_to_method_in_args__type_arg,
  NULL,
};

const GDBusMethodInfo atspi_component_interface__scroll_to_method =
{
  -1,  /* ref count */
  (gchar *) "ScrollTo",
  (GDBusArgInfo **) atspi_component_interface__scroll_to_method_in_args,
  NULL,  /* no out args */
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_component_interface__scroll_to_point_method_in_args__type_arg =
{
  -1,  /* ref count */
  (gchar *) "type",
  (gchar *) "u",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_component_interface__scroll_to_point_method_in_args__x_arg =
{
  -1,  /* ref count */
  (gchar *) "x",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_component_interface__scroll_to_point_method_in_args__y_arg =
{
  -1,  /* ref count */
  (gchar *) "y",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_component_interface__scroll_to_point_method_in_args[] =
{
  &atspi_component_interface__scroll_to_point_method_in_args__type_arg,
  &atspi_component_interface__scroll_to_point_method_in_args__x_arg,
  &atspi_component_interface__scroll_to_point_method_in_args__y_arg,
  NULL,
};

const GDBusMethodInfo atspi_component_interface__scroll_to_point_method =
{
  -1,  /* ref count */
  (gchar *) "ScrollToPoint",
  (GDBusArgInfo **) atspi_component_interface__scroll_to_point_method_in_args,
  NULL,  /* no out args */
  NULL,  /* no annotations */
};

const GDBusMethodInfo * const atspi_component_interface_methods[] =
{
  &atspi_component_interface__contains_method,
  &atspi_component_interface__get_accessible_at_point_method,
  &atspi_component_interface__get_extents_method,
  &atspi_component_interface__get_position_method,
  &atspi_component_interface__get_size_method,
  &atspi_component_interface__get_layer_method,
  &atspi_component_interface__get_mdizorder_method,
  &atspi_component_interface__grab_focus_method,
  &atspi_component_interface__get_alpha_method,
  &atspi_component_interface__set_extents_method,
  &atspi_component_interface__set_position_method,
  &atspi_component_interface__set_size_method,
  &atspi_component_interface__scroll_to_method,
  &atspi_component_interface__scroll_to_point_method,
  NULL,
};

const GDBusInterfaceInfo atspi_component_interface =
{
  -1,  /* ref count */
  (gchar *) "org.a11y.atspi.Component",
  (GDBusMethodInfo **) atspi_component_interface_methods,
  NULL,  /* no signals */
NULL,  /* no properties */
  NULL,  /* no annotations */
};

