/*
 * This file is generated by gdbus-codegen, do not modify it.
 *
 * The license of this code is the same as for the D-Bus interface description
 * it was derived from. Note that it links to GLib, so must comply with the
 * LGPL linking clauses.
 */

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include "atspi-accessible.h"

#include <string.h>

/* ------------------------------------------------------------------------ */
/* Definitions for org.a11y.atspi.Accessible */

const GDBusArgInfo atspi_accessible_interface__get_child_at_index_method_in_args__index_arg =
{
  -1,  /* ref count */
  (gchar *) "index",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_accessible_interface__get_child_at_index_method_in_args[] =
{
  &atspi_accessible_interface__get_child_at_index_method_in_args__index_arg,
  NULL,
};

const GDBusArgInfo atspi_accessible_interface__get_child_at_index_method_out_args__unnamed_arg1_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg1",
  (gchar *) "(so)",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_accessible_interface__get_child_at_index_method_out_args[] =
{
  &atspi_accessible_interface__get_child_at_index_method_out_args__unnamed_arg1_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_accessible_interface__get_child_at_index_method_annotations__org_qtproject__qt_dbus__qt_type_name__out0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.Out0",
  (gchar *) "QSpiObjectReference",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_accessible_interface__get_child_at_index_method_annotations[] =
{
  &atspi_accessible_interface__get_child_at_index_method_annotations__org_qtproject__qt_dbus__qt_type_name__out0_annotation,
  NULL,
};

const GDBusMethodInfo atspi_accessible_interface__get_child_at_index_method =
{
  -1,  /* ref count */
  (gchar *) "GetChildAtIndex",
  (GDBusArgInfo **) atspi_accessible_interface__get_child_at_index_method_in_args,
  (GDBusArgInfo **) atspi_accessible_interface__get_child_at_index_method_out_args,
  (GDBusAnnotationInfo **) atspi_accessible_interface__get_child_at_index_method_annotations,
};

const GDBusArgInfo atspi_accessible_interface__get_children_method_out_args__unnamed_arg0_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg0",
  (gchar *) "a(so)",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_accessible_interface__get_children_method_out_args[] =
{
  &atspi_accessible_interface__get_children_method_out_args__unnamed_arg0_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_accessible_interface__get_children_method_annotations__org_qtproject__qt_dbus__qt_type_name__out0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.Out0",
  (gchar *) "QSpiObjectReferenceArray",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_accessible_interface__get_children_method_annotations[] =
{
  &atspi_accessible_interface__get_children_method_annotations__org_qtproject__qt_dbus__qt_type_name__out0_annotation,
  NULL,
};

const GDBusMethodInfo atspi_accessible_interface__get_children_method =
{
  -1,  /* ref count */
  (gchar *) "GetChildren",
  NULL,  /* no in args */
  (GDBusArgInfo **) atspi_accessible_interface__get_children_method_out_args,
  (GDBusAnnotationInfo **) atspi_accessible_interface__get_children_method_annotations,
};

const GDBusArgInfo atspi_accessible_interface__get_index_in_parent_method_out_args__unnamed_arg0_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg0",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_accessible_interface__get_index_in_parent_method_out_args[] =
{
  &atspi_accessible_interface__get_index_in_parent_method_out_args__unnamed_arg0_arg,
  NULL,
};

const GDBusMethodInfo atspi_accessible_interface__get_index_in_parent_method =
{
  -1,  /* ref count */
  (gchar *) "GetIndexInParent",
  NULL,  /* no in args */
  (GDBusArgInfo **) atspi_accessible_interface__get_index_in_parent_method_out_args,
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_accessible_interface__get_relation_set_method_out_args__unnamed_arg0_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg0",
  (gchar *) "a(ua(so))",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_accessible_interface__get_relation_set_method_out_args[] =
{
  &atspi_accessible_interface__get_relation_set_method_out_args__unnamed_arg0_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_accessible_interface__get_relation_set_method_annotations__org_qtproject__qt_dbus__qt_type_name__out0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.Out0",
  (gchar *) "QSpiRelationArray",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_accessible_interface__get_relation_set_method_annotations[] =
{
  &atspi_accessible_interface__get_relation_set_method_annotations__org_qtproject__qt_dbus__qt_type_name__out0_annotation,
  NULL,
};

const GDBusMethodInfo atspi_accessible_interface__get_relation_set_method =
{
  -1,  /* ref count */
  (gchar *) "GetRelationSet",
  NULL,  /* no in args */
  (GDBusArgInfo **) atspi_accessible_interface__get_relation_set_method_out_args,
  (GDBusAnnotationInfo **) atspi_accessible_interface__get_relation_set_method_annotations,
};

const GDBusArgInfo atspi_accessible_interface__get_role_method_out_args__unnamed_arg0_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg0",
  (gchar *) "u",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_accessible_interface__get_role_method_out_args[] =
{
  &atspi_accessible_interface__get_role_method_out_args__unnamed_arg0_arg,
  NULL,
};

const GDBusMethodInfo atspi_accessible_interface__get_role_method =
{
  -1,  /* ref count */
  (gchar *) "GetRole",
  NULL,  /* no in args */
  (GDBusArgInfo **) atspi_accessible_interface__get_role_method_out_args,
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_accessible_interface__get_role_name_method_out_args__unnamed_arg0_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg0",
  (gchar *) "s",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_accessible_interface__get_role_name_method_out_args[] =
{
  &atspi_accessible_interface__get_role_name_method_out_args__unnamed_arg0_arg,
  NULL,
};

const GDBusMethodInfo atspi_accessible_interface__get_role_name_method =
{
  -1,  /* ref count */
  (gchar *) "GetRoleName",
  NULL,  /* no in args */
  (GDBusArgInfo **) atspi_accessible_interface__get_role_name_method_out_args,
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_accessible_interface__get_localized_role_name_method_out_args__unnamed_arg0_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg0",
  (gchar *) "s",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_accessible_interface__get_localized_role_name_method_out_args[] =
{
  &atspi_accessible_interface__get_localized_role_name_method_out_args__unnamed_arg0_arg,
  NULL,
};

const GDBusMethodInfo atspi_accessible_interface__get_localized_role_name_method =
{
  -1,  /* ref count */
  (gchar *) "GetLocalizedRoleName",
  NULL,  /* no in args */
  (GDBusArgInfo **) atspi_accessible_interface__get_localized_role_name_method_out_args,
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_accessible_interface__get_state_method_out_args__unnamed_arg0_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg0",
  (gchar *) "au",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_accessible_interface__get_state_method_out_args[] =
{
  &atspi_accessible_interface__get_state_method_out_args__unnamed_arg0_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_accessible_interface__get_state_method_annotations__org_qtproject__qt_dbus__qt_type_name__out0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.Out0",
  (gchar *) "QSpiIntList",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_accessible_interface__get_state_method_annotations[] =
{
  &atspi_accessible_interface__get_state_method_annotations__org_qtproject__qt_dbus__qt_type_name__out0_annotation,
  NULL,
};

const GDBusMethodInfo atspi_accessible_interface__get_state_method =
{
  -1,  /* ref count */
  (gchar *) "GetState",
  NULL,  /* no in args */
  (GDBusArgInfo **) atspi_accessible_interface__get_state_method_out_args,
  (GDBusAnnotationInfo **) atspi_accessible_interface__get_state_method_annotations,
};

const GDBusArgInfo atspi_accessible_interface__get_attributes_method_out_args__unnamed_arg0_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg0",
  (gchar *) "a{ss}",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_accessible_interface__get_attributes_method_out_args[] =
{
  &atspi_accessible_interface__get_attributes_method_out_args__unnamed_arg0_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_accessible_interface__get_attributes_method_annotations__org_qtproject__qt_dbus__qt_type_name__out0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.Out0",
  (gchar *) "QSpiAttributeSet",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_accessible_interface__get_attributes_method_annotations[] =
{
  &atspi_accessible_interface__get_attributes_method_annotations__org_qtproject__qt_dbus__qt_type_name__out0_annotation,
  NULL,
};

const GDBusMethodInfo atspi_accessible_interface__get_attributes_method =
{
  -1,  /* ref count */
  (gchar *) "GetAttributes",
  NULL,  /* no in args */
  (GDBusArgInfo **) atspi_accessible_interface__get_attributes_method_out_args,
  (GDBusAnnotationInfo **) atspi_accessible_interface__get_attributes_method_annotations,
};

const GDBusArgInfo atspi_accessible_interface__get_application_method_out_args__unnamed_arg0_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg0",
  (gchar *) "(so)",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_accessible_interface__get_application_method_out_args[] =
{
  &atspi_accessible_interface__get_application_method_out_args__unnamed_arg0_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_accessible_interface__get_application_method_annotations__org_qtproject__qt_dbus__qt_type_name__out0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.Out0",
  (gchar *) "QSpiObjectReference",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_accessible_interface__get_application_method_annotations[] =
{
  &atspi_accessible_interface__get_application_method_annotations__org_qtproject__qt_dbus__qt_type_name__out0_annotation,
  NULL,
};

const GDBusMethodInfo atspi_accessible_interface__get_application_method =
{
  -1,  /* ref count */
  (gchar *) "GetApplication",
  NULL,  /* no in args */
  (GDBusArgInfo **) atspi_accessible_interface__get_application_method_out_args,
  (GDBusAnnotationInfo **) atspi_accessible_interface__get_application_method_annotations,
};

const GDBusArgInfo atspi_accessible_interface__get_interfaces_method_out_args__unnamed_arg0_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg0",
  (gchar *) "as",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_accessible_interface__get_interfaces_method_out_args[] =
{
  &atspi_accessible_interface__get_interfaces_method_out_args__unnamed_arg0_arg,
  NULL,
};

const GDBusMethodInfo atspi_accessible_interface__get_interfaces_method =
{
  -1,  /* ref count */
  (gchar *) "GetInterfaces",
  NULL,  /* no in args */
  (GDBusArgInfo **) atspi_accessible_interface__get_interfaces_method_out_args,
  NULL,  /* no annotations */
};

const GDBusMethodInfo * const atspi_accessible_interface_methods[] =
{
  &atspi_accessible_interface__get_child_at_index_method,
  &atspi_accessible_interface__get_children_method,
  &atspi_accessible_interface__get_index_in_parent_method,
  &atspi_accessible_interface__get_relation_set_method,
  &atspi_accessible_interface__get_role_method,
  &atspi_accessible_interface__get_role_name_method,
  &atspi_accessible_interface__get_localized_role_name_method,
  &atspi_accessible_interface__get_state_method,
  &atspi_accessible_interface__get_attributes_method,
  &atspi_accessible_interface__get_application_method,
  &atspi_accessible_interface__get_interfaces_method,
  NULL,
};

const GDBusPropertyInfo atspi_accessible_interface__name_property =
{
  -1,  /* ref count */
  (gchar *) "Name",
  (gchar *) "s",
  G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
  NULL,  /* no annotations */
};

const GDBusPropertyInfo atspi_accessible_interface__description_property =
{
  -1,  /* ref count */
  (gchar *) "Description",
  (gchar *) "s",
  G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo atspi_accessible_interface__parent_property_annotations__org_qtproject__qt_dbus__qt_type_name_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName",
  (gchar *) "QSpiObjectReference",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_accessible_interface__parent_property_annotations[] =
{
  &atspi_accessible_interface__parent_property_annotations__org_qtproject__qt_dbus__qt_type_name_annotation,
  NULL,
};

const GDBusPropertyInfo atspi_accessible_interface__parent_property =
{
  -1,  /* ref count */
  (gchar *) "Parent",
  (gchar *) "(so)",
  G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
  (GDBusAnnotationInfo **) atspi_accessible_interface__parent_property_annotations,
};

const GDBusPropertyInfo atspi_accessible_interface__child_count_property =
{
  -1,  /* ref count */
  (gchar *) "ChildCount",
  (gchar *) "i",
  G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
  NULL,  /* no annotations */
};

const GDBusPropertyInfo atspi_accessible_interface__locale_property =
{
  -1,  /* ref count */
  (gchar *) "Locale",
  (gchar *) "s",
  G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
  NULL,  /* no annotations */
};

const GDBusPropertyInfo atspi_accessible_interface__accessible_id_property =
{
  -1,  /* ref count */
  (gchar *) "AccessibleId",
  (gchar *) "s",
  G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
  NULL,  /* no annotations */
};

const GDBusPropertyInfo * const atspi_accessible_interface_properties[] =
{
  &atspi_accessible_interface__name_property,
  &atspi_accessible_interface__description_property,
  &atspi_accessible_interface__parent_property,
  &atspi_accessible_interface__child_count_property,
  &atspi_accessible_interface__locale_property,
  &atspi_accessible_interface__accessible_id_property,
  NULL,
};

const GDBusInterfaceInfo atspi_accessible_interface =
{
  -1,  /* ref count */
  (gchar *) "org.a11y.atspi.Accessible",
  (GDBusMethodInfo **) atspi_accessible_interface_methods,
  NULL,  /* no signals */
  (GDBusPropertyInfo **) atspi_accessible_interface_properties,
  NULL,  /* no annotations */
};

