
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#include "config.h"
#include "gtkcssenumtypes.h"
#include <gtkcss.h>
/* enumerations from "gtkcssenums.h" */
GType
gtk_css_parser_error_get_type (void)
{
  static gsize g_define_type_id__volatile = 0;

  if (g_once_init_enter (&g_define_type_id__volatile))
    {
      static const GEnumValue values[] = {
        { GTK_CSS_PARSER_ERROR_FAILED, "GTK_CSS_PARSER_ERROR_FAILED", "failed" },
        { GTK_CSS_PARSER_ERROR_SYNTAX, "GTK_CSS_PARSER_ERROR_SYNTAX", "syntax" },
        { GTK_CSS_PARSER_ERROR_IMPORT, "GTK_CSS_PARSER_ERROR_IMPORT", "import" },
        { GTK_CSS_PARSER_ERROR_NAME, "GTK_CSS_PARSER_ERROR_NAME", "name" },
        { GTK_CSS_PARSER_ERROR_UNKNOWN_VALUE, "GTK_CSS_PARSER_ERROR_UNKNOWN_VALUE", "unknown-value" },
        { 0, NULL, NULL }
      };
      GType g_define_type_id =
        g_enum_register_static (g_intern_static_string ("GtkCssParserError"), values);
      g_once_init_leave (&g_define_type_id__volatile, g_define_type_id);
    }

  return g_define_type_id__volatile;
}
GType
gtk_css_parser_warning_get_type (void)
{
  static gsize g_define_type_id__volatile = 0;

  if (g_once_init_enter (&g_define_type_id__volatile))
    {
      static const GEnumValue values[] = {
        { GTK_CSS_PARSER_WARNING_DEPRECATED, "GTK_CSS_PARSER_WARNING_DEPRECATED", "deprecated" },
        { GTK_CSS_PARSER_WARNING_SYNTAX, "GTK_CSS_PARSER_WARNING_SYNTAX", "syntax" },
        { GTK_CSS_PARSER_WARNING_UNIMPLEMENTED, "GTK_CSS_PARSER_WARNING_UNIMPLEMENTED", "unimplemented" },
        { 0, NULL, NULL }
      };
      GType g_define_type_id =
        g_enum_register_static (g_intern_static_string ("GtkCssParserWarning"), values);
      g_once_init_leave (&g_define_type_id__volatile, g_define_type_id);
    }

  return g_define_type_id__volatile;
}

/* Generated data ends here */

