/*
 * This file is generated by gdbus-codegen, do not modify it.
 *
 * The license of this code is the same as for the D-Bus interface description
 * it was derived from. Note that it links to GLib, so must comply with the
 * LGPL linking clauses.
 */

#pragma once

#include <gio/gio.h>

G_BEGIN_DECLS

extern  const GDBusInterfaceInfo atspi_cache_interface;

G_END_DECLS
