/*
 * This file is generated by gdbus-codegen, do not modify it.
 *
 * The license of this code is the same as for the D-Bus interface description
 * it was derived from. Note that it links to GLib, so must comply with the
 * LGPL linking clauses.
 */

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include "atspi-event.h"

#include <string.h>

/* ------------------------------------------------------------------------ */
/* Definitions for org.a11y.atspi.Event.Object */

const GDBusArgInfo atspi_event_object_interface__property_change_signal_args__unnamed_arg0_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg0",
  (gchar *) "(suuv)",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_event_object_interface__property_change_signal_args[] =
{
  &atspi_event_object_interface__property_change_signal_args__unnamed_arg0_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_event_object_interface__property_change_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.In0",
  (gchar *) "QSpiEvent",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_event_object_interface__property_change_signal_annotations[] =
{
  &atspi_event_object_interface__property_change_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation,
  NULL,
};

const GDBusSignalInfo atspi_event_object_interface__property_change_signal =
{
  -1,  /* ref count */
  (gchar *) "PropertyChange",
  (GDBusArgInfo **) atspi_event_object_interface__property_change_signal_args,
  (GDBusAnnotationInfo **) atspi_event_object_interface__property_change_signal_annotations,
};

const GDBusArgInfo atspi_event_object_interface__bounds_changed_signal_args__unnamed_arg0_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg0",
  (gchar *) "(suuv)",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_event_object_interface__bounds_changed_signal_args[] =
{
  &atspi_event_object_interface__bounds_changed_signal_args__unnamed_arg0_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_event_object_interface__bounds_changed_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.In0",
  (gchar *) "QSpiEvent",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_event_object_interface__bounds_changed_signal_annotations[] =
{
  &atspi_event_object_interface__bounds_changed_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation,
  NULL,
};

const GDBusSignalInfo atspi_event_object_interface__bounds_changed_signal =
{
  -1,  /* ref count */
  (gchar *) "BoundsChanged",
  (GDBusArgInfo **) atspi_event_object_interface__bounds_changed_signal_args,
  (GDBusAnnotationInfo **) atspi_event_object_interface__bounds_changed_signal_annotations,
};

const GDBusArgInfo atspi_event_object_interface__link_selected_signal_args__unnamed_arg0_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg0",
  (gchar *) "(suuv)",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_event_object_interface__link_selected_signal_args[] =
{
  &atspi_event_object_interface__link_selected_signal_args__unnamed_arg0_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_event_object_interface__link_selected_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.In0",
  (gchar *) "QSpiEvent",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_event_object_interface__link_selected_signal_annotations[] =
{
  &atspi_event_object_interface__link_selected_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation,
  NULL,
};

const GDBusSignalInfo atspi_event_object_interface__link_selected_signal =
{
  -1,  /* ref count */
  (gchar *) "LinkSelected",
  (GDBusArgInfo **) atspi_event_object_interface__link_selected_signal_args,
  (GDBusAnnotationInfo **) atspi_event_object_interface__link_selected_signal_annotations,
};

const GDBusArgInfo atspi_event_object_interface__state_changed_signal_args__unnamed_arg0_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg0",
  (gchar *) "(suuv)",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_event_object_interface__state_changed_signal_args[] =
{
  &atspi_event_object_interface__state_changed_signal_args__unnamed_arg0_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_event_object_interface__state_changed_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.In0",
  (gchar *) "QSpiEvent",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_event_object_interface__state_changed_signal_annotations[] =
{
  &atspi_event_object_interface__state_changed_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation,
  NULL,
};

const GDBusSignalInfo atspi_event_object_interface__state_changed_signal =
{
  -1,  /* ref count */
  (gchar *) "StateChanged",
  (GDBusArgInfo **) atspi_event_object_interface__state_changed_signal_args,
  (GDBusAnnotationInfo **) atspi_event_object_interface__state_changed_signal_annotations,
};

const GDBusArgInfo atspi_event_object_interface__children_changed_signal_args__unnamed_arg0_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg0",
  (gchar *) "(suuv)",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_event_object_interface__children_changed_signal_args[] =
{
  &atspi_event_object_interface__children_changed_signal_args__unnamed_arg0_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_event_object_interface__children_changed_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.In0",
  (gchar *) "QSpiEvent",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_event_object_interface__children_changed_signal_annotations[] =
{
  &atspi_event_object_interface__children_changed_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation,
  NULL,
};

const GDBusSignalInfo atspi_event_object_interface__children_changed_signal =
{
  -1,  /* ref count */
  (gchar *) "ChildrenChanged",
  (GDBusArgInfo **) atspi_event_object_interface__children_changed_signal_args,
  (GDBusAnnotationInfo **) atspi_event_object_interface__children_changed_signal_annotations,
};

const GDBusArgInfo atspi_event_object_interface__visible_data_changed_signal_args__unnamed_arg0_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg0",
  (gchar *) "(suuv)",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_event_object_interface__visible_data_changed_signal_args[] =
{
  &atspi_event_object_interface__visible_data_changed_signal_args__unnamed_arg0_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_event_object_interface__visible_data_changed_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.In0",
  (gchar *) "QSpiEvent",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_event_object_interface__visible_data_changed_signal_annotations[] =
{
  &atspi_event_object_interface__visible_data_changed_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation,
  NULL,
};

const GDBusSignalInfo atspi_event_object_interface__visible_data_changed_signal =
{
  -1,  /* ref count */
  (gchar *) "VisibleDataChanged",
  (GDBusArgInfo **) atspi_event_object_interface__visible_data_changed_signal_args,
  (GDBusAnnotationInfo **) atspi_event_object_interface__visible_data_changed_signal_annotations,
};

const GDBusArgInfo atspi_event_object_interface__selection_changed_signal_args__unnamed_arg0_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg0",
  (gchar *) "(suuv)",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_event_object_interface__selection_changed_signal_args[] =
{
  &atspi_event_object_interface__selection_changed_signal_args__unnamed_arg0_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_event_object_interface__selection_changed_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.In0",
  (gchar *) "QSpiEvent",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_event_object_interface__selection_changed_signal_annotations[] =
{
  &atspi_event_object_interface__selection_changed_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation,
  NULL,
};

const GDBusSignalInfo atspi_event_object_interface__selection_changed_signal =
{
  -1,  /* ref count */
  (gchar *) "SelectionChanged",
  (GDBusArgInfo **) atspi_event_object_interface__selection_changed_signal_args,
  (GDBusAnnotationInfo **) atspi_event_object_interface__selection_changed_signal_annotations,
};

const GDBusArgInfo atspi_event_object_interface__model_changed_signal_args__unnamed_arg0_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg0",
  (gchar *) "(suuv)",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_event_object_interface__model_changed_signal_args[] =
{
  &atspi_event_object_interface__model_changed_signal_args__unnamed_arg0_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_event_object_interface__model_changed_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.In0",
  (gchar *) "QSpiEvent",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_event_object_interface__model_changed_signal_annotations[] =
{
  &atspi_event_object_interface__model_changed_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation,
  NULL,
};

const GDBusSignalInfo atspi_event_object_interface__model_changed_signal =
{
  -1,  /* ref count */
  (gchar *) "ModelChanged",
  (GDBusArgInfo **) atspi_event_object_interface__model_changed_signal_args,
  (GDBusAnnotationInfo **) atspi_event_object_interface__model_changed_signal_annotations,
};

const GDBusArgInfo atspi_event_object_interface__active_descendant_changed_signal_args__unnamed_arg0_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg0",
  (gchar *) "(suuv)",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_event_object_interface__active_descendant_changed_signal_args[] =
{
  &atspi_event_object_interface__active_descendant_changed_signal_args__unnamed_arg0_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_event_object_interface__active_descendant_changed_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.In0",
  (gchar *) "QSpiEvent",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_event_object_interface__active_descendant_changed_signal_annotations[] =
{
  &atspi_event_object_interface__active_descendant_changed_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation,
  NULL,
};

const GDBusSignalInfo atspi_event_object_interface__active_descendant_changed_signal =
{
  -1,  /* ref count */
  (gchar *) "ActiveDescendantChanged",
  (GDBusArgInfo **) atspi_event_object_interface__active_descendant_changed_signal_args,
  (GDBusAnnotationInfo **) atspi_event_object_interface__active_descendant_changed_signal_annotations,
};

const GDBusArgInfo atspi_event_object_interface__row_inserted_signal_args__unnamed_arg0_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg0",
  (gchar *) "(suuv)",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_event_object_interface__row_inserted_signal_args[] =
{
  &atspi_event_object_interface__row_inserted_signal_args__unnamed_arg0_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_event_object_interface__row_inserted_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.In0",
  (gchar *) "QSpiEvent",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_event_object_interface__row_inserted_signal_annotations[] =
{
  &atspi_event_object_interface__row_inserted_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation,
  NULL,
};

const GDBusSignalInfo atspi_event_object_interface__row_inserted_signal =
{
  -1,  /* ref count */
  (gchar *) "RowInserted",
  (GDBusArgInfo **) atspi_event_object_interface__row_inserted_signal_args,
  (GDBusAnnotationInfo **) atspi_event_object_interface__row_inserted_signal_annotations,
};

const GDBusArgInfo atspi_event_object_interface__row_reordered_signal_args__unnamed_arg0_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg0",
  (gchar *) "(suuv)",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_event_object_interface__row_reordered_signal_args[] =
{
  &atspi_event_object_interface__row_reordered_signal_args__unnamed_arg0_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_event_object_interface__row_reordered_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.In0",
  (gchar *) "QSpiEvent",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_event_object_interface__row_reordered_signal_annotations[] =
{
  &atspi_event_object_interface__row_reordered_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation,
  NULL,
};

const GDBusSignalInfo atspi_event_object_interface__row_reordered_signal =
{
  -1,  /* ref count */
  (gchar *) "RowReordered",
  (GDBusArgInfo **) atspi_event_object_interface__row_reordered_signal_args,
  (GDBusAnnotationInfo **) atspi_event_object_interface__row_reordered_signal_annotations,
};

const GDBusArgInfo atspi_event_object_interface__row_deleted_signal_args__unnamed_arg0_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg0",
  (gchar *) "(suuv)",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_event_object_interface__row_deleted_signal_args[] =
{
  &atspi_event_object_interface__row_deleted_signal_args__unnamed_arg0_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_event_object_interface__row_deleted_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.In0",
  (gchar *) "QSpiEvent",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_event_object_interface__row_deleted_signal_annotations[] =
{
  &atspi_event_object_interface__row_deleted_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation,
  NULL,
};

const GDBusSignalInfo atspi_event_object_interface__row_deleted_signal =
{
  -1,  /* ref count */
  (gchar *) "RowDeleted",
  (GDBusArgInfo **) atspi_event_object_interface__row_deleted_signal_args,
  (GDBusAnnotationInfo **) atspi_event_object_interface__row_deleted_signal_annotations,
};

const GDBusArgInfo atspi_event_object_interface__column_inserted_signal_args__unnamed_arg0_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg0",
  (gchar *) "(suuv)",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_event_object_interface__column_inserted_signal_args[] =
{
  &atspi_event_object_interface__column_inserted_signal_args__unnamed_arg0_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_event_object_interface__column_inserted_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.In0",
  (gchar *) "QSpiEvent",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_event_object_interface__column_inserted_signal_annotations[] =
{
  &atspi_event_object_interface__column_inserted_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation,
  NULL,
};

const GDBusSignalInfo atspi_event_object_interface__column_inserted_signal =
{
  -1,  /* ref count */
  (gchar *) "ColumnInserted",
  (GDBusArgInfo **) atspi_event_object_interface__column_inserted_signal_args,
  (GDBusAnnotationInfo **) atspi_event_object_interface__column_inserted_signal_annotations,
};

const GDBusArgInfo atspi_event_object_interface__column_reordered_signal_args__unnamed_arg0_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg0",
  (gchar *) "(suuv)",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_event_object_interface__column_reordered_signal_args[] =
{
  &atspi_event_object_interface__column_reordered_signal_args__unnamed_arg0_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_event_object_interface__column_reordered_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.In0",
  (gchar *) "QSpiEvent",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_event_object_interface__column_reordered_signal_annotations[] =
{
  &atspi_event_object_interface__column_reordered_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation,
  NULL,
};

const GDBusSignalInfo atspi_event_object_interface__column_reordered_signal =
{
  -1,  /* ref count */
  (gchar *) "ColumnReordered",
  (GDBusArgInfo **) atspi_event_object_interface__column_reordered_signal_args,
  (GDBusAnnotationInfo **) atspi_event_object_interface__column_reordered_signal_annotations,
};

const GDBusArgInfo atspi_event_object_interface__column_deleted_signal_args__unnamed_arg0_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg0",
  (gchar *) "(suuv)",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_event_object_interface__column_deleted_signal_args[] =
{
  &atspi_event_object_interface__column_deleted_signal_args__unnamed_arg0_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_event_object_interface__column_deleted_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.In0",
  (gchar *) "QSpiEvent",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_event_object_interface__column_deleted_signal_annotations[] =
{
  &atspi_event_object_interface__column_deleted_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation,
  NULL,
};

const GDBusSignalInfo atspi_event_object_interface__column_deleted_signal =
{
  -1,  /* ref count */
  (gchar *) "ColumnDeleted",
  (GDBusArgInfo **) atspi_event_object_interface__column_deleted_signal_args,
  (GDBusAnnotationInfo **) atspi_event_object_interface__column_deleted_signal_annotations,
};

const GDBusArgInfo atspi_event_object_interface__text_bounds_changed_signal_args__unnamed_arg0_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg0",
  (gchar *) "(suuv)",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_event_object_interface__text_bounds_changed_signal_args[] =
{
  &atspi_event_object_interface__text_bounds_changed_signal_args__unnamed_arg0_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_event_object_interface__text_bounds_changed_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.In0",
  (gchar *) "QSpiEvent",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_event_object_interface__text_bounds_changed_signal_annotations[] =
{
  &atspi_event_object_interface__text_bounds_changed_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation,
  NULL,
};

const GDBusSignalInfo atspi_event_object_interface__text_bounds_changed_signal =
{
  -1,  /* ref count */
  (gchar *) "TextBoundsChanged",
  (GDBusArgInfo **) atspi_event_object_interface__text_bounds_changed_signal_args,
  (GDBusAnnotationInfo **) atspi_event_object_interface__text_bounds_changed_signal_annotations,
};

const GDBusArgInfo atspi_event_object_interface__text_selection_changed_signal_args__unnamed_arg0_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg0",
  (gchar *) "(suuv)",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_event_object_interface__text_selection_changed_signal_args[] =
{
  &atspi_event_object_interface__text_selection_changed_signal_args__unnamed_arg0_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_event_object_interface__text_selection_changed_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.In0",
  (gchar *) "QSpiEvent",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_event_object_interface__text_selection_changed_signal_annotations[] =
{
  &atspi_event_object_interface__text_selection_changed_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation,
  NULL,
};

const GDBusSignalInfo atspi_event_object_interface__text_selection_changed_signal =
{
  -1,  /* ref count */
  (gchar *) "TextSelectionChanged",
  (GDBusArgInfo **) atspi_event_object_interface__text_selection_changed_signal_args,
  (GDBusAnnotationInfo **) atspi_event_object_interface__text_selection_changed_signal_annotations,
};

const GDBusArgInfo atspi_event_object_interface__text_changed_signal_args__unnamed_arg0_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg0",
  (gchar *) "(suuv)",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_event_object_interface__text_changed_signal_args[] =
{
  &atspi_event_object_interface__text_changed_signal_args__unnamed_arg0_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_event_object_interface__text_changed_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.In0",
  (gchar *) "QSpiEvent",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_event_object_interface__text_changed_signal_annotations[] =
{
  &atspi_event_object_interface__text_changed_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation,
  NULL,
};

const GDBusSignalInfo atspi_event_object_interface__text_changed_signal =
{
  -1,  /* ref count */
  (gchar *) "TextChanged",
  (GDBusArgInfo **) atspi_event_object_interface__text_changed_signal_args,
  (GDBusAnnotationInfo **) atspi_event_object_interface__text_changed_signal_annotations,
};

const GDBusArgInfo atspi_event_object_interface__text_attributes_changed_signal_args__unnamed_arg0_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg0",
  (gchar *) "(suuv)",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_event_object_interface__text_attributes_changed_signal_args[] =
{
  &atspi_event_object_interface__text_attributes_changed_signal_args__unnamed_arg0_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_event_object_interface__text_attributes_changed_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.In0",
  (gchar *) "QSpiEvent",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_event_object_interface__text_attributes_changed_signal_annotations[] =
{
  &atspi_event_object_interface__text_attributes_changed_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation,
  NULL,
};

const GDBusSignalInfo atspi_event_object_interface__text_attributes_changed_signal =
{
  -1,  /* ref count */
  (gchar *) "TextAttributesChanged",
  (GDBusArgInfo **) atspi_event_object_interface__text_attributes_changed_signal_args,
  (GDBusAnnotationInfo **) atspi_event_object_interface__text_attributes_changed_signal_annotations,
};

const GDBusArgInfo atspi_event_object_interface__text_caret_moved_signal_args__unnamed_arg0_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg0",
  (gchar *) "(suuv)",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_event_object_interface__text_caret_moved_signal_args[] =
{
  &atspi_event_object_interface__text_caret_moved_signal_args__unnamed_arg0_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_event_object_interface__text_caret_moved_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.In0",
  (gchar *) "QSpiEvent",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_event_object_interface__text_caret_moved_signal_annotations[] =
{
  &atspi_event_object_interface__text_caret_moved_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation,
  NULL,
};

const GDBusSignalInfo atspi_event_object_interface__text_caret_moved_signal =
{
  -1,  /* ref count */
  (gchar *) "TextCaretMoved",
  (GDBusArgInfo **) atspi_event_object_interface__text_caret_moved_signal_args,
  (GDBusAnnotationInfo **) atspi_event_object_interface__text_caret_moved_signal_annotations,
};

const GDBusArgInfo atspi_event_object_interface__attributes_changed_signal_args__unnamed_arg0_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg0",
  (gchar *) "(suuv)",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_event_object_interface__attributes_changed_signal_args[] =
{
  &atspi_event_object_interface__attributes_changed_signal_args__unnamed_arg0_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_event_object_interface__attributes_changed_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.In0",
  (gchar *) "QSpiEvent",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_event_object_interface__attributes_changed_signal_annotations[] =
{
  &atspi_event_object_interface__attributes_changed_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation,
  NULL,
};

const GDBusSignalInfo atspi_event_object_interface__attributes_changed_signal =
{
  -1,  /* ref count */
  (gchar *) "AttributesChanged",
  (GDBusArgInfo **) atspi_event_object_interface__attributes_changed_signal_args,
  (GDBusAnnotationInfo **) atspi_event_object_interface__attributes_changed_signal_annotations,
};

const GDBusSignalInfo * const atspi_event_object_interface_signals[] =
{
  &atspi_event_object_interface__property_change_signal,
  &atspi_event_object_interface__bounds_changed_signal,
  &atspi_event_object_interface__link_selected_signal,
  &atspi_event_object_interface__state_changed_signal,
  &atspi_event_object_interface__children_changed_signal,
  &atspi_event_object_interface__visible_data_changed_signal,
  &atspi_event_object_interface__selection_changed_signal,
  &atspi_event_object_interface__model_changed_signal,
  &atspi_event_object_interface__active_descendant_changed_signal,
  &atspi_event_object_interface__row_inserted_signal,
  &atspi_event_object_interface__row_reordered_signal,
  &atspi_event_object_interface__row_deleted_signal,
  &atspi_event_object_interface__column_inserted_signal,
  &atspi_event_object_interface__column_reordered_signal,
  &atspi_event_object_interface__column_deleted_signal,
  &atspi_event_object_interface__text_bounds_changed_signal,
  &atspi_event_object_interface__text_selection_changed_signal,
  &atspi_event_object_interface__text_changed_signal,
  &atspi_event_object_interface__text_attributes_changed_signal,
  &atspi_event_object_interface__text_caret_moved_signal,
  &atspi_event_object_interface__attributes_changed_signal,
  NULL,
};

const GDBusInterfaceInfo atspi_event_object_interface =
{
  -1,  /* ref count */
  (gchar *) "org.a11y.atspi.Event.Object",
  NULL,  /* no methods */
  (GDBusSignalInfo **) atspi_event_object_interface_signals,
NULL,  /* no properties */
  NULL,  /* no annotations */
};

/* ------------------------------------------------------------------------ */
/* Definitions for org.a11y.atspi.Event.Window */

const GDBusArgInfo atspi_event_window_interface__property_change_signal_args__unnamed_arg0_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg0",
  (gchar *) "(suuv)",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_event_window_interface__property_change_signal_args[] =
{
  &atspi_event_window_interface__property_change_signal_args__unnamed_arg0_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_event_window_interface__property_change_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.In0",
  (gchar *) "QSpiEvent",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_event_window_interface__property_change_signal_annotations[] =
{
  &atspi_event_window_interface__property_change_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation,
  NULL,
};

const GDBusSignalInfo atspi_event_window_interface__property_change_signal =
{
  -1,  /* ref count */
  (gchar *) "PropertyChange",
  (GDBusArgInfo **) atspi_event_window_interface__property_change_signal_args,
  (GDBusAnnotationInfo **) atspi_event_window_interface__property_change_signal_annotations,
};

const GDBusArgInfo atspi_event_window_interface__minimize_signal_args__unnamed_arg0_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg0",
  (gchar *) "(suuv)",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_event_window_interface__minimize_signal_args[] =
{
  &atspi_event_window_interface__minimize_signal_args__unnamed_arg0_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_event_window_interface__minimize_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.In0",
  (gchar *) "QSpiEvent",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_event_window_interface__minimize_signal_annotations[] =
{
  &atspi_event_window_interface__minimize_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation,
  NULL,
};

const GDBusSignalInfo atspi_event_window_interface__minimize_signal =
{
  -1,  /* ref count */
  (gchar *) "Minimize",
  (GDBusArgInfo **) atspi_event_window_interface__minimize_signal_args,
  (GDBusAnnotationInfo **) atspi_event_window_interface__minimize_signal_annotations,
};

const GDBusArgInfo atspi_event_window_interface__maximize_signal_args__unnamed_arg0_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg0",
  (gchar *) "(suuv)",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_event_window_interface__maximize_signal_args[] =
{
  &atspi_event_window_interface__maximize_signal_args__unnamed_arg0_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_event_window_interface__maximize_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.In0",
  (gchar *) "QSpiEvent",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_event_window_interface__maximize_signal_annotations[] =
{
  &atspi_event_window_interface__maximize_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation,
  NULL,
};

const GDBusSignalInfo atspi_event_window_interface__maximize_signal =
{
  -1,  /* ref count */
  (gchar *) "Maximize",
  (GDBusArgInfo **) atspi_event_window_interface__maximize_signal_args,
  (GDBusAnnotationInfo **) atspi_event_window_interface__maximize_signal_annotations,
};

const GDBusArgInfo atspi_event_window_interface__restore_signal_args__unnamed_arg0_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg0",
  (gchar *) "(suuv)",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_event_window_interface__restore_signal_args[] =
{
  &atspi_event_window_interface__restore_signal_args__unnamed_arg0_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_event_window_interface__restore_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.In0",
  (gchar *) "QSpiEvent",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_event_window_interface__restore_signal_annotations[] =
{
  &atspi_event_window_interface__restore_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation,
  NULL,
};

const GDBusSignalInfo atspi_event_window_interface__restore_signal =
{
  -1,  /* ref count */
  (gchar *) "Restore",
  (GDBusArgInfo **) atspi_event_window_interface__restore_signal_args,
  (GDBusAnnotationInfo **) atspi_event_window_interface__restore_signal_annotations,
};

const GDBusArgInfo atspi_event_window_interface__close_signal_args__unnamed_arg0_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg0",
  (gchar *) "(suuv)",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_event_window_interface__close_signal_args[] =
{
  &atspi_event_window_interface__close_signal_args__unnamed_arg0_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_event_window_interface__close_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.In0",
  (gchar *) "QSpiEvent",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_event_window_interface__close_signal_annotations[] =
{
  &atspi_event_window_interface__close_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation,
  NULL,
};

const GDBusSignalInfo atspi_event_window_interface__close_signal =
{
  -1,  /* ref count */
  (gchar *) "Close",
  (GDBusArgInfo **) atspi_event_window_interface__close_signal_args,
  (GDBusAnnotationInfo **) atspi_event_window_interface__close_signal_annotations,
};

const GDBusArgInfo atspi_event_window_interface__create_signal_args__unnamed_arg0_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg0",
  (gchar *) "(suuv)",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_event_window_interface__create_signal_args[] =
{
  &atspi_event_window_interface__create_signal_args__unnamed_arg0_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_event_window_interface__create_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.In0",
  (gchar *) "QSpiEvent",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_event_window_interface__create_signal_annotations[] =
{
  &atspi_event_window_interface__create_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation,
  NULL,
};

const GDBusSignalInfo atspi_event_window_interface__create_signal =
{
  -1,  /* ref count */
  (gchar *) "Create",
  (GDBusArgInfo **) atspi_event_window_interface__create_signal_args,
  (GDBusAnnotationInfo **) atspi_event_window_interface__create_signal_annotations,
};

const GDBusArgInfo atspi_event_window_interface__reparent_signal_args__unnamed_arg0_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg0",
  (gchar *) "(suuv)",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_event_window_interface__reparent_signal_args[] =
{
  &atspi_event_window_interface__reparent_signal_args__unnamed_arg0_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_event_window_interface__reparent_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.In0",
  (gchar *) "QSpiEvent",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_event_window_interface__reparent_signal_annotations[] =
{
  &atspi_event_window_interface__reparent_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation,
  NULL,
};

const GDBusSignalInfo atspi_event_window_interface__reparent_signal =
{
  -1,  /* ref count */
  (gchar *) "Reparent",
  (GDBusArgInfo **) atspi_event_window_interface__reparent_signal_args,
  (GDBusAnnotationInfo **) atspi_event_window_interface__reparent_signal_annotations,
};

const GDBusArgInfo atspi_event_window_interface__desktop_create_signal_args__unnamed_arg0_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg0",
  (gchar *) "(suuv)",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_event_window_interface__desktop_create_signal_args[] =
{
  &atspi_event_window_interface__desktop_create_signal_args__unnamed_arg0_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_event_window_interface__desktop_create_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.In0",
  (gchar *) "QSpiEvent",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_event_window_interface__desktop_create_signal_annotations[] =
{
  &atspi_event_window_interface__desktop_create_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation,
  NULL,
};

const GDBusSignalInfo atspi_event_window_interface__desktop_create_signal =
{
  -1,  /* ref count */
  (gchar *) "DesktopCreate",
  (GDBusArgInfo **) atspi_event_window_interface__desktop_create_signal_args,
  (GDBusAnnotationInfo **) atspi_event_window_interface__desktop_create_signal_annotations,
};

const GDBusArgInfo atspi_event_window_interface__desktop_destroy_signal_args__unnamed_arg0_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg0",
  (gchar *) "(suuv)",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_event_window_interface__desktop_destroy_signal_args[] =
{
  &atspi_event_window_interface__desktop_destroy_signal_args__unnamed_arg0_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_event_window_interface__desktop_destroy_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.In0",
  (gchar *) "QSpiEvent",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_event_window_interface__desktop_destroy_signal_annotations[] =
{
  &atspi_event_window_interface__desktop_destroy_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation,
  NULL,
};

const GDBusSignalInfo atspi_event_window_interface__desktop_destroy_signal =
{
  -1,  /* ref count */
  (gchar *) "DesktopDestroy",
  (GDBusArgInfo **) atspi_event_window_interface__desktop_destroy_signal_args,
  (GDBusAnnotationInfo **) atspi_event_window_interface__desktop_destroy_signal_annotations,
};

const GDBusArgInfo atspi_event_window_interface__destroy_signal_args__unnamed_arg0_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg0",
  (gchar *) "(suuv)",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_event_window_interface__destroy_signal_args[] =
{
  &atspi_event_window_interface__destroy_signal_args__unnamed_arg0_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_event_window_interface__destroy_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.In0",
  (gchar *) "QSpiEvent",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_event_window_interface__destroy_signal_annotations[] =
{
  &atspi_event_window_interface__destroy_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation,
  NULL,
};

const GDBusSignalInfo atspi_event_window_interface__destroy_signal =
{
  -1,  /* ref count */
  (gchar *) "Destroy",
  (GDBusArgInfo **) atspi_event_window_interface__destroy_signal_args,
  (GDBusAnnotationInfo **) atspi_event_window_interface__destroy_signal_annotations,
};

const GDBusArgInfo atspi_event_window_interface__activate_signal_args__unnamed_arg0_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg0",
  (gchar *) "(suuv)",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_event_window_interface__activate_signal_args[] =
{
  &atspi_event_window_interface__activate_signal_args__unnamed_arg0_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_event_window_interface__activate_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.In0",
  (gchar *) "QSpiEvent",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_event_window_interface__activate_signal_annotations[] =
{
  &atspi_event_window_interface__activate_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation,
  NULL,
};

const GDBusSignalInfo atspi_event_window_interface__activate_signal =
{
  -1,  /* ref count */
  (gchar *) "Activate",
  (GDBusArgInfo **) atspi_event_window_interface__activate_signal_args,
  (GDBusAnnotationInfo **) atspi_event_window_interface__activate_signal_annotations,
};

const GDBusArgInfo atspi_event_window_interface__deactivate_signal_args__unnamed_arg0_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg0",
  (gchar *) "(suuv)",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_event_window_interface__deactivate_signal_args[] =
{
  &atspi_event_window_interface__deactivate_signal_args__unnamed_arg0_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_event_window_interface__deactivate_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.In0",
  (gchar *) "QSpiEvent",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_event_window_interface__deactivate_signal_annotations[] =
{
  &atspi_event_window_interface__deactivate_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation,
  NULL,
};

const GDBusSignalInfo atspi_event_window_interface__deactivate_signal =
{
  -1,  /* ref count */
  (gchar *) "Deactivate",
  (GDBusArgInfo **) atspi_event_window_interface__deactivate_signal_args,
  (GDBusAnnotationInfo **) atspi_event_window_interface__deactivate_signal_annotations,
};

const GDBusArgInfo atspi_event_window_interface__raise_signal_args__unnamed_arg0_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg0",
  (gchar *) "(suuv)",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_event_window_interface__raise_signal_args[] =
{
  &atspi_event_window_interface__raise_signal_args__unnamed_arg0_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_event_window_interface__raise_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.In0",
  (gchar *) "QSpiEvent",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_event_window_interface__raise_signal_annotations[] =
{
  &atspi_event_window_interface__raise_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation,
  NULL,
};

const GDBusSignalInfo atspi_event_window_interface__raise_signal =
{
  -1,  /* ref count */
  (gchar *) "Raise",
  (GDBusArgInfo **) atspi_event_window_interface__raise_signal_args,
  (GDBusAnnotationInfo **) atspi_event_window_interface__raise_signal_annotations,
};

const GDBusArgInfo atspi_event_window_interface__lower_signal_args__unnamed_arg0_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg0",
  (gchar *) "(suuv)",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_event_window_interface__lower_signal_args[] =
{
  &atspi_event_window_interface__lower_signal_args__unnamed_arg0_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_event_window_interface__lower_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.In0",
  (gchar *) "QSpiEvent",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_event_window_interface__lower_signal_annotations[] =
{
  &atspi_event_window_interface__lower_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation,
  NULL,
};

const GDBusSignalInfo atspi_event_window_interface__lower_signal =
{
  -1,  /* ref count */
  (gchar *) "Lower",
  (GDBusArgInfo **) atspi_event_window_interface__lower_signal_args,
  (GDBusAnnotationInfo **) atspi_event_window_interface__lower_signal_annotations,
};

const GDBusArgInfo atspi_event_window_interface__move_signal_args__unnamed_arg0_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg0",
  (gchar *) "(suuv)",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_event_window_interface__move_signal_args[] =
{
  &atspi_event_window_interface__move_signal_args__unnamed_arg0_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_event_window_interface__move_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.In0",
  (gchar *) "QSpiEvent",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_event_window_interface__move_signal_annotations[] =
{
  &atspi_event_window_interface__move_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation,
  NULL,
};

const GDBusSignalInfo atspi_event_window_interface__move_signal =
{
  -1,  /* ref count */
  (gchar *) "Move",
  (GDBusArgInfo **) atspi_event_window_interface__move_signal_args,
  (GDBusAnnotationInfo **) atspi_event_window_interface__move_signal_annotations,
};

const GDBusArgInfo atspi_event_window_interface__resize_signal_args__unnamed_arg0_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg0",
  (gchar *) "(suuv)",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_event_window_interface__resize_signal_args[] =
{
  &atspi_event_window_interface__resize_signal_args__unnamed_arg0_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_event_window_interface__resize_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.In0",
  (gchar *) "QSpiEvent",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_event_window_interface__resize_signal_annotations[] =
{
  &atspi_event_window_interface__resize_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation,
  NULL,
};

const GDBusSignalInfo atspi_event_window_interface__resize_signal =
{
  -1,  /* ref count */
  (gchar *) "Resize",
  (GDBusArgInfo **) atspi_event_window_interface__resize_signal_args,
  (GDBusAnnotationInfo **) atspi_event_window_interface__resize_signal_annotations,
};

const GDBusArgInfo atspi_event_window_interface__shade_signal_args__unnamed_arg0_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg0",
  (gchar *) "(suuv)",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_event_window_interface__shade_signal_args[] =
{
  &atspi_event_window_interface__shade_signal_args__unnamed_arg0_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_event_window_interface__shade_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.In0",
  (gchar *) "QSpiEvent",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_event_window_interface__shade_signal_annotations[] =
{
  &atspi_event_window_interface__shade_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation,
  NULL,
};

const GDBusSignalInfo atspi_event_window_interface__shade_signal =
{
  -1,  /* ref count */
  (gchar *) "Shade",
  (GDBusArgInfo **) atspi_event_window_interface__shade_signal_args,
  (GDBusAnnotationInfo **) atspi_event_window_interface__shade_signal_annotations,
};

const GDBusArgInfo atspi_event_window_interface__u_ushade_signal_args__unnamed_arg0_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg0",
  (gchar *) "(suuv)",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_event_window_interface__u_ushade_signal_args[] =
{
  &atspi_event_window_interface__u_ushade_signal_args__unnamed_arg0_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_event_window_interface__u_ushade_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.In0",
  (gchar *) "QSpiEvent",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_event_window_interface__u_ushade_signal_annotations[] =
{
  &atspi_event_window_interface__u_ushade_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation,
  NULL,
};

const GDBusSignalInfo atspi_event_window_interface__u_ushade_signal =
{
  -1,  /* ref count */
  (gchar *) "uUshade",
  (GDBusArgInfo **) atspi_event_window_interface__u_ushade_signal_args,
  (GDBusAnnotationInfo **) atspi_event_window_interface__u_ushade_signal_annotations,
};

const GDBusArgInfo atspi_event_window_interface__restyle_signal_args__unnamed_arg0_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg0",
  (gchar *) "(suuv)",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_event_window_interface__restyle_signal_args[] =
{
  &atspi_event_window_interface__restyle_signal_args__unnamed_arg0_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_event_window_interface__restyle_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.In0",
  (gchar *) "QSpiEvent",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_event_window_interface__restyle_signal_annotations[] =
{
  &atspi_event_window_interface__restyle_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation,
  NULL,
};

const GDBusSignalInfo atspi_event_window_interface__restyle_signal =
{
  -1,  /* ref count */
  (gchar *) "Restyle",
  (GDBusArgInfo **) atspi_event_window_interface__restyle_signal_args,
  (GDBusAnnotationInfo **) atspi_event_window_interface__restyle_signal_annotations,
};

const GDBusSignalInfo * const atspi_event_window_interface_signals[] =
{
  &atspi_event_window_interface__property_change_signal,
  &atspi_event_window_interface__minimize_signal,
  &atspi_event_window_interface__maximize_signal,
  &atspi_event_window_interface__restore_signal,
  &atspi_event_window_interface__close_signal,
  &atspi_event_window_interface__create_signal,
  &atspi_event_window_interface__reparent_signal,
  &atspi_event_window_interface__desktop_create_signal,
  &atspi_event_window_interface__desktop_destroy_signal,
  &atspi_event_window_interface__destroy_signal,
  &atspi_event_window_interface__activate_signal,
  &atspi_event_window_interface__deactivate_signal,
  &atspi_event_window_interface__raise_signal,
  &atspi_event_window_interface__lower_signal,
  &atspi_event_window_interface__move_signal,
  &atspi_event_window_interface__resize_signal,
  &atspi_event_window_interface__shade_signal,
  &atspi_event_window_interface__u_ushade_signal,
  &atspi_event_window_interface__restyle_signal,
  NULL,
};

const GDBusInterfaceInfo atspi_event_window_interface =
{
  -1,  /* ref count */
  (gchar *) "org.a11y.atspi.Event.Window",
  NULL,  /* no methods */
  (GDBusSignalInfo **) atspi_event_window_interface_signals,
NULL,  /* no properties */
  NULL,  /* no annotations */
};

/* ------------------------------------------------------------------------ */
/* Definitions for org.a11y.atspi.Event.Mouse */

const GDBusArgInfo atspi_event_mouse_interface__abs_signal_args__unnamed_arg0_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg0",
  (gchar *) "(suuv)",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_event_mouse_interface__abs_signal_args[] =
{
  &atspi_event_mouse_interface__abs_signal_args__unnamed_arg0_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_event_mouse_interface__abs_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.In0",
  (gchar *) "QSpiEvent",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_event_mouse_interface__abs_signal_annotations[] =
{
  &atspi_event_mouse_interface__abs_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation,
  NULL,
};

const GDBusSignalInfo atspi_event_mouse_interface__abs_signal =
{
  -1,  /* ref count */
  (gchar *) "Abs",
  (GDBusArgInfo **) atspi_event_mouse_interface__abs_signal_args,
  (GDBusAnnotationInfo **) atspi_event_mouse_interface__abs_signal_annotations,
};

const GDBusArgInfo atspi_event_mouse_interface__rel_signal_args__unnamed_arg0_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg0",
  (gchar *) "(suuv)",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_event_mouse_interface__rel_signal_args[] =
{
  &atspi_event_mouse_interface__rel_signal_args__unnamed_arg0_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_event_mouse_interface__rel_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.In0",
  (gchar *) "QSpiEvent",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_event_mouse_interface__rel_signal_annotations[] =
{
  &atspi_event_mouse_interface__rel_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation,
  NULL,
};

const GDBusSignalInfo atspi_event_mouse_interface__rel_signal =
{
  -1,  /* ref count */
  (gchar *) "Rel",
  (GDBusArgInfo **) atspi_event_mouse_interface__rel_signal_args,
  (GDBusAnnotationInfo **) atspi_event_mouse_interface__rel_signal_annotations,
};

const GDBusArgInfo atspi_event_mouse_interface__button_signal_args__unnamed_arg0_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg0",
  (gchar *) "(suuv)",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_event_mouse_interface__button_signal_args[] =
{
  &atspi_event_mouse_interface__button_signal_args__unnamed_arg0_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_event_mouse_interface__button_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.In0",
  (gchar *) "QSpiEvent",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_event_mouse_interface__button_signal_annotations[] =
{
  &atspi_event_mouse_interface__button_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation,
  NULL,
};

const GDBusSignalInfo atspi_event_mouse_interface__button_signal =
{
  -1,  /* ref count */
  (gchar *) "Button",
  (GDBusArgInfo **) atspi_event_mouse_interface__button_signal_args,
  (GDBusAnnotationInfo **) atspi_event_mouse_interface__button_signal_annotations,
};

const GDBusSignalInfo * const atspi_event_mouse_interface_signals[] =
{
  &atspi_event_mouse_interface__abs_signal,
  &atspi_event_mouse_interface__rel_signal,
  &atspi_event_mouse_interface__button_signal,
  NULL,
};

const GDBusInterfaceInfo atspi_event_mouse_interface =
{
  -1,  /* ref count */
  (gchar *) "org.a11y.atspi.Event.Mouse",
  NULL,  /* no methods */
  (GDBusSignalInfo **) atspi_event_mouse_interface_signals,
NULL,  /* no properties */
  NULL,  /* no annotations */
};

/* ------------------------------------------------------------------------ */
/* Definitions for org.a11y.atspi.Event.Keyboard */

const GDBusArgInfo atspi_event_keyboard_interface__modifiers_signal_args__unnamed_arg0_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg0",
  (gchar *) "(suuv)",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_event_keyboard_interface__modifiers_signal_args[] =
{
  &atspi_event_keyboard_interface__modifiers_signal_args__unnamed_arg0_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_event_keyboard_interface__modifiers_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.In0",
  (gchar *) "QSpiEvent",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_event_keyboard_interface__modifiers_signal_annotations[] =
{
  &atspi_event_keyboard_interface__modifiers_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation,
  NULL,
};

const GDBusSignalInfo atspi_event_keyboard_interface__modifiers_signal =
{
  -1,  /* ref count */
  (gchar *) "Modifiers",
  (GDBusArgInfo **) atspi_event_keyboard_interface__modifiers_signal_args,
  (GDBusAnnotationInfo **) atspi_event_keyboard_interface__modifiers_signal_annotations,
};

const GDBusSignalInfo * const atspi_event_keyboard_interface_signals[] =
{
  &atspi_event_keyboard_interface__modifiers_signal,
  NULL,
};

const GDBusInterfaceInfo atspi_event_keyboard_interface =
{
  -1,  /* ref count */
  (gchar *) "org.a11y.atspi.Event.Keyboard",
  NULL,  /* no methods */
  (GDBusSignalInfo **) atspi_event_keyboard_interface_signals,
NULL,  /* no properties */
  NULL,  /* no annotations */
};

/* ------------------------------------------------------------------------ */
/* Definitions for org.a11y.atspi.Event.Terminal */

const GDBusArgInfo atspi_event_terminal_interface__line_changed_signal_args__unnamed_arg0_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg0",
  (gchar *) "(suuv)",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_event_terminal_interface__line_changed_signal_args[] =
{
  &atspi_event_terminal_interface__line_changed_signal_args__unnamed_arg0_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_event_terminal_interface__line_changed_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.In0",
  (gchar *) "QSpiEvent",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_event_terminal_interface__line_changed_signal_annotations[] =
{
  &atspi_event_terminal_interface__line_changed_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation,
  NULL,
};

const GDBusSignalInfo atspi_event_terminal_interface__line_changed_signal =
{
  -1,  /* ref count */
  (gchar *) "LineChanged",
  (GDBusArgInfo **) atspi_event_terminal_interface__line_changed_signal_args,
  (GDBusAnnotationInfo **) atspi_event_terminal_interface__line_changed_signal_annotations,
};

const GDBusArgInfo atspi_event_terminal_interface__columncount_changed_signal_args__unnamed_arg0_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg0",
  (gchar *) "(suuv)",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_event_terminal_interface__columncount_changed_signal_args[] =
{
  &atspi_event_terminal_interface__columncount_changed_signal_args__unnamed_arg0_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_event_terminal_interface__columncount_changed_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.In0",
  (gchar *) "QSpiEvent",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_event_terminal_interface__columncount_changed_signal_annotations[] =
{
  &atspi_event_terminal_interface__columncount_changed_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation,
  NULL,
};

const GDBusSignalInfo atspi_event_terminal_interface__columncount_changed_signal =
{
  -1,  /* ref count */
  (gchar *) "ColumncountChanged",
  (GDBusArgInfo **) atspi_event_terminal_interface__columncount_changed_signal_args,
  (GDBusAnnotationInfo **) atspi_event_terminal_interface__columncount_changed_signal_annotations,
};

const GDBusArgInfo atspi_event_terminal_interface__linecount_changed_signal_args__unnamed_arg0_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg0",
  (gchar *) "(suuv)",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_event_terminal_interface__linecount_changed_signal_args[] =
{
  &atspi_event_terminal_interface__linecount_changed_signal_args__unnamed_arg0_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_event_terminal_interface__linecount_changed_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.In0",
  (gchar *) "QSpiEvent",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_event_terminal_interface__linecount_changed_signal_annotations[] =
{
  &atspi_event_terminal_interface__linecount_changed_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation,
  NULL,
};

const GDBusSignalInfo atspi_event_terminal_interface__linecount_changed_signal =
{
  -1,  /* ref count */
  (gchar *) "LinecountChanged",
  (GDBusArgInfo **) atspi_event_terminal_interface__linecount_changed_signal_args,
  (GDBusAnnotationInfo **) atspi_event_terminal_interface__linecount_changed_signal_annotations,
};

const GDBusArgInfo atspi_event_terminal_interface__application_changed_signal_args__unnamed_arg0_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg0",
  (gchar *) "(suuv)",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_event_terminal_interface__application_changed_signal_args[] =
{
  &atspi_event_terminal_interface__application_changed_signal_args__unnamed_arg0_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_event_terminal_interface__application_changed_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.In0",
  (gchar *) "QSpiEvent",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_event_terminal_interface__application_changed_signal_annotations[] =
{
  &atspi_event_terminal_interface__application_changed_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation,
  NULL,
};

const GDBusSignalInfo atspi_event_terminal_interface__application_changed_signal =
{
  -1,  /* ref count */
  (gchar *) "ApplicationChanged",
  (GDBusArgInfo **) atspi_event_terminal_interface__application_changed_signal_args,
  (GDBusAnnotationInfo **) atspi_event_terminal_interface__application_changed_signal_annotations,
};

const GDBusArgInfo atspi_event_terminal_interface__charwidth_changed_signal_args__unnamed_arg0_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg0",
  (gchar *) "(suuv)",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_event_terminal_interface__charwidth_changed_signal_args[] =
{
  &atspi_event_terminal_interface__charwidth_changed_signal_args__unnamed_arg0_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_event_terminal_interface__charwidth_changed_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.In0",
  (gchar *) "QSpiEvent",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_event_terminal_interface__charwidth_changed_signal_annotations[] =
{
  &atspi_event_terminal_interface__charwidth_changed_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation,
  NULL,
};

const GDBusSignalInfo atspi_event_terminal_interface__charwidth_changed_signal =
{
  -1,  /* ref count */
  (gchar *) "CharwidthChanged",
  (GDBusArgInfo **) atspi_event_terminal_interface__charwidth_changed_signal_args,
  (GDBusAnnotationInfo **) atspi_event_terminal_interface__charwidth_changed_signal_annotations,
};

const GDBusSignalInfo * const atspi_event_terminal_interface_signals[] =
{
  &atspi_event_terminal_interface__line_changed_signal,
  &atspi_event_terminal_interface__columncount_changed_signal,
  &atspi_event_terminal_interface__linecount_changed_signal,
  &atspi_event_terminal_interface__application_changed_signal,
  &atspi_event_terminal_interface__charwidth_changed_signal,
  NULL,
};

const GDBusInterfaceInfo atspi_event_terminal_interface =
{
  -1,  /* ref count */
  (gchar *) "org.a11y.atspi.Event.Terminal",
  NULL,  /* no methods */
  (GDBusSignalInfo **) atspi_event_terminal_interface_signals,
NULL,  /* no properties */
  NULL,  /* no annotations */
};

/* ------------------------------------------------------------------------ */
/* Definitions for org.a11y.atspi.Event.Document */

const GDBusArgInfo atspi_event_document_interface__load_complete_signal_args__unnamed_arg0_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg0",
  (gchar *) "(suuv)",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_event_document_interface__load_complete_signal_args[] =
{
  &atspi_event_document_interface__load_complete_signal_args__unnamed_arg0_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_event_document_interface__load_complete_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.In0",
  (gchar *) "QSpiEvent",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_event_document_interface__load_complete_signal_annotations[] =
{
  &atspi_event_document_interface__load_complete_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation,
  NULL,
};

const GDBusSignalInfo atspi_event_document_interface__load_complete_signal =
{
  -1,  /* ref count */
  (gchar *) "LoadComplete",
  (GDBusArgInfo **) atspi_event_document_interface__load_complete_signal_args,
  (GDBusAnnotationInfo **) atspi_event_document_interface__load_complete_signal_annotations,
};

const GDBusArgInfo atspi_event_document_interface__reload_signal_args__unnamed_arg0_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg0",
  (gchar *) "(suuv)",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_event_document_interface__reload_signal_args[] =
{
  &atspi_event_document_interface__reload_signal_args__unnamed_arg0_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_event_document_interface__reload_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.In0",
  (gchar *) "QSpiEvent",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_event_document_interface__reload_signal_annotations[] =
{
  &atspi_event_document_interface__reload_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation,
  NULL,
};

const GDBusSignalInfo atspi_event_document_interface__reload_signal =
{
  -1,  /* ref count */
  (gchar *) "Reload",
  (GDBusArgInfo **) atspi_event_document_interface__reload_signal_args,
  (GDBusAnnotationInfo **) atspi_event_document_interface__reload_signal_annotations,
};

const GDBusArgInfo atspi_event_document_interface__load_stopped_signal_args__unnamed_arg0_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg0",
  (gchar *) "(suuv)",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_event_document_interface__load_stopped_signal_args[] =
{
  &atspi_event_document_interface__load_stopped_signal_args__unnamed_arg0_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_event_document_interface__load_stopped_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.In0",
  (gchar *) "QSpiEvent",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_event_document_interface__load_stopped_signal_annotations[] =
{
  &atspi_event_document_interface__load_stopped_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation,
  NULL,
};

const GDBusSignalInfo atspi_event_document_interface__load_stopped_signal =
{
  -1,  /* ref count */
  (gchar *) "LoadStopped",
  (GDBusArgInfo **) atspi_event_document_interface__load_stopped_signal_args,
  (GDBusAnnotationInfo **) atspi_event_document_interface__load_stopped_signal_annotations,
};

const GDBusArgInfo atspi_event_document_interface__content_changed_signal_args__unnamed_arg0_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg0",
  (gchar *) "(suuv)",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_event_document_interface__content_changed_signal_args[] =
{
  &atspi_event_document_interface__content_changed_signal_args__unnamed_arg0_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_event_document_interface__content_changed_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.In0",
  (gchar *) "QSpiEvent",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_event_document_interface__content_changed_signal_annotations[] =
{
  &atspi_event_document_interface__content_changed_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation,
  NULL,
};

const GDBusSignalInfo atspi_event_document_interface__content_changed_signal =
{
  -1,  /* ref count */
  (gchar *) "ContentChanged",
  (GDBusArgInfo **) atspi_event_document_interface__content_changed_signal_args,
  (GDBusAnnotationInfo **) atspi_event_document_interface__content_changed_signal_annotations,
};

const GDBusArgInfo atspi_event_document_interface__attributes_changed_signal_args__unnamed_arg0_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg0",
  (gchar *) "(suuv)",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_event_document_interface__attributes_changed_signal_args[] =
{
  &atspi_event_document_interface__attributes_changed_signal_args__unnamed_arg0_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_event_document_interface__attributes_changed_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.In0",
  (gchar *) "QSpiEvent",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_event_document_interface__attributes_changed_signal_annotations[] =
{
  &atspi_event_document_interface__attributes_changed_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation,
  NULL,
};

const GDBusSignalInfo atspi_event_document_interface__attributes_changed_signal =
{
  -1,  /* ref count */
  (gchar *) "AttributesChanged",
  (GDBusArgInfo **) atspi_event_document_interface__attributes_changed_signal_args,
  (GDBusAnnotationInfo **) atspi_event_document_interface__attributes_changed_signal_annotations,
};

const GDBusArgInfo atspi_event_document_interface__page_changed_signal_args__unnamed_arg0_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg0",
  (gchar *) "(suuv)",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_event_document_interface__page_changed_signal_args[] =
{
  &atspi_event_document_interface__page_changed_signal_args__unnamed_arg0_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_event_document_interface__page_changed_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.In0",
  (gchar *) "QSpiEvent",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_event_document_interface__page_changed_signal_annotations[] =
{
  &atspi_event_document_interface__page_changed_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation,
  NULL,
};

const GDBusSignalInfo atspi_event_document_interface__page_changed_signal =
{
  -1,  /* ref count */
  (gchar *) "PageChanged",
  (GDBusArgInfo **) atspi_event_document_interface__page_changed_signal_args,
  (GDBusAnnotationInfo **) atspi_event_document_interface__page_changed_signal_annotations,
};

const GDBusSignalInfo * const atspi_event_document_interface_signals[] =
{
  &atspi_event_document_interface__load_complete_signal,
  &atspi_event_document_interface__reload_signal,
  &atspi_event_document_interface__load_stopped_signal,
  &atspi_event_document_interface__content_changed_signal,
  &atspi_event_document_interface__attributes_changed_signal,
  &atspi_event_document_interface__page_changed_signal,
  NULL,
};

const GDBusInterfaceInfo atspi_event_document_interface =
{
  -1,  /* ref count */
  (gchar *) "org.a11y.atspi.Event.Document",
  NULL,  /* no methods */
  (GDBusSignalInfo **) atspi_event_document_interface_signals,
NULL,  /* no properties */
  NULL,  /* no annotations */
};

/* ------------------------------------------------------------------------ */
/* Definitions for org.a11y.atspi.Event.Focus */

const GDBusArgInfo atspi_event_focus_interface__focus_signal_args__unnamed_arg0_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg0",
  (gchar *) "(suuv)",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_event_focus_interface__focus_signal_args[] =
{
  &atspi_event_focus_interface__focus_signal_args__unnamed_arg0_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_event_focus_interface__focus_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.In0",
  (gchar *) "QSpiEvent",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_event_focus_interface__focus_signal_annotations[] =
{
  &atspi_event_focus_interface__focus_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation,
  NULL,
};

const GDBusSignalInfo atspi_event_focus_interface__focus_signal =
{
  -1,  /* ref count */
  (gchar *) "Focus",
  (GDBusArgInfo **) atspi_event_focus_interface__focus_signal_args,
  (GDBusAnnotationInfo **) atspi_event_focus_interface__focus_signal_annotations,
};

const GDBusSignalInfo * const atspi_event_focus_interface_signals[] =
{
  &atspi_event_focus_interface__focus_signal,
  NULL,
};

const GDBusInterfaceInfo atspi_event_focus_interface =
{
  -1,  /* ref count */
  (gchar *) "org.a11y.atspi.Event.Focus",
  NULL,  /* no methods */
  (GDBusSignalInfo **) atspi_event_focus_interface_signals,
NULL,  /* no properties */
  NULL,  /* no annotations */
};

