
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#if !defined (__GTK_H_INSIDE__) && !defined (GTK_COMPILATION)
#error "Only <gtk/gtk.h> can be included directly."
#endif

#ifndef __GTK_PRIVATE_TYPE_BUILTINS_H__
#define __GTK_PRIVATE_TYPE_BUILTINS_H__

#include <glib-object.h>
#include <gdk/gdk.h>

G_BEGIN_DECLS

/* enumerations from "gtkcsstypesprivate.h" */
GType _gtk_css_affects_get_type (void) G_GNUC_CONST;
#define GTK_TYPE_CSS_AFFECTS (_gtk_css_affects_get_type ())

/* enumerations from "gtkplacessidebarprivate.h" */
GType _gtk_places_open_flags_get_type (void) G_GNUC_CONST;
#define GTK_TYPE_PLACES_OPEN_FLAGS (_gtk_places_open_flags_get_type ())
GType _gtk_places_section_type_get_type (void) G_GNUC_CONST;
#define GTK_TYPE_PLACES_SECTION_TYPE (_gtk_places_section_type_get_type ())
GType _gtk_places_place_type_get_type (void) G_GNUC_CONST;
#define GTK_TYPE_PLACES_PLACE_TYPE (_gtk_places_place_type_get_type ())

/* enumerations from "gtktexthandleprivate.h" */
GType _gtk_text_handle_role_get_type (void) G_GNUC_CONST;
#define GTK_TYPE_TEXT_HANDLE_ROLE (_gtk_text_handle_role_get_type ())
G_END_DECLS

#endif /* __GTK_PRIVATE_TYPE_BUILTINS_H__ */

/* Generated data ends here */

