#pragma once

void do_simplify   (int *argc, const char ***argv);
void do_validate   (int *argc, const char ***argv);
void do_enumerate  (int *argc, const char ***argv);
void do_preview    (int *argc, const char ***argv);
void do_screenshot (int *argc, const char ***argv);
