/*
 * This file is generated by gdbus-codegen, do not modify it.
 *
 * The license of this code is the same as for the D-Bus interface description
 * it was derived from. Note that it links to GLib, so must comply with the
 * LGPL linking clauses.
 */

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include "atspi-document.h"

#include <string.h>

/* ------------------------------------------------------------------------ */
/* Definitions for org.a11y.atspi.Document */

const GDBusArgInfo atspi_document_interface__get_locale_method_out_args__unnamed_arg0_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg0",
  (gchar *) "s",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_document_interface__get_locale_method_out_args[] =
{
  &atspi_document_interface__get_locale_method_out_args__unnamed_arg0_arg,
  NULL,
};

const GDBusMethodInfo atspi_document_interface__get_locale_method =
{
  -1,  /* ref count */
  (gchar *) "GetLocale",
  NULL,  /* no in args */
  (GDBusArgInfo **) atspi_document_interface__get_locale_method_out_args,
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_document_interface__get_attribute_value_method_in_args__attributename_arg =
{
  -1,  /* ref count */
  (gchar *) "attributename",
  (gchar *) "s",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_document_interface__get_attribute_value_method_in_args[] =
{
  &atspi_document_interface__get_attribute_value_method_in_args__attributename_arg,
  NULL,
};

const GDBusArgInfo atspi_document_interface__get_attribute_value_method_out_args__unnamed_arg1_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg1",
  (gchar *) "s",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_document_interface__get_attribute_value_method_out_args[] =
{
  &atspi_document_interface__get_attribute_value_method_out_args__unnamed_arg1_arg,
  NULL,
};

const GDBusMethodInfo atspi_document_interface__get_attribute_value_method =
{
  -1,  /* ref count */
  (gchar *) "GetAttributeValue",
  (GDBusArgInfo **) atspi_document_interface__get_attribute_value_method_in_args,
  (GDBusArgInfo **) atspi_document_interface__get_attribute_value_method_out_args,
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_document_interface__get_attributes_method_out_args__unnamed_arg0_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg0",
  (gchar *) "{ss}",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_document_interface__get_attributes_method_out_args[] =
{
  &atspi_document_interface__get_attributes_method_out_args__unnamed_arg0_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_document_interface__get_attributes_method_annotations__org_qtproject__qt_dbus__qt_type_name__out0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.Out0",
  (gchar *) "QSpiAttributeSet",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_document_interface__get_attributes_method_annotations[] =
{
  &atspi_document_interface__get_attributes_method_annotations__org_qtproject__qt_dbus__qt_type_name__out0_annotation,
  NULL,
};

const GDBusMethodInfo atspi_document_interface__get_attributes_method =
{
  -1,  /* ref count */
  (gchar *) "GetAttributes",
  NULL,  /* no in args */
  (GDBusArgInfo **) atspi_document_interface__get_attributes_method_out_args,
  (GDBusAnnotationInfo **) atspi_document_interface__get_attributes_method_annotations,
};

const GDBusMethodInfo * const atspi_document_interface_methods[] =
{
  &atspi_document_interface__get_locale_method,
  &atspi_document_interface__get_attribute_value_method,
  &atspi_document_interface__get_attributes_method,
  NULL,
};

const GDBusPropertyInfo atspi_document_interface__current_page_number_property =
{
  -1,  /* ref count */
  (gchar *) "CurrentPageNumber",
  (gchar *) "i",
  G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
  NULL,  /* no annotations */
};

const GDBusPropertyInfo atspi_document_interface__page_count_property =
{
  -1,  /* ref count */
  (gchar *) "PageCount",
  (gchar *) "i",
  G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
  NULL,  /* no annotations */
};

const GDBusPropertyInfo * const atspi_document_interface_properties[] =
{
  &atspi_document_interface__current_page_number_property,
  &atspi_document_interface__page_count_property,
  NULL,
};

const GDBusInterfaceInfo atspi_document_interface =
{
  -1,  /* ref count */
  (gchar *) "org.a11y.atspi.Document",
  (GDBusMethodInfo **) atspi_document_interface_methods,
  NULL,  /* no signals */
  (GDBusPropertyInfo **) atspi_document_interface_properties,
  NULL,  /* no annotations */
};

