/*
 * This file is generated by gdbus-codegen, do not modify it.
 *
 * The license of this code is the same as for the D-Bus interface description
 * it was derived from. Note that it links to GLib, so must comply with the
 * LGPL linking clauses.
 */

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include "atspi-collection.h"

#include <string.h>

/* ------------------------------------------------------------------------ */
/* Definitions for org.a11y.atspi.Collection */

const GDBusArgInfo atspi_collection_interface__get_matches_method_in_args__rule_arg =
{
  -1,  /* ref count */
  (gchar *) "rule",
  (gchar *) "(auuasuauusub)",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_collection_interface__get_matches_method_in_args__sortby_arg =
{
  -1,  /* ref count */
  (gchar *) "sortby",
  (gchar *) "u",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_collection_interface__get_matches_method_in_args__count_arg =
{
  -1,  /* ref count */
  (gchar *) "count",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_collection_interface__get_matches_method_in_args__traverse_arg =
{
  -1,  /* ref count */
  (gchar *) "traverse",
  (gchar *) "b",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_collection_interface__get_matches_method_in_args[] =
{
  &atspi_collection_interface__get_matches_method_in_args__rule_arg,
  &atspi_collection_interface__get_matches_method_in_args__sortby_arg,
  &atspi_collection_interface__get_matches_method_in_args__count_arg,
  &atspi_collection_interface__get_matches_method_in_args__traverse_arg,
  NULL,
};

const GDBusArgInfo atspi_collection_interface__get_matches_method_out_args__unnamed_arg4_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg4",
  (gchar *) "a(so)",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_collection_interface__get_matches_method_out_args[] =
{
  &atspi_collection_interface__get_matches_method_out_args__unnamed_arg4_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_collection_interface__get_matches_method_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.In0",
  (gchar *) "QSpiMatchRule",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo atspi_collection_interface__get_matches_method_annotations__org_qtproject__qt_dbus__qt_type_name__out0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.Out0",
  (gchar *) "QSpiReferenceSet",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_collection_interface__get_matches_method_annotations[] =
{
  &atspi_collection_interface__get_matches_method_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation,
  &atspi_collection_interface__get_matches_method_annotations__org_qtproject__qt_dbus__qt_type_name__out0_annotation,
  NULL,
};

const GDBusMethodInfo atspi_collection_interface__get_matches_method =
{
  -1,  /* ref count */
  (gchar *) "GetMatches",
  (GDBusArgInfo **) atspi_collection_interface__get_matches_method_in_args,
  (GDBusArgInfo **) atspi_collection_interface__get_matches_method_out_args,
  (GDBusAnnotationInfo **) atspi_collection_interface__get_matches_method_annotations,
};

const GDBusArgInfo atspi_collection_interface__get_matches_to_method_in_args__current_object_arg =
{
  -1,  /* ref count */
  (gchar *) "current_object",
  (gchar *) "o",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_collection_interface__get_matches_to_method_in_args__rule_arg =
{
  -1,  /* ref count */
  (gchar *) "rule",
  (gchar *) "(auuasuauusub)",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_collection_interface__get_matches_to_method_in_args__sortby_arg =
{
  -1,  /* ref count */
  (gchar *) "sortby",
  (gchar *) "u",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_collection_interface__get_matches_to_method_in_args__tree_arg =
{
  -1,  /* ref count */
  (gchar *) "tree",
  (gchar *) "u",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_collection_interface__get_matches_to_method_in_args__limit_scope_arg =
{
  -1,  /* ref count */
  (gchar *) "limit_scope",
  (gchar *) "b",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_collection_interface__get_matches_to_method_in_args__count_arg =
{
  -1,  /* ref count */
  (gchar *) "count",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_collection_interface__get_matches_to_method_in_args__traverse_arg =
{
  -1,  /* ref count */
  (gchar *) "traverse",
  (gchar *) "b",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_collection_interface__get_matches_to_method_in_args[] =
{
  &atspi_collection_interface__get_matches_to_method_in_args__current_object_arg,
  &atspi_collection_interface__get_matches_to_method_in_args__rule_arg,
  &atspi_collection_interface__get_matches_to_method_in_args__sortby_arg,
  &atspi_collection_interface__get_matches_to_method_in_args__tree_arg,
  &atspi_collection_interface__get_matches_to_method_in_args__limit_scope_arg,
  &atspi_collection_interface__get_matches_to_method_in_args__count_arg,
  &atspi_collection_interface__get_matches_to_method_in_args__traverse_arg,
  NULL,
};

const GDBusArgInfo atspi_collection_interface__get_matches_to_method_out_args__unnamed_arg7_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg7",
  (gchar *) "a(so)",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_collection_interface__get_matches_to_method_out_args[] =
{
  &atspi_collection_interface__get_matches_to_method_out_args__unnamed_arg7_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_collection_interface__get_matches_to_method_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.In0",
  (gchar *) "QSpiObjectReference",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo atspi_collection_interface__get_matches_to_method_annotations__org_qtproject__qt_dbus__qt_type_name__in1_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.In1",
  (gchar *) "QSpiMatchRule",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo atspi_collection_interface__get_matches_to_method_annotations__org_qtproject__qt_dbus__qt_type_name__out0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.Out0",
  (gchar *) "QSpiReferenceSet",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_collection_interface__get_matches_to_method_annotations[] =
{
  &atspi_collection_interface__get_matches_to_method_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation,
  &atspi_collection_interface__get_matches_to_method_annotations__org_qtproject__qt_dbus__qt_type_name__in1_annotation,
  &atspi_collection_interface__get_matches_to_method_annotations__org_qtproject__qt_dbus__qt_type_name__out0_annotation,
  NULL,
};

const GDBusMethodInfo atspi_collection_interface__get_matches_to_method =
{
  -1,  /* ref count */
  (gchar *) "GetMatchesTo",
  (GDBusArgInfo **) atspi_collection_interface__get_matches_to_method_in_args,
  (GDBusArgInfo **) atspi_collection_interface__get_matches_to_method_out_args,
  (GDBusAnnotationInfo **) atspi_collection_interface__get_matches_to_method_annotations,
};

const GDBusArgInfo atspi_collection_interface__get_matches_from_method_in_args__current_object_arg =
{
  -1,  /* ref count */
  (gchar *) "current_object",
  (gchar *) "o",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_collection_interface__get_matches_from_method_in_args__rule_arg =
{
  -1,  /* ref count */
  (gchar *) "rule",
  (gchar *) "(auuasuauusub)",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_collection_interface__get_matches_from_method_in_args__sortby_arg =
{
  -1,  /* ref count */
  (gchar *) "sortby",
  (gchar *) "u",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_collection_interface__get_matches_from_method_in_args__tree_arg =
{
  -1,  /* ref count */
  (gchar *) "tree",
  (gchar *) "u",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_collection_interface__get_matches_from_method_in_args__count_arg =
{
  -1,  /* ref count */
  (gchar *) "count",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_collection_interface__get_matches_from_method_in_args__traverse_arg =
{
  -1,  /* ref count */
  (gchar *) "traverse",
  (gchar *) "b",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_collection_interface__get_matches_from_method_in_args[] =
{
  &atspi_collection_interface__get_matches_from_method_in_args__current_object_arg,
  &atspi_collection_interface__get_matches_from_method_in_args__rule_arg,
  &atspi_collection_interface__get_matches_from_method_in_args__sortby_arg,
  &atspi_collection_interface__get_matches_from_method_in_args__tree_arg,
  &atspi_collection_interface__get_matches_from_method_in_args__count_arg,
  &atspi_collection_interface__get_matches_from_method_in_args__traverse_arg,
  NULL,
};

const GDBusArgInfo atspi_collection_interface__get_matches_from_method_out_args__unnamed_arg6_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg6",
  (gchar *) "a(so)",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_collection_interface__get_matches_from_method_out_args[] =
{
  &atspi_collection_interface__get_matches_from_method_out_args__unnamed_arg6_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_collection_interface__get_matches_from_method_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.In0",
  (gchar *) "QSpiObjectReference",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo atspi_collection_interface__get_matches_from_method_annotations__org_qtproject__qt_dbus__qt_type_name__in1_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.In1",
  (gchar *) "QSpiMatchRule",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo atspi_collection_interface__get_matches_from_method_annotations__org_qtproject__qt_dbus__qt_type_name__out0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.Out0",
  (gchar *) "QSpiReferenceSet",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_collection_interface__get_matches_from_method_annotations[] =
{
  &atspi_collection_interface__get_matches_from_method_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation,
  &atspi_collection_interface__get_matches_from_method_annotations__org_qtproject__qt_dbus__qt_type_name__in1_annotation,
  &atspi_collection_interface__get_matches_from_method_annotations__org_qtproject__qt_dbus__qt_type_name__out0_annotation,
  NULL,
};

const GDBusMethodInfo atspi_collection_interface__get_matches_from_method =
{
  -1,  /* ref count */
  (gchar *) "GetMatchesFrom",
  (GDBusArgInfo **) atspi_collection_interface__get_matches_from_method_in_args,
  (GDBusArgInfo **) atspi_collection_interface__get_matches_from_method_out_args,
  (GDBusAnnotationInfo **) atspi_collection_interface__get_matches_from_method_annotations,
};

const GDBusArgInfo atspi_collection_interface__get_active_descendant_method_out_args__unnamed_arg0_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg0",
  (gchar *) "(so)",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_collection_interface__get_active_descendant_method_out_args[] =
{
  &atspi_collection_interface__get_active_descendant_method_out_args__unnamed_arg0_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_collection_interface__get_active_descendant_method_annotations__org_qtproject__qt_dbus__qt_type_name__out0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.Out0",
  (gchar *) "QSpiReferenceSet",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_collection_interface__get_active_descendant_method_annotations[] =
{
  &atspi_collection_interface__get_active_descendant_method_annotations__org_qtproject__qt_dbus__qt_type_name__out0_annotation,
  NULL,
};

const GDBusMethodInfo atspi_collection_interface__get_active_descendant_method =
{
  -1,  /* ref count */
  (gchar *) "GetActiveDescendant",
  NULL,  /* no in args */
  (GDBusArgInfo **) atspi_collection_interface__get_active_descendant_method_out_args,
  (GDBusAnnotationInfo **) atspi_collection_interface__get_active_descendant_method_annotations,
};

const GDBusMethodInfo * const atspi_collection_interface_methods[] =
{
  &atspi_collection_interface__get_matches_method,
  &atspi_collection_interface__get_matches_to_method,
  &atspi_collection_interface__get_matches_from_method,
  &atspi_collection_interface__get_active_descendant_method,
  NULL,
};

const GDBusInterfaceInfo atspi_collection_interface =
{
  -1,  /* ref count */
  (gchar *) "org.a11y.atspi.Collection",
  (GDBusMethodInfo **) atspi_collection_interface_methods,
  NULL,  /* no signals */
NULL,  /* no properties */
  NULL,  /* no annotations */
};

