/* GDK - The GIMP Drawing Kit
 *
 * gdkglcontext-wayland.h: Private Wayland specific OpenGL wrappers
 *
 * Copyright © 2014  Emmanuele Bassi
 * Copyright © 2014  Red Hat, Int
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library. If not, see <http://www.gnu.org/licenses/>.
 */

#pragma once

#include "gdkglcontextprivate.h"
#include "gdkdisplayprivate.h"
#include "gdksurface.h"

#include <epoxy/egl.h>

G_BEGIN_DECLS

struct _GdkWaylandGLContext
{
  GdkGLContext parent_instance;
};

struct _GdkWaylandGLContextClass
{
  GdkGLContextClass parent_class;
};

GdkGLContext *  gdk_wayland_display_init_gl                         (GdkDisplay        *display,
                                                                     GError           **error);

G_END_DECLS

