G_GNUC_BEGIN_IGNORE_DEPRECATIONS
*tp++ = gdk_anchor_hints_get_type();
*tp++ = gdk_app_launch_context_get_type();
*tp++ = gdk_axis_flags_get_type();
*tp++ = gdk_axis_use_get_type();
*tp++ = gdk_button_event_get_type();
*tp++ = gdk_cairo_context_get_type();
*tp++ = gdk_clipboard_get_type();
*tp++ = gdk_content_deserializer_get_type();
*tp++ = gdk_content_formats_builder_get_type();
*tp++ = gdk_content_formats_get_type();
*tp++ = gdk_content_provider_get_type();
*tp++ = gdk_content_serializer_get_type();
*tp++ = gdk_crossing_event_get_type();
*tp++ = gdk_crossing_mode_get_type();
*tp++ = gdk_cursor_get_type();
*tp++ = gdk_delete_event_get_type();
*tp++ = gdk_device_get_type();
*tp++ = gdk_device_pad_feature_get_type();
*tp++ = gdk_device_pad_get_type();
*tp++ = gdk_device_tool_get_type();
*tp++ = gdk_device_tool_type_get_type();
*tp++ = gdk_display_get_type();
*tp++ = gdk_display_manager_get_type();
*tp++ = gdk_dmabuf_error_get_type();
*tp++ = gdk_dmabuf_formats_get_type();
*tp++ = gdk_dmabuf_texture_builder_get_type();
*tp++ = gdk_dmabuf_texture_get_type();
*tp++ = gdk_dnd_event_get_type();
*tp++ = gdk_drag_action_get_type();
*tp++ = gdk_drag_cancel_reason_get_type();
*tp++ = gdk_drag_get_type();
*tp++ = gdk_drag_surface_get_type();
*tp++ = gdk_drag_surface_size_get_type();
*tp++ = gdk_draw_context_get_type();
*tp++ = gdk_drop_get_type();
*tp++ = gdk_event_get_type();
*tp++ = gdk_event_sequence_get_type();
*tp++ = gdk_event_type_get_type();
*tp++ = gdk_file_list_get_type();
*tp++ = gdk_focus_event_get_type();
*tp++ = gdk_frame_clock_get_type();
*tp++ = gdk_frame_clock_phase_get_type();
*tp++ = gdk_frame_timings_get_type();
*tp++ = gdk_fullscreen_mode_get_type();
*tp++ = gdk_gl_api_get_type();
*tp++ = gdk_gl_context_get_type();
*tp++ = gdk_gl_error_get_type();
*tp++ = gdk_gl_texture_builder_get_type();
*tp++ = gdk_gl_texture_get_type();
*tp++ = gdk_grab_broken_event_get_type();
*tp++ = gdk_gravity_get_type();
*tp++ = gdk_input_source_get_type();
*tp++ = gdk_key_event_get_type();
*tp++ = gdk_key_match_get_type();
*tp++ = gdk_memory_format_get_type();
*tp++ = gdk_memory_texture_get_type();
*tp++ = gdk_modifier_type_get_type();
*tp++ = gdk_monitor_get_type();
*tp++ = gdk_motion_event_get_type();
*tp++ = gdk_notify_type_get_type();
*tp++ = gdk_pad_event_get_type();
*tp++ = gdk_paintable_flags_get_type();
*tp++ = gdk_paintable_get_type();
*tp++ = gdk_popup_get_type();
*tp++ = gdk_popup_layout_get_type();
*tp++ = gdk_proximity_event_get_type();
*tp++ = gdk_rectangle_get_type();
*tp++ = gdk_rgba_get_type();
*tp++ = gdk_scroll_direction_get_type();
*tp++ = gdk_scroll_event_get_type();
*tp++ = gdk_scroll_unit_get_type();
*tp++ = gdk_seat_capabilities_get_type();
*tp++ = gdk_seat_get_type();
*tp++ = gdk_snapshot_get_type();
*tp++ = gdk_subpixel_layout_get_type();
*tp++ = gdk_surface_edge_get_type();
*tp++ = gdk_surface_get_type();
*tp++ = gdk_texture_downloader_get_type();
*tp++ = gdk_texture_error_get_type();
*tp++ = gdk_texture_get_type();
*tp++ = gdk_titlebar_gesture_get_type();
*tp++ = gdk_toplevel_get_type();
*tp++ = gdk_toplevel_layout_get_type();
*tp++ = gdk_toplevel_size_get_type();
*tp++ = gdk_toplevel_state_get_type();
*tp++ = gdk_touch_event_get_type();
*tp++ = gdk_touchpad_event_get_type();
*tp++ = gdk_touchpad_gesture_phase_get_type();
*tp++ = gdk_vulkan_context_get_type();
*tp++ = gdk_vulkan_error_get_type();
*tp++ = gsk_blend_mode_get_type();
*tp++ = gsk_blend_node_get_type();
*tp++ = gsk_blur_node_get_type();
*tp++ = gsk_border_node_get_type();
*tp++ = gsk_cairo_node_get_type();
*tp++ = gsk_cairo_renderer_get_type();
*tp++ = gsk_clip_node_get_type();
*tp++ = gsk_color_matrix_node_get_type();
*tp++ = gsk_color_node_get_type();
*tp++ = gsk_conic_gradient_node_get_type();
*tp++ = gsk_container_node_get_type();
*tp++ = gsk_corner_get_type();
*tp++ = gsk_cross_fade_node_get_type();
*tp++ = gsk_debug_node_get_type();
*tp++ = gsk_fill_node_get_type();
*tp++ = gsk_fill_rule_get_type();
*tp++ = gsk_gl_renderer_get_type();
*tp++ = gsk_gl_shader_get_type();
*tp++ = gsk_gl_shader_node_get_type();
*tp++ = gsk_gl_uniform_type_get_type();
*tp++ = gsk_inset_shadow_node_get_type();
*tp++ = gsk_line_cap_get_type();
*tp++ = gsk_line_join_get_type();
*tp++ = gsk_linear_gradient_node_get_type();
*tp++ = gsk_mask_mode_get_type();
*tp++ = gsk_mask_node_get_type();

*tp++ = gsk_opacity_node_get_type();
*tp++ = gsk_outset_shadow_node_get_type();
*tp++ = gsk_path_builder_get_type();
*tp++ = gsk_path_direction_get_type();
*tp++ = gsk_path_foreach_flags_get_type();
*tp++ = gsk_path_get_type();
*tp++ = gsk_path_measure_get_type();
*tp++ = gsk_path_operation_get_type();
*tp++ = gsk_path_point_get_type();
*tp++ = gsk_radial_gradient_node_get_type();
*tp++ = gsk_render_node_get_type();
*tp++ = gsk_render_node_type_get_type();
*tp++ = gsk_renderer_get_type();
*tp++ = gsk_repeat_node_get_type();
*tp++ = gsk_repeating_linear_gradient_node_get_type();
*tp++ = gsk_repeating_radial_gradient_node_get_type();
*tp++ = gsk_rounded_clip_node_get_type();
*tp++ = gsk_scaling_filter_get_type();
*tp++ = gsk_serialization_error_get_type();
*tp++ = gsk_shader_args_builder_get_type();
*tp++ = gsk_shadow_node_get_type();
*tp++ = gsk_stroke_get_type();
*tp++ = gsk_stroke_node_get_type();
*tp++ = gsk_subsurface_node_get_type();
*tp++ = gsk_text_node_get_type();
*tp++ = gsk_texture_node_get_type();
*tp++ = gsk_texture_scale_node_get_type();
*tp++ = gsk_transform_category_get_type();
*tp++ = gsk_transform_get_type();
*tp++ = gsk_transform_node_get_type();
#ifdef GDK_RENDERING_VULKAN
*tp++ = gsk_vulkan_renderer_get_type();
#endif
*tp++ = gtk_about_dialog_get_type();
*tp++ = gtk_accessible_announcement_priority_get_type();
*tp++ = gtk_accessible_autocomplete_get_type();
*tp++ = gtk_accessible_get_type();
*tp++ = gtk_accessible_invalid_state_get_type();
*tp++ = gtk_accessible_list_get_type();
*tp++ = gtk_accessible_platform_state_get_type();
*tp++ = gtk_accessible_property_get_type();
*tp++ = gtk_accessible_range_get_type();
*tp++ = gtk_accessible_relation_get_type();
*tp++ = gtk_accessible_role_get_type();
*tp++ = gtk_accessible_sort_get_type();
*tp++ = gtk_accessible_state_get_type();
*tp++ = gtk_accessible_text_content_change_get_type();
*tp++ = gtk_accessible_text_get_type();
*tp++ = gtk_accessible_text_granularity_get_type();
*tp++ = gtk_accessible_tristate_get_type();
*tp++ = gtk_action_bar_get_type();
*tp++ = gtk_actionable_get_type();
*tp++ = gtk_activate_action_get_type();
*tp++ = gtk_adjustment_get_type();
*tp++ = gtk_alert_dialog_get_type();
*tp++ = gtk_align_get_type();
*tp++ = gtk_alternative_trigger_get_type();
*tp++ = gtk_any_filter_get_type();
*tp++ = gtk_app_chooser_button_get_type();
*tp++ = gtk_app_chooser_dialog_get_type();
*tp++ = gtk_app_chooser_get_type();
*tp++ = gtk_app_chooser_widget_get_type();
*tp++ = gtk_application_get_type();
*tp++ = gtk_application_inhibit_flags_get_type();
*tp++ = gtk_application_window_get_type();
*tp++ = gtk_arrow_type_get_type();
*tp++ = gtk_aspect_frame_get_type();
*tp++ = gtk_assistant_get_type();
*tp++ = gtk_assistant_page_get_type();
*tp++ = gtk_assistant_page_type_get_type();
*tp++ = gtk_at_context_get_type();
*tp++ = gtk_baseline_position_get_type();
*tp++ = gtk_bin_layout_get_type();
*tp++ = gtk_bitset_get_type();
*tp++ = gtk_bitset_iter_get_type();
*tp++ = gtk_bookmark_list_get_type();
*tp++ = gtk_bool_filter_get_type();
*tp++ = gtk_border_get_type();
*tp++ = gtk_border_style_get_type();
*tp++ = gtk_box_get_type();
*tp++ = gtk_box_layout_get_type();
*tp++ = gtk_buildable_get_type();
*tp++ = gtk_builder_closure_flags_get_type();
*tp++ = gtk_builder_cscope_get_type();
*tp++ = gtk_builder_error_get_type();
*tp++ = gtk_builder_get_type();
*tp++ = gtk_builder_list_item_factory_get_type();
*tp++ = gtk_builder_scope_get_type();
*tp++ = gtk_button_get_type();
*tp++ = gtk_buttons_type_get_type();
*tp++ = gtk_calendar_get_type();
*tp++ = gtk_callback_action_get_type();
*tp++ = gtk_cclosure_expression_get_type();
*tp++ = gtk_cell_area_box_get_type();
*tp++ = gtk_cell_area_context_get_type();
*tp++ = gtk_cell_area_get_type();
*tp++ = gtk_cell_editable_get_type();
*tp++ = gtk_cell_layout_get_type();
*tp++ = gtk_cell_renderer_accel_get_type();
*tp++ = gtk_cell_renderer_accel_mode_get_type();
*tp++ = gtk_cell_renderer_combo_get_type();
*tp++ = gtk_cell_renderer_get_type();
*tp++ = gtk_cell_renderer_mode_get_type();
*tp++ = gtk_cell_renderer_pixbuf_get_type();
*tp++ = gtk_cell_renderer_progress_get_type();
*tp++ = gtk_cell_renderer_spin_get_type();
*tp++ = gtk_cell_renderer_spinner_get_type();
*tp++ = gtk_cell_renderer_state_get_type();
*tp++ = gtk_cell_renderer_text_get_type();
*tp++ = gtk_cell_renderer_toggle_get_type();
*tp++ = gtk_cell_view_get_type();
*tp++ = gtk_center_box_get_type();
*tp++ = gtk_center_layout_get_type();
*tp++ = gtk_check_button_get_type();
*tp++ = gtk_closure_expression_get_type();
*tp++ = gtk_collation_get_type();
*tp++ = gtk_color_button_get_type();
*tp++ = gtk_color_chooser_dialog_get_type();
*tp++ = gtk_color_chooser_get_type();
*tp++ = gtk_color_chooser_widget_get_type();
*tp++ = gtk_color_dialog_button_get_type();
*tp++ = gtk_color_dialog_get_type();
*tp++ = gtk_column_view_cell_get_type();
*tp++ = gtk_column_view_column_get_type();
*tp++ = gtk_column_view_get_type();
*tp++ = gtk_column_view_row_get_type();
*tp++ = gtk_column_view_sorter_get_type();
*tp++ = gtk_combo_box_get_type();
*tp++ = gtk_combo_box_text_get_type();
*tp++ = gtk_constant_expression_get_type();
*tp++ = gtk_constraint_attribute_get_type();
*tp++ = gtk_constraint_get_type();
*tp++ = gtk_constraint_guide_get_type();
*tp++ = gtk_constraint_layout_child_get_type();
*tp++ = gtk_constraint_layout_get_type();
*tp++ = gtk_constraint_relation_get_type();
*tp++ = gtk_constraint_strength_get_type();
*tp++ = gtk_constraint_target_get_type();
*tp++ = gtk_constraint_vfl_parser_error_get_type();
*tp++ = gtk_content_fit_get_type();
*tp++ = gtk_corner_type_get_type();
*tp++ = gtk_css_parser_error_get_type();
*tp++ = gtk_css_parser_warning_get_type();
*tp++ = gtk_css_provider_get_type();
*tp++ = gtk_custom_filter_get_type();
*tp++ = gtk_custom_layout_get_type();
*tp++ = gtk_custom_sorter_get_type();
*tp++ = gtk_debug_flags_get_type();
*tp++ = gtk_delete_type_get_type();
*tp++ = gtk_dialog_error_get_type();
*tp++ = gtk_dialog_flags_get_type();
*tp++ = gtk_dialog_get_type();
*tp++ = gtk_direction_type_get_type();
*tp++ = gtk_directory_list_get_type();
*tp++ = gtk_drag_icon_get_type();
*tp++ = gtk_drag_source_get_type();
*tp++ = gtk_drawing_area_get_type();
*tp++ = gtk_drop_controller_motion_get_type();
*tp++ = gtk_drop_down_get_type();
*tp++ = gtk_drop_target_async_get_type();
*tp++ = gtk_drop_target_get_type();
*tp++ = gtk_editable_get_type();
*tp++ = gtk_editable_label_get_type();
*tp++ = gtk_editable_properties_get_type();
*tp++ = gtk_emoji_chooser_get_type();
*tp++ = gtk_entry_buffer_get_type();
*tp++ = gtk_entry_completion_get_type();
*tp++ = gtk_entry_get_type();
*tp++ = gtk_entry_icon_position_get_type();
*tp++ = gtk_event_controller_focus_get_type();
*tp++ = gtk_event_controller_get_type();
*tp++ = gtk_event_controller_key_get_type();
*tp++ = gtk_event_controller_legacy_get_type();
*tp++ = gtk_event_controller_motion_get_type();
*tp++ = gtk_event_controller_scroll_flags_get_type();
*tp++ = gtk_event_controller_scroll_get_type();
*tp++ = gtk_event_sequence_state_get_type();
*tp++ = gtk_every_filter_get_type();
*tp++ = gtk_expander_get_type();
*tp++ = gtk_expression_get_type();
*tp++ = gtk_expression_watch_get_type();
*tp++ = gtk_file_chooser_action_get_type();
*tp++ = gtk_file_chooser_dialog_get_type();
*tp++ = gtk_file_chooser_error_get_type();
*tp++ = gtk_file_chooser_get_type();
*tp++ = gtk_file_chooser_native_get_type();
*tp++ = gtk_file_chooser_widget_get_type();
*tp++ = gtk_file_dialog_get_type();
*tp++ = gtk_file_filter_get_type();
*tp++ = gtk_file_launcher_get_type();
*tp++ = gtk_filter_change_get_type();
*tp++ = gtk_filter_get_type();
*tp++ = gtk_filter_list_model_get_type();
*tp++ = gtk_filter_match_get_type();
*tp++ = gtk_fixed_get_type();
*tp++ = gtk_fixed_layout_child_get_type();
*tp++ = gtk_fixed_layout_get_type();
*tp++ = gtk_flatten_list_model_get_type();
*tp++ = gtk_flow_box_child_get_type();
*tp++ = gtk_flow_box_get_type();
*tp++ = gtk_font_button_get_type();
*tp++ = gtk_font_chooser_dialog_get_type();
*tp++ = gtk_font_chooser_get_type();
*tp++ = gtk_font_chooser_level_get_type();
*tp++ = gtk_font_chooser_widget_get_type();
*tp++ = gtk_font_dialog_button_get_type();
*tp++ = gtk_font_dialog_get_type();
*tp++ = gtk_font_level_get_type();
*tp++ = gtk_frame_get_type();
*tp++ = gtk_gesture_click_get_type();
*tp++ = gtk_gesture_drag_get_type();
*tp++ = gtk_gesture_get_type();
*tp++ = gtk_gesture_long_press_get_type();
*tp++ = gtk_gesture_pan_get_type();
*tp++ = gtk_gesture_rotate_get_type();
*tp++ = gtk_gesture_single_get_type();
*tp++ = gtk_gesture_stylus_get_type();
*tp++ = gtk_gesture_swipe_get_type();
*tp++ = gtk_gesture_zoom_get_type();
*tp++ = gtk_gl_area_get_type();
*tp++ = gtk_graphics_offload_enabled_get_type();
*tp++ = gtk_graphics_offload_get_type();
*tp++ = gtk_grid_get_type();
*tp++ = gtk_grid_layout_child_get_type();
*tp++ = gtk_grid_layout_get_type();
*tp++ = gtk_grid_view_get_type();
*tp++ = gtk_header_bar_get_type();
*tp++ = gtk_icon_lookup_flags_get_type();
*tp++ = gtk_icon_paintable_get_type();
*tp++ = gtk_icon_size_get_type();
*tp++ = gtk_icon_theme_error_get_type();
*tp++ = gtk_icon_theme_get_type();
*tp++ = gtk_icon_view_drop_position_get_type();
*tp++ = gtk_icon_view_get_type();
*tp++ = gtk_im_context_get_type();
*tp++ = gtk_im_context_simple_get_type();
*tp++ = gtk_im_multicontext_get_type();
*tp++ = gtk_image_get_type();
*tp++ = gtk_image_type_get_type();
*tp++ = gtk_info_bar_get_type();
*tp++ = gtk_input_hints_get_type();
*tp++ = gtk_input_purpose_get_type();
*tp++ = gtk_inscription_get_type();
*tp++ = gtk_inscription_overflow_get_type();
*tp++ = gtk_justification_get_type();
*tp++ = gtk_keyval_trigger_get_type();
*tp++ = gtk_label_get_type();
*tp++ = gtk_layout_child_get_type();
*tp++ = gtk_layout_manager_get_type();
*tp++ = gtk_level_bar_get_type();
*tp++ = gtk_level_bar_mode_get_type();
*tp++ = gtk_license_get_type();
*tp++ = gtk_link_button_get_type();
*tp++ = gtk_list_base_get_type();
*tp++ = gtk_list_box_get_type();
*tp++ = gtk_list_box_row_get_type();
*tp++ = gtk_list_header_get_type();
*tp++ = gtk_list_item_factory_get_type();
*tp++ = gtk_list_item_get_type();
*tp++ = gtk_list_scroll_flags_get_type();
*tp++ = gtk_list_store_get_type();
*tp++ = gtk_list_tab_behavior_get_type();
*tp++ = gtk_list_view_get_type();
*tp++ = gtk_lock_button_get_type();
*tp++ = gtk_map_list_model_get_type();
*tp++ = gtk_media_controls_get_type();
*tp++ = gtk_media_file_get_type();
*tp++ = gtk_media_stream_get_type();
*tp++ = gtk_menu_button_get_type();
*tp++ = gtk_message_dialog_get_type();
*tp++ = gtk_message_type_get_type();
*tp++ = gtk_mnemonic_action_get_type();
*tp++ = gtk_mnemonic_trigger_get_type();
*tp++ = gtk_mount_operation_get_type();
*tp++ = gtk_movement_step_get_type();
*tp++ = gtk_multi_filter_get_type();
*tp++ = gtk_multi_selection_get_type();
*tp++ = gtk_multi_sorter_get_type();
*tp++ = gtk_named_action_get_type();
*tp++ = gtk_native_dialog_get_type();
*tp++ = gtk_native_get_type();
*tp++ = gtk_natural_wrap_mode_get_type();
*tp++ = gtk_never_trigger_get_type();
*tp++ = gtk_no_selection_get_type();
*tp++ = gtk_notebook_get_type();
*tp++ = gtk_notebook_page_get_type();
*tp++ = gtk_notebook_tab_get_type();
*tp++ = gtk_nothing_action_get_type();
*tp++ = gtk_number_up_layout_get_type();
*tp++ = gtk_numeric_sorter_get_type();
*tp++ = gtk_object_expression_get_type();
*tp++ = gtk_ordering_get_type();
*tp++ = gtk_orientable_get_type();
*tp++ = gtk_orientation_get_type();
*tp++ = gtk_overflow_get_type();
*tp++ = gtk_overlay_get_type();
*tp++ = gtk_overlay_layout_child_get_type();
*tp++ = gtk_overlay_layout_get_type();
*tp++ = gtk_pack_type_get_type();
*tp++ = gtk_pad_action_type_get_type();
*tp++ = gtk_pad_controller_get_type();
*tp++ = gtk_page_orientation_get_type();
*tp++ = gtk_page_set_get_type();
*tp++ = gtk_pan_direction_get_type();
*tp++ = gtk_paned_get_type();
*tp++ = gtk_param_expression_get_type();
*tp++ = gtk_password_entry_buffer_get_type();
*tp++ = gtk_password_entry_get_type();
*tp++ = gtk_pick_flags_get_type();
*tp++ = gtk_picture_get_type();
*tp++ = gtk_policy_type_get_type();
*tp++ = gtk_popover_get_type();
*tp++ = gtk_popover_menu_bar_get_type();
*tp++ = gtk_popover_menu_flags_get_type();
*tp++ = gtk_popover_menu_get_type();
*tp++ = gtk_position_type_get_type();
*tp++ = gtk_print_dialog_get_type();
*tp++ = gtk_print_duplex_get_type();
*tp++ = gtk_print_error_get_type();
*tp++ = gtk_print_operation_action_get_type();
*tp++ = gtk_print_operation_result_get_type();
*tp++ = gtk_print_pages_get_type();
*tp++ = gtk_print_quality_get_type();
*tp++ = gtk_print_setup_get_type();
*tp++ = gtk_print_status_get_type();
*tp++ = gtk_progress_bar_get_type();
*tp++ = gtk_propagation_limit_get_type();
*tp++ = gtk_propagation_phase_get_type();
*tp++ = gtk_property_expression_get_type();
*tp++ = gtk_range_get_type();
*tp++ = gtk_recent_info_get_type();
*tp++ = gtk_recent_manager_error_get_type();
*tp++ = gtk_recent_manager_get_type();
*tp++ = gtk_requisition_get_type();
*tp++ = gtk_response_type_get_type();
*tp++ = gtk_revealer_get_type();
*tp++ = gtk_revealer_transition_type_get_type();
*tp++ = gtk_root_get_type();
*tp++ = gtk_scale_button_get_type();
*tp++ = gtk_scale_get_type();
*tp++ = gtk_scroll_info_get_type();
*tp++ = gtk_scroll_step_get_type();
*tp++ = gtk_scroll_type_get_type();
*tp++ = gtk_scrollable_get_type();
*tp++ = gtk_scrollable_policy_get_type();
*tp++ = gtk_scrollbar_get_type();
*tp++ = gtk_scrolled_window_get_type();
*tp++ = gtk_search_bar_get_type();
*tp++ = gtk_search_entry_get_type();
*tp++ = gtk_section_model_get_type();
*tp++ = gtk_selection_filter_model_get_type();
*tp++ = gtk_selection_mode_get_type();
*tp++ = gtk_selection_model_get_type();
*tp++ = gtk_sensitivity_type_get_type();
*tp++ = gtk_separator_get_type();
*tp++ = gtk_settings_get_type();
*tp++ = gtk_shortcut_action_flags_get_type();
*tp++ = gtk_shortcut_action_get_type();
*tp++ = gtk_shortcut_controller_get_type();
*tp++ = gtk_shortcut_get_type();
*tp++ = gtk_shortcut_label_get_type();
*tp++ = gtk_shortcut_manager_get_type();
*tp++ = gtk_shortcut_scope_get_type();
*tp++ = gtk_shortcut_trigger_get_type();
*tp++ = gtk_shortcut_type_get_type();
*tp++ = gtk_shortcuts_group_get_type();
*tp++ = gtk_shortcuts_section_get_type();
*tp++ = gtk_shortcuts_shortcut_get_type();
*tp++ = gtk_shortcuts_window_get_type();
*tp++ = gtk_signal_action_get_type();
*tp++ = gtk_signal_list_item_factory_get_type();
*tp++ = gtk_single_selection_get_type();
*tp++ = gtk_size_group_get_type();
*tp++ = gtk_size_group_mode_get_type();
*tp++ = gtk_size_request_mode_get_type();
*tp++ = gtk_slice_list_model_get_type();
*tp++ = gtk_snapshot_get_type();
*tp++ = gtk_sort_list_model_get_type();
*tp++ = gtk_sort_type_get_type();
*tp++ = gtk_sorter_change_get_type();
*tp++ = gtk_sorter_get_type();
*tp++ = gtk_sorter_order_get_type();
*tp++ = gtk_spin_button_get_type();
*tp++ = gtk_spin_button_update_policy_get_type();
*tp++ = gtk_spin_type_get_type();
*tp++ = gtk_spinner_get_type();
*tp++ = gtk_stack_get_type();
*tp++ = gtk_stack_page_get_type();
*tp++ = gtk_stack_sidebar_get_type();
*tp++ = gtk_stack_switcher_get_type();
*tp++ = gtk_stack_transition_type_get_type();
*tp++ = gtk_state_flags_get_type();
*tp++ = gtk_statusbar_get_type();
*tp++ = gtk_string_filter_get_type();
*tp++ = gtk_string_filter_match_mode_get_type();
*tp++ = gtk_string_list_get_type();
*tp++ = gtk_string_object_get_type();
*tp++ = gtk_string_sorter_get_type();
*tp++ = gtk_style_context_get_type();
*tp++ = gtk_style_context_print_flags_get_type();
*tp++ = gtk_style_provider_get_type();
*tp++ = gtk_switch_get_type();
*tp++ = gtk_symbolic_color_get_type();
*tp++ = gtk_symbolic_paintable_get_type();
*tp++ = gtk_system_setting_get_type();
*tp++ = gtk_text_buffer_get_type();
*tp++ = gtk_text_child_anchor_get_type();
*tp++ = gtk_text_direction_get_type();
*tp++ = gtk_text_extend_selection_get_type();
*tp++ = gtk_text_get_type();
*tp++ = gtk_text_iter_get_type();
*tp++ = gtk_text_mark_get_type();
*tp++ = gtk_text_search_flags_get_type();
*tp++ = gtk_text_tag_get_type();
*tp++ = gtk_text_tag_table_get_type();
*tp++ = gtk_text_view_get_type();
*tp++ = gtk_text_view_layer_get_type();
*tp++ = gtk_text_window_type_get_type();
*tp++ = gtk_toggle_button_get_type();
*tp++ = gtk_tooltip_get_type();
*tp++ = gtk_tree_drag_dest_get_type();
*tp++ = gtk_tree_drag_source_get_type();
*tp++ = gtk_tree_expander_get_type();
*tp++ = gtk_tree_iter_get_type();
*tp++ = gtk_tree_list_model_get_type();
*tp++ = gtk_tree_list_row_get_type();
*tp++ = gtk_tree_list_row_sorter_get_type();
*tp++ = gtk_tree_model_filter_get_type();
*tp++ = gtk_tree_model_flags_get_type();
*tp++ = gtk_tree_model_get_type();
*tp++ = gtk_tree_model_sort_get_type();
*tp++ = gtk_tree_path_get_type();
*tp++ = gtk_tree_row_data_get_type();
*tp++ = gtk_tree_row_reference_get_type();
*tp++ = gtk_tree_selection_get_type();
*tp++ = gtk_tree_sortable_get_type();
*tp++ = gtk_tree_store_get_type();
*tp++ = gtk_tree_view_column_get_type();
*tp++ = gtk_tree_view_column_sizing_get_type();
*tp++ = gtk_tree_view_drop_position_get_type();
*tp++ = gtk_tree_view_get_type();
*tp++ = gtk_tree_view_grid_lines_get_type();
*tp++ = gtk_unit_get_type();
*tp++ = gtk_uri_launcher_get_type();
*tp++ = gtk_video_get_type();
*tp++ = gtk_viewport_get_type();
*tp++ = gtk_volume_button_get_type();
*tp++ = gtk_widget_get_type();
*tp++ = gtk_widget_paintable_get_type();
*tp++ = gtk_window_controls_get_type();
*tp++ = gtk_window_get_type();
*tp++ = gtk_window_group_get_type();
*tp++ = gtk_window_handle_get_type();
*tp++ = gtk_wrap_mode_get_type();
G_GNUC_END_IGNORE_DEPRECATIONS