/*
 * This file is generated by gdbus-codegen, do not modify it.
 *
 * The license of this code is the same as for the D-Bus interface description
 * it was derived from. Note that it links to GLib, so must comply with the
 * LGPL linking clauses.
 */

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include "atspi-application.h"

#include <string.h>

/* ------------------------------------------------------------------------ */
/* Definitions for org.a11y.atspi.Application */

const GDBusArgInfo atspi_application_interface__get_locale_method_in_args__lctype_arg =
{
  -1,  /* ref count */
  (gchar *) "lctype",
  (gchar *) "u",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_application_interface__get_locale_method_in_args[] =
{
  &atspi_application_interface__get_locale_method_in_args__lctype_arg,
  NULL,
};

const GDBusArgInfo atspi_application_interface__get_locale_method_out_args__unnamed_arg1_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg1",
  (gchar *) "s",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_application_interface__get_locale_method_out_args[] =
{
  &atspi_application_interface__get_locale_method_out_args__unnamed_arg1_arg,
  NULL,
};

const GDBusMethodInfo atspi_application_interface__get_locale_method =
{
  -1,  /* ref count */
  (gchar *) "GetLocale",
  (GDBusArgInfo **) atspi_application_interface__get_locale_method_in_args,
  (GDBusArgInfo **) atspi_application_interface__get_locale_method_out_args,
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_application_interface__register_event_listener_method_in_args__event_arg =
{
  -1,  /* ref count */
  (gchar *) "event",
  (gchar *) "s",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_application_interface__register_event_listener_method_in_args[] =
{
  &atspi_application_interface__register_event_listener_method_in_args__event_arg,
  NULL,
};

const GDBusMethodInfo atspi_application_interface__register_event_listener_method =
{
  -1,  /* ref count */
  (gchar *) "RegisterEventListener",
  (GDBusArgInfo **) atspi_application_interface__register_event_listener_method_in_args,
  NULL,  /* no out args */
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_application_interface__deregister_event_listener_method_in_args__event_arg =
{
  -1,  /* ref count */
  (gchar *) "event",
  (gchar *) "s",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_application_interface__deregister_event_listener_method_in_args[] =
{
  &atspi_application_interface__deregister_event_listener_method_in_args__event_arg,
  NULL,
};

const GDBusMethodInfo atspi_application_interface__deregister_event_listener_method =
{
  -1,  /* ref count */
  (gchar *) "DeregisterEventListener",
  (GDBusArgInfo **) atspi_application_interface__deregister_event_listener_method_in_args,
  NULL,  /* no out args */
  NULL,  /* no annotations */
};

const GDBusMethodInfo * const atspi_application_interface_methods[] =
{
  &atspi_application_interface__get_locale_method,
  &atspi_application_interface__register_event_listener_method,
  &atspi_application_interface__deregister_event_listener_method,
  NULL,
};

const GDBusPropertyInfo atspi_application_interface__toolkit_name_property =
{
  -1,  /* ref count */
  (gchar *) "ToolkitName",
  (gchar *) "s",
  G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
  NULL,  /* no annotations */
};

const GDBusPropertyInfo atspi_application_interface__version_property =
{
  -1,  /* ref count */
  (gchar *) "Version",
  (gchar *) "s",
  G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
  NULL,  /* no annotations */
};

const GDBusPropertyInfo atspi_application_interface__atspi_version_property =
{
  -1,  /* ref count */
  (gchar *) "AtspiVersion",
  (gchar *) "s",
  G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
  NULL,  /* no annotations */
};

const GDBusPropertyInfo atspi_application_interface__id_property =
{
  -1,  /* ref count */
  (gchar *) "Id",
  (gchar *) "i",
  G_DBUS_PROPERTY_INFO_FLAGS_READABLE | G_DBUS_PROPERTY_INFO_FLAGS_WRITABLE,
  NULL,  /* no annotations */
};

const GDBusPropertyInfo * const atspi_application_interface_properties[] =
{
  &atspi_application_interface__toolkit_name_property,
  &atspi_application_interface__version_property,
  &atspi_application_interface__atspi_version_property,
  &atspi_application_interface__id_property,
  NULL,
};

const GDBusInterfaceInfo atspi_application_interface =
{
  -1,  /* ref count */
  (gchar *) "org.a11y.atspi.Application",
  (GDBusMethodInfo **) atspi_application_interface_methods,
  NULL,  /* no signals */
  (GDBusPropertyInfo **) atspi_application_interface_properties,
  NULL,  /* no annotations */
};

