/*
 * This file is generated by gdbus-codegen, do not modify it.
 *
 * The license of this code is the same as for the D-Bus interface description
 * it was derived from. Note that it links to GLib, so must comply with the
 * LGPL linking clauses.
 */

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include "atspi-tablecell.h"

#include <string.h>

/* ------------------------------------------------------------------------ */
/* Definitions for org.a11y.atspi.TableCell */

const GDBusArgInfo atspi_table_cell_interface__get_row_column_span_method_out_args__unnamed_arg0_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg0",
  (gchar *) "b",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_table_cell_interface__get_row_column_span_method_out_args__row_arg =
{
  -1,  /* ref count */
  (gchar *) "row",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_table_cell_interface__get_row_column_span_method_out_args__col_arg =
{
  -1,  /* ref count */
  (gchar *) "col",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_table_cell_interface__get_row_column_span_method_out_args__row_extents_arg =
{
  -1,  /* ref count */
  (gchar *) "row_extents",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_table_cell_interface__get_row_column_span_method_out_args__col_extents_arg =
{
  -1,  /* ref count */
  (gchar *) "col_extents",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_table_cell_interface__get_row_column_span_method_out_args[] =
{
  &atspi_table_cell_interface__get_row_column_span_method_out_args__unnamed_arg0_arg,
  &atspi_table_cell_interface__get_row_column_span_method_out_args__row_arg,
  &atspi_table_cell_interface__get_row_column_span_method_out_args__col_arg,
  &atspi_table_cell_interface__get_row_column_span_method_out_args__row_extents_arg,
  &atspi_table_cell_interface__get_row_column_span_method_out_args__col_extents_arg,
  NULL,
};

const GDBusMethodInfo atspi_table_cell_interface__get_row_column_span_method =
{
  -1,  /* ref count */
  (gchar *) "GetRowColumnSpan",
  NULL,  /* no in args */
  (GDBusArgInfo **) atspi_table_cell_interface__get_row_column_span_method_out_args,
  NULL,  /* no annotations */
};

const GDBusMethodInfo * const atspi_table_cell_interface_methods[] =
{
  &atspi_table_cell_interface__get_row_column_span_method,
  NULL,
};

const GDBusPropertyInfo atspi_table_cell_interface__column_span_property =
{
  -1,  /* ref count */
  (gchar *) "ColumnSpan",
  (gchar *) "i",
  G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo atspi_table_cell_interface__position_property_annotations__org_qtproject__qt_dbus__qt_type_name_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName",
  (gchar *) "QPoint",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_table_cell_interface__position_property_annotations[] =
{
  &atspi_table_cell_interface__position_property_annotations__org_qtproject__qt_dbus__qt_type_name_annotation,
  NULL,
};

const GDBusPropertyInfo atspi_table_cell_interface__position_property =
{
  -1,  /* ref count */
  (gchar *) "Position",
  (gchar *) "(ii)",
  G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
  (GDBusAnnotationInfo **) atspi_table_cell_interface__position_property_annotations,
};

const GDBusPropertyInfo atspi_table_cell_interface__row_span_property =
{
  -1,  /* ref count */
  (gchar *) "RowSpan",
  (gchar *) "i",
  G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo atspi_table_cell_interface__table_property_annotations__org_qtproject__qt_dbus__qt_type_name_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName",
  (gchar *) "QSpiObjectReference",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_table_cell_interface__table_property_annotations[] =
{
  &atspi_table_cell_interface__table_property_annotations__org_qtproject__qt_dbus__qt_type_name_annotation,
  NULL,
};

const GDBusPropertyInfo atspi_table_cell_interface__table_property =
{
  -1,  /* ref count */
  (gchar *) "Table",
  (gchar *) "(so)",
  G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
  (GDBusAnnotationInfo **) atspi_table_cell_interface__table_property_annotations,
};

const GDBusPropertyInfo * const atspi_table_cell_interface_properties[] =
{
  &atspi_table_cell_interface__column_span_property,
  &atspi_table_cell_interface__position_property,
  &atspi_table_cell_interface__row_span_property,
  &atspi_table_cell_interface__table_property,
  NULL,
};

const GDBusInterfaceInfo atspi_table_cell_interface =
{
  -1,  /* ref count */
  (gchar *) "org.a11y.atspi.TableCell",
  (GDBusMethodInfo **) atspi_table_cell_interface_methods,
  NULL,  /* no signals */
  (GDBusPropertyInfo **) atspi_table_cell_interface_properties,
  NULL,  /* no annotations */
};

