/*
 * This file is generated by gdbus-codegen, do not modify it.
 *
 * The license of this code is the same as for the D-Bus interface description
 * it was derived from. Note that it links to GLib, so must comply with the
 * LGPL linking clauses.
 */

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include "atspi-socket.h"

#include <string.h>

/* ------------------------------------------------------------------------ */
/* Definitions for org.a11y.atspi.Socket */

const GDBusArgInfo atspi_socket_interface__embed_method_in_args__plug_arg =
{
  -1,  /* ref count */
  (gchar *) "plug",
  (gchar *) "(so)",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_socket_interface__embed_method_in_args[] =
{
  &atspi_socket_interface__embed_method_in_args__plug_arg,
  NULL,
};

const GDBusArgInfo atspi_socket_interface__embed_method_out_args__socket_arg =
{
  -1,  /* ref count */
  (gchar *) "socket",
  (gchar *) "(so)",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_socket_interface__embed_method_out_args[] =
{
  &atspi_socket_interface__embed_method_out_args__socket_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_socket_interface__embed_method_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.In0",
  (gchar *) "QSpiObjectReference",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo atspi_socket_interface__embed_method_annotations__org_qtproject__qt_dbus__qt_type_name__out0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.Out0",
  (gchar *) "QSpiObjectReference",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_socket_interface__embed_method_annotations[] =
{
  &atspi_socket_interface__embed_method_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation,
  &atspi_socket_interface__embed_method_annotations__org_qtproject__qt_dbus__qt_type_name__out0_annotation,
  NULL,
};

const GDBusMethodInfo atspi_socket_interface__embed_method =
{
  -1,  /* ref count */
  (gchar *) "Embed",
  (GDBusArgInfo **) atspi_socket_interface__embed_method_in_args,
  (GDBusArgInfo **) atspi_socket_interface__embed_method_out_args,
  (GDBusAnnotationInfo **) atspi_socket_interface__embed_method_annotations,
};

const GDBusArgInfo atspi_socket_interface__unembed_method_in_args__plug_arg =
{
  -1,  /* ref count */
  (gchar *) "plug",
  (gchar *) "(so)",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_socket_interface__unembed_method_in_args[] =
{
  &atspi_socket_interface__unembed_method_in_args__plug_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_socket_interface__unembed_method_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.In0",
  (gchar *) "QSpiObjectReference",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_socket_interface__unembed_method_annotations[] =
{
  &atspi_socket_interface__unembed_method_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation,
  NULL,
};

const GDBusMethodInfo atspi_socket_interface__unembed_method =
{
  -1,  /* ref count */
  (gchar *) "Unembed",
  (GDBusArgInfo **) atspi_socket_interface__unembed_method_in_args,
  NULL,  /* no out args */
  (GDBusAnnotationInfo **) atspi_socket_interface__unembed_method_annotations,
};

const GDBusMethodInfo * const atspi_socket_interface_methods[] =
{
  &atspi_socket_interface__embed_method,
  &atspi_socket_interface__unembed_method,
  NULL,
};

const GDBusArgInfo atspi_socket_interface__available_signal_args__socket_arg =
{
  -1,  /* ref count */
  (gchar *) "socket",
  (gchar *) "(so)",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_socket_interface__available_signal_args[] =
{
  &atspi_socket_interface__available_signal_args__socket_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_socket_interface__available_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.In0",
  (gchar *) "QSpiObjectReference",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_socket_interface__available_signal_annotations[] =
{
  &atspi_socket_interface__available_signal_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation,
  NULL,
};

const GDBusSignalInfo atspi_socket_interface__available_signal =
{
  -1,  /* ref count */
  (gchar *) "Available",
  (GDBusArgInfo **) atspi_socket_interface__available_signal_args,
  (GDBusAnnotationInfo **) atspi_socket_interface__available_signal_annotations,
};

const GDBusSignalInfo * const atspi_socket_interface_signals[] =
{
  &atspi_socket_interface__available_signal,
  NULL,
};

const GDBusInterfaceInfo atspi_socket_interface =
{
  -1,  /* ref count */
  (gchar *) "org.a11y.atspi.Socket",
  (GDBusMethodInfo **) atspi_socket_interface_methods,
  (GDBusSignalInfo **) atspi_socket_interface_signals,
NULL,  /* no properties */
  NULL,  /* no annotations */
};

