
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#include "config.h"
#include "gdkenumtypes.h"
#include "gdk.h"
#include "gdkapplaunchcontext.h"
#include "gdkcairo.h"
#include "gdkcairocontext.h"
#include "gdkclipboard.h"
#include "gdkcontentdeserializer.h"
#include "gdkcontentformats.h"
#include "gdkcontentprovider.h"
#include "gdkcontentproviderimpl.h"
#include "gdkcontentserializer.h"
#include "gdkcursor.h"
#include "gdkdevice.h"
#include "gdkdevicepad.h"
#include "gdkdevicetool.h"
#include "gdkdisplay.h"
#include "gdkdisplaymanager.h"
#include "gdkdrag.h"
#include "gdkdmabufformats.h"
#include "gdkdmabuftexture.h"
#include "gdkdmabuftexturebuilder.h"
#include "gdkdragsurfacesize.h"
#include "gdkdrawcontext.h"
#include "gdkdrop.h"
#include "gdkenums.h"
#include "gdkevents.h"
#include "gdkframeclock.h"
#include "gdkframetimings.h"
#include "gdkglcontext.h"
#include "gdkgltexture.h"
#include "gdkgltexturebuilder.h"
#include "gdkkeys.h"
#include "gdkkeysyms.h"
#include "gdkmemorytexture.h"
#include "gdkmonitor.h"
#include "gdkpaintable.h"
#include "gdkpango.h"
#include "gdkrectangle.h"
#include "gdkrgba.h"
#include "gdkseat.h"
#include "gdksnapshot.h"
#include "gdktexture.h"
#include "gdktexturedownloader.h"
#include "gdktypes.h"
#include "gdkvulkancontext.h"
#include "gdksurface.h"
#include "gdkpopuplayout.h"
#include "gdkpopup.h"
#include "gdktoplevellayout.h"
#include "gdktoplevelsize.h"
#include "gdktoplevel.h"
#include "gdkdragsurface.h"

#define C_ENUM(v) ((gint) v)
#define C_FLAGS(v) ((guint) v)

/* enumerations from "gdkdevice.h" */

GType
gdk_input_source_get_type (void)
{
static gsize gtype_id = 0;
static const GEnumValue values[] = {
    { C_ENUM(GDK_SOURCE_MOUSE), "GDK_SOURCE_MOUSE", "mouse" },
    { C_ENUM(GDK_SOURCE_PEN), "GDK_SOURCE_PEN", "pen" },
    { C_ENUM(GDK_SOURCE_KEYBOARD), "GDK_SOURCE_KEYBOARD", "keyboard" },
    { C_ENUM(GDK_SOURCE_TOUCHSCREEN), "GDK_SOURCE_TOUCHSCREEN", "touchscreen" },
    { C_ENUM(GDK_SOURCE_TOUCHPAD), "GDK_SOURCE_TOUCHPAD", "touchpad" },
    { C_ENUM(GDK_SOURCE_TRACKPOINT), "GDK_SOURCE_TRACKPOINT", "trackpoint" },
    { C_ENUM(GDK_SOURCE_TABLET_PAD), "GDK_SOURCE_TABLET_PAD", "tablet-pad" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_enum_register_static (g_intern_static_string ("GdkInputSource"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

/* enumerations from "gdkdevicepad.h" */

GType
gdk_device_pad_feature_get_type (void)
{
static gsize gtype_id = 0;
static const GEnumValue values[] = {
    { C_ENUM(GDK_DEVICE_PAD_FEATURE_BUTTON), "GDK_DEVICE_PAD_FEATURE_BUTTON", "button" },
    { C_ENUM(GDK_DEVICE_PAD_FEATURE_RING), "GDK_DEVICE_PAD_FEATURE_RING", "ring" },
    { C_ENUM(GDK_DEVICE_PAD_FEATURE_STRIP), "GDK_DEVICE_PAD_FEATURE_STRIP", "strip" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_enum_register_static (g_intern_static_string ("GdkDevicePadFeature"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

/* enumerations from "gdkdevicetool.h" */

GType
gdk_device_tool_type_get_type (void)
{
static gsize gtype_id = 0;
static const GEnumValue values[] = {
    { C_ENUM(GDK_DEVICE_TOOL_TYPE_UNKNOWN), "GDK_DEVICE_TOOL_TYPE_UNKNOWN", "unknown" },
    { C_ENUM(GDK_DEVICE_TOOL_TYPE_PEN), "GDK_DEVICE_TOOL_TYPE_PEN", "pen" },
    { C_ENUM(GDK_DEVICE_TOOL_TYPE_ERASER), "GDK_DEVICE_TOOL_TYPE_ERASER", "eraser" },
    { C_ENUM(GDK_DEVICE_TOOL_TYPE_BRUSH), "GDK_DEVICE_TOOL_TYPE_BRUSH", "brush" },
    { C_ENUM(GDK_DEVICE_TOOL_TYPE_PENCIL), "GDK_DEVICE_TOOL_TYPE_PENCIL", "pencil" },
    { C_ENUM(GDK_DEVICE_TOOL_TYPE_AIRBRUSH), "GDK_DEVICE_TOOL_TYPE_AIRBRUSH", "airbrush" },
    { C_ENUM(GDK_DEVICE_TOOL_TYPE_MOUSE), "GDK_DEVICE_TOOL_TYPE_MOUSE", "mouse" },
    { C_ENUM(GDK_DEVICE_TOOL_TYPE_LENS), "GDK_DEVICE_TOOL_TYPE_LENS", "lens" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_enum_register_static (g_intern_static_string ("GdkDeviceToolType"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

/* enumerations from "gdkdrag.h" */

GType
gdk_drag_cancel_reason_get_type (void)
{
static gsize gtype_id = 0;
static const GEnumValue values[] = {
    { C_ENUM(GDK_DRAG_CANCEL_NO_TARGET), "GDK_DRAG_CANCEL_NO_TARGET", "no-target" },
    { C_ENUM(GDK_DRAG_CANCEL_USER_CANCELLED), "GDK_DRAG_CANCEL_USER_CANCELLED", "user-cancelled" },
    { C_ENUM(GDK_DRAG_CANCEL_ERROR), "GDK_DRAG_CANCEL_ERROR", "error" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_enum_register_static (g_intern_static_string ("GdkDragCancelReason"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

/* enumerations from "gdkenums.h" */

GType
gdk_gl_api_get_type (void)
{
static gsize gtype_id = 0;
static const GFlagsValue values[] = {
    { C_FLAGS(GDK_GL_API_GL), "GDK_GL_API_GL", "gl" },
    { C_FLAGS(GDK_GL_API_GLES), "GDK_GL_API_GLES", "gles" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_flags_register_static (g_intern_static_string ("GdkGLAPI"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

GType
gdk_gravity_get_type (void)
{
static gsize gtype_id = 0;
static const GEnumValue values[] = {
    { C_ENUM(GDK_GRAVITY_NORTH_WEST), "GDK_GRAVITY_NORTH_WEST", "north-west" },
    { C_ENUM(GDK_GRAVITY_NORTH), "GDK_GRAVITY_NORTH", "north" },
    { C_ENUM(GDK_GRAVITY_NORTH_EAST), "GDK_GRAVITY_NORTH_EAST", "north-east" },
    { C_ENUM(GDK_GRAVITY_WEST), "GDK_GRAVITY_WEST", "west" },
    { C_ENUM(GDK_GRAVITY_CENTER), "GDK_GRAVITY_CENTER", "center" },
    { C_ENUM(GDK_GRAVITY_EAST), "GDK_GRAVITY_EAST", "east" },
    { C_ENUM(GDK_GRAVITY_SOUTH_WEST), "GDK_GRAVITY_SOUTH_WEST", "south-west" },
    { C_ENUM(GDK_GRAVITY_SOUTH), "GDK_GRAVITY_SOUTH", "south" },
    { C_ENUM(GDK_GRAVITY_SOUTH_EAST), "GDK_GRAVITY_SOUTH_EAST", "south-east" },
    { C_ENUM(GDK_GRAVITY_STATIC), "GDK_GRAVITY_STATIC", "static" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_enum_register_static (g_intern_static_string ("GdkGravity"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

GType
gdk_modifier_type_get_type (void)
{
static gsize gtype_id = 0;
static const GFlagsValue values[] = {
    { C_FLAGS(GDK_NO_MODIFIER_MASK), "GDK_NO_MODIFIER_MASK", "no-modifier-mask" },
    { C_FLAGS(GDK_SHIFT_MASK), "GDK_SHIFT_MASK", "shift-mask" },
    { C_FLAGS(GDK_LOCK_MASK), "GDK_LOCK_MASK", "lock-mask" },
    { C_FLAGS(GDK_CONTROL_MASK), "GDK_CONTROL_MASK", "control-mask" },
    { C_FLAGS(GDK_ALT_MASK), "GDK_ALT_MASK", "alt-mask" },
    { C_FLAGS(GDK_BUTTON1_MASK), "GDK_BUTTON1_MASK", "button1-mask" },
    { C_FLAGS(GDK_BUTTON2_MASK), "GDK_BUTTON2_MASK", "button2-mask" },
    { C_FLAGS(GDK_BUTTON3_MASK), "GDK_BUTTON3_MASK", "button3-mask" },
    { C_FLAGS(GDK_BUTTON4_MASK), "GDK_BUTTON4_MASK", "button4-mask" },
    { C_FLAGS(GDK_BUTTON5_MASK), "GDK_BUTTON5_MASK", "button5-mask" },
    { C_FLAGS(GDK_SUPER_MASK), "GDK_SUPER_MASK", "super-mask" },
    { C_FLAGS(GDK_HYPER_MASK), "GDK_HYPER_MASK", "hyper-mask" },
    { C_FLAGS(GDK_META_MASK), "GDK_META_MASK", "meta-mask" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_flags_register_static (g_intern_static_string ("GdkModifierType"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

GType
gdk_dmabuf_error_get_type (void)
{
static gsize gtype_id = 0;
static const GEnumValue values[] = {
    { C_ENUM(GDK_DMABUF_ERROR_NOT_AVAILABLE), "GDK_DMABUF_ERROR_NOT_AVAILABLE", "not-available" },
    { C_ENUM(GDK_DMABUF_ERROR_UNSUPPORTED_FORMAT), "GDK_DMABUF_ERROR_UNSUPPORTED_FORMAT", "unsupported-format" },
    { C_ENUM(GDK_DMABUF_ERROR_CREATION_FAILED), "GDK_DMABUF_ERROR_CREATION_FAILED", "creation-failed" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_enum_register_static (g_intern_static_string ("GdkDmabufError"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

GType
gdk_gl_error_get_type (void)
{
static gsize gtype_id = 0;
static const GEnumValue values[] = {
    { C_ENUM(GDK_GL_ERROR_NOT_AVAILABLE), "GDK_GL_ERROR_NOT_AVAILABLE", "not-available" },
    { C_ENUM(GDK_GL_ERROR_UNSUPPORTED_FORMAT), "GDK_GL_ERROR_UNSUPPORTED_FORMAT", "unsupported-format" },
    { C_ENUM(GDK_GL_ERROR_UNSUPPORTED_PROFILE), "GDK_GL_ERROR_UNSUPPORTED_PROFILE", "unsupported-profile" },
    { C_ENUM(GDK_GL_ERROR_COMPILATION_FAILED), "GDK_GL_ERROR_COMPILATION_FAILED", "compilation-failed" },
    { C_ENUM(GDK_GL_ERROR_LINK_FAILED), "GDK_GL_ERROR_LINK_FAILED", "link-failed" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_enum_register_static (g_intern_static_string ("GdkGLError"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

GType
gdk_vulkan_error_get_type (void)
{
static gsize gtype_id = 0;
static const GEnumValue values[] = {
    { C_ENUM(GDK_VULKAN_ERROR_UNSUPPORTED), "GDK_VULKAN_ERROR_UNSUPPORTED", "unsupported" },
    { C_ENUM(GDK_VULKAN_ERROR_NOT_AVAILABLE), "GDK_VULKAN_ERROR_NOT_AVAILABLE", "not-available" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_enum_register_static (g_intern_static_string ("GdkVulkanError"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

GType
gdk_axis_use_get_type (void)
{
static gsize gtype_id = 0;
static const GEnumValue values[] = {
    { C_ENUM(GDK_AXIS_IGNORE), "GDK_AXIS_IGNORE", "ignore" },
    { C_ENUM(GDK_AXIS_X), "GDK_AXIS_X", "x" },
    { C_ENUM(GDK_AXIS_Y), "GDK_AXIS_Y", "y" },
    { C_ENUM(GDK_AXIS_DELTA_X), "GDK_AXIS_DELTA_X", "delta-x" },
    { C_ENUM(GDK_AXIS_DELTA_Y), "GDK_AXIS_DELTA_Y", "delta-y" },
    { C_ENUM(GDK_AXIS_PRESSURE), "GDK_AXIS_PRESSURE", "pressure" },
    { C_ENUM(GDK_AXIS_XTILT), "GDK_AXIS_XTILT", "xtilt" },
    { C_ENUM(GDK_AXIS_YTILT), "GDK_AXIS_YTILT", "ytilt" },
    { C_ENUM(GDK_AXIS_WHEEL), "GDK_AXIS_WHEEL", "wheel" },
    { C_ENUM(GDK_AXIS_DISTANCE), "GDK_AXIS_DISTANCE", "distance" },
    { C_ENUM(GDK_AXIS_ROTATION), "GDK_AXIS_ROTATION", "rotation" },
    { C_ENUM(GDK_AXIS_SLIDER), "GDK_AXIS_SLIDER", "slider" },
    { C_ENUM(GDK_AXIS_LAST), "GDK_AXIS_LAST", "last" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_enum_register_static (g_intern_static_string ("GdkAxisUse"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

GType
gdk_axis_flags_get_type (void)
{
static gsize gtype_id = 0;
static const GFlagsValue values[] = {
    { C_FLAGS(GDK_AXIS_FLAG_X), "GDK_AXIS_FLAG_X", "x" },
    { C_FLAGS(GDK_AXIS_FLAG_Y), "GDK_AXIS_FLAG_Y", "y" },
    { C_FLAGS(GDK_AXIS_FLAG_DELTA_X), "GDK_AXIS_FLAG_DELTA_X", "delta-x" },
    { C_FLAGS(GDK_AXIS_FLAG_DELTA_Y), "GDK_AXIS_FLAG_DELTA_Y", "delta-y" },
    { C_FLAGS(GDK_AXIS_FLAG_PRESSURE), "GDK_AXIS_FLAG_PRESSURE", "pressure" },
    { C_FLAGS(GDK_AXIS_FLAG_XTILT), "GDK_AXIS_FLAG_XTILT", "xtilt" },
    { C_FLAGS(GDK_AXIS_FLAG_YTILT), "GDK_AXIS_FLAG_YTILT", "ytilt" },
    { C_FLAGS(GDK_AXIS_FLAG_WHEEL), "GDK_AXIS_FLAG_WHEEL", "wheel" },
    { C_FLAGS(GDK_AXIS_FLAG_DISTANCE), "GDK_AXIS_FLAG_DISTANCE", "distance" },
    { C_FLAGS(GDK_AXIS_FLAG_ROTATION), "GDK_AXIS_FLAG_ROTATION", "rotation" },
    { C_FLAGS(GDK_AXIS_FLAG_SLIDER), "GDK_AXIS_FLAG_SLIDER", "slider" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_flags_register_static (g_intern_static_string ("GdkAxisFlags"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

GType
gdk_drag_action_get_type (void)
{
static gsize gtype_id = 0;
static const GFlagsValue values[] = {
    { C_FLAGS(GDK_ACTION_COPY), "GDK_ACTION_COPY", "copy" },
    { C_FLAGS(GDK_ACTION_MOVE), "GDK_ACTION_MOVE", "move" },
    { C_FLAGS(GDK_ACTION_LINK), "GDK_ACTION_LINK", "link" },
    { C_FLAGS(GDK_ACTION_ASK), "GDK_ACTION_ASK", "ask" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_flags_register_static (g_intern_static_string ("GdkDragAction"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

GType
gdk_memory_format_get_type (void)
{
static gsize gtype_id = 0;
static const GEnumValue values[] = {
    { C_ENUM(GDK_MEMORY_B8G8R8A8_PREMULTIPLIED), "GDK_MEMORY_B8G8R8A8_PREMULTIPLIED", "b8g8r8a8-premultiplied" },
    { C_ENUM(GDK_MEMORY_A8R8G8B8_PREMULTIPLIED), "GDK_MEMORY_A8R8G8B8_PREMULTIPLIED", "a8r8g8b8-premultiplied" },
    { C_ENUM(GDK_MEMORY_R8G8B8A8_PREMULTIPLIED), "GDK_MEMORY_R8G8B8A8_PREMULTIPLIED", "r8g8b8a8-premultiplied" },
    { C_ENUM(GDK_MEMORY_B8G8R8A8), "GDK_MEMORY_B8G8R8A8", "b8g8r8a8" },
    { C_ENUM(GDK_MEMORY_A8R8G8B8), "GDK_MEMORY_A8R8G8B8", "a8r8g8b8" },
    { C_ENUM(GDK_MEMORY_R8G8B8A8), "GDK_MEMORY_R8G8B8A8", "r8g8b8a8" },
    { C_ENUM(GDK_MEMORY_A8B8G8R8), "GDK_MEMORY_A8B8G8R8", "a8b8g8r8" },
    { C_ENUM(GDK_MEMORY_R8G8B8), "GDK_MEMORY_R8G8B8", "r8g8b8" },
    { C_ENUM(GDK_MEMORY_B8G8R8), "GDK_MEMORY_B8G8R8", "b8g8r8" },
    { C_ENUM(GDK_MEMORY_R16G16B16), "GDK_MEMORY_R16G16B16", "r16g16b16" },
    { C_ENUM(GDK_MEMORY_R16G16B16A16_PREMULTIPLIED), "GDK_MEMORY_R16G16B16A16_PREMULTIPLIED", "r16g16b16a16-premultiplied" },
    { C_ENUM(GDK_MEMORY_R16G16B16A16), "GDK_MEMORY_R16G16B16A16", "r16g16b16a16" },
    { C_ENUM(GDK_MEMORY_R16G16B16_FLOAT), "GDK_MEMORY_R16G16B16_FLOAT", "r16g16b16-float" },
    { C_ENUM(GDK_MEMORY_R16G16B16A16_FLOAT_PREMULTIPLIED), "GDK_MEMORY_R16G16B16A16_FLOAT_PREMULTIPLIED", "r16g16b16a16-float-premultiplied" },
    { C_ENUM(GDK_MEMORY_R16G16B16A16_FLOAT), "GDK_MEMORY_R16G16B16A16_FLOAT", "r16g16b16a16-float" },
    { C_ENUM(GDK_MEMORY_R32G32B32_FLOAT), "GDK_MEMORY_R32G32B32_FLOAT", "r32g32b32-float" },
    { C_ENUM(GDK_MEMORY_R32G32B32A32_FLOAT_PREMULTIPLIED), "GDK_MEMORY_R32G32B32A32_FLOAT_PREMULTIPLIED", "r32g32b32a32-float-premultiplied" },
    { C_ENUM(GDK_MEMORY_R32G32B32A32_FLOAT), "GDK_MEMORY_R32G32B32A32_FLOAT", "r32g32b32a32-float" },
    { C_ENUM(GDK_MEMORY_G8A8_PREMULTIPLIED), "GDK_MEMORY_G8A8_PREMULTIPLIED", "g8a8-premultiplied" },
    { C_ENUM(GDK_MEMORY_G8A8), "GDK_MEMORY_G8A8", "g8a8" },
    { C_ENUM(GDK_MEMORY_G8), "GDK_MEMORY_G8", "g8" },
    { C_ENUM(GDK_MEMORY_G16A16_PREMULTIPLIED), "GDK_MEMORY_G16A16_PREMULTIPLIED", "g16a16-premultiplied" },
    { C_ENUM(GDK_MEMORY_G16A16), "GDK_MEMORY_G16A16", "g16a16" },
    { C_ENUM(GDK_MEMORY_G16), "GDK_MEMORY_G16", "g16" },
    { C_ENUM(GDK_MEMORY_A8), "GDK_MEMORY_A8", "a8" },
    { C_ENUM(GDK_MEMORY_A16), "GDK_MEMORY_A16", "a16" },
    { C_ENUM(GDK_MEMORY_A16_FLOAT), "GDK_MEMORY_A16_FLOAT", "a16-float" },
    { C_ENUM(GDK_MEMORY_A32_FLOAT), "GDK_MEMORY_A32_FLOAT", "a32-float" },
    { C_ENUM(GDK_MEMORY_A8B8G8R8_PREMULTIPLIED), "GDK_MEMORY_A8B8G8R8_PREMULTIPLIED", "a8b8g8r8-premultiplied" },
    { C_ENUM(GDK_MEMORY_B8G8R8X8), "GDK_MEMORY_B8G8R8X8", "b8g8r8x8" },
    { C_ENUM(GDK_MEMORY_X8R8G8B8), "GDK_MEMORY_X8R8G8B8", "x8r8g8b8" },
    { C_ENUM(GDK_MEMORY_R8G8B8X8), "GDK_MEMORY_R8G8B8X8", "r8g8b8x8" },
    { C_ENUM(GDK_MEMORY_X8B8G8R8), "GDK_MEMORY_X8B8G8R8", "x8b8g8r8" },
    { C_ENUM(GDK_MEMORY_N_FORMATS), "GDK_MEMORY_N_FORMATS", "n-formats" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_enum_register_static (g_intern_static_string ("GdkMemoryFormat"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

/* enumerations from "gdkevents.h" */

GType
gdk_event_type_get_type (void)
{
static gsize gtype_id = 0;
static const GEnumValue values[] = {
    { C_ENUM(GDK_DELETE), "GDK_DELETE", "delete" },
    { C_ENUM(GDK_MOTION_NOTIFY), "GDK_MOTION_NOTIFY", "motion-notify" },
    { C_ENUM(GDK_BUTTON_PRESS), "GDK_BUTTON_PRESS", "button-press" },
    { C_ENUM(GDK_BUTTON_RELEASE), "GDK_BUTTON_RELEASE", "button-release" },
    { C_ENUM(GDK_KEY_PRESS), "GDK_KEY_PRESS", "key-press" },
    { C_ENUM(GDK_KEY_RELEASE), "GDK_KEY_RELEASE", "key-release" },
    { C_ENUM(GDK_ENTER_NOTIFY), "GDK_ENTER_NOTIFY", "enter-notify" },
    { C_ENUM(GDK_LEAVE_NOTIFY), "GDK_LEAVE_NOTIFY", "leave-notify" },
    { C_ENUM(GDK_FOCUS_CHANGE), "GDK_FOCUS_CHANGE", "focus-change" },
    { C_ENUM(GDK_PROXIMITY_IN), "GDK_PROXIMITY_IN", "proximity-in" },
    { C_ENUM(GDK_PROXIMITY_OUT), "GDK_PROXIMITY_OUT", "proximity-out" },
    { C_ENUM(GDK_DRAG_ENTER), "GDK_DRAG_ENTER", "drag-enter" },
    { C_ENUM(GDK_DRAG_LEAVE), "GDK_DRAG_LEAVE", "drag-leave" },
    { C_ENUM(GDK_DRAG_MOTION), "GDK_DRAG_MOTION", "drag-motion" },
    { C_ENUM(GDK_DROP_START), "GDK_DROP_START", "drop-start" },
    { C_ENUM(GDK_SCROLL), "GDK_SCROLL", "scroll" },
    { C_ENUM(GDK_GRAB_BROKEN), "GDK_GRAB_BROKEN", "grab-broken" },
    { C_ENUM(GDK_TOUCH_BEGIN), "GDK_TOUCH_BEGIN", "touch-begin" },
    { C_ENUM(GDK_TOUCH_UPDATE), "GDK_TOUCH_UPDATE", "touch-update" },
    { C_ENUM(GDK_TOUCH_END), "GDK_TOUCH_END", "touch-end" },
    { C_ENUM(GDK_TOUCH_CANCEL), "GDK_TOUCH_CANCEL", "touch-cancel" },
    { C_ENUM(GDK_TOUCHPAD_SWIPE), "GDK_TOUCHPAD_SWIPE", "touchpad-swipe" },
    { C_ENUM(GDK_TOUCHPAD_PINCH), "GDK_TOUCHPAD_PINCH", "touchpad-pinch" },
    { C_ENUM(GDK_PAD_BUTTON_PRESS), "GDK_PAD_BUTTON_PRESS", "pad-button-press" },
    { C_ENUM(GDK_PAD_BUTTON_RELEASE), "GDK_PAD_BUTTON_RELEASE", "pad-button-release" },
    { C_ENUM(GDK_PAD_RING), "GDK_PAD_RING", "pad-ring" },
    { C_ENUM(GDK_PAD_STRIP), "GDK_PAD_STRIP", "pad-strip" },
    { C_ENUM(GDK_PAD_GROUP_MODE), "GDK_PAD_GROUP_MODE", "pad-group-mode" },
    { C_ENUM(GDK_TOUCHPAD_HOLD), "GDK_TOUCHPAD_HOLD", "touchpad-hold" },
    { C_ENUM(GDK_EVENT_LAST), "GDK_EVENT_LAST", "event-last" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_enum_register_static (g_intern_static_string ("GdkEventType"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

GType
gdk_touchpad_gesture_phase_get_type (void)
{
static gsize gtype_id = 0;
static const GEnumValue values[] = {
    { C_ENUM(GDK_TOUCHPAD_GESTURE_PHASE_BEGIN), "GDK_TOUCHPAD_GESTURE_PHASE_BEGIN", "begin" },
    { C_ENUM(GDK_TOUCHPAD_GESTURE_PHASE_UPDATE), "GDK_TOUCHPAD_GESTURE_PHASE_UPDATE", "update" },
    { C_ENUM(GDK_TOUCHPAD_GESTURE_PHASE_END), "GDK_TOUCHPAD_GESTURE_PHASE_END", "end" },
    { C_ENUM(GDK_TOUCHPAD_GESTURE_PHASE_CANCEL), "GDK_TOUCHPAD_GESTURE_PHASE_CANCEL", "cancel" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_enum_register_static (g_intern_static_string ("GdkTouchpadGesturePhase"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

GType
gdk_scroll_direction_get_type (void)
{
static gsize gtype_id = 0;
static const GEnumValue values[] = {
    { C_ENUM(GDK_SCROLL_UP), "GDK_SCROLL_UP", "up" },
    { C_ENUM(GDK_SCROLL_DOWN), "GDK_SCROLL_DOWN", "down" },
    { C_ENUM(GDK_SCROLL_LEFT), "GDK_SCROLL_LEFT", "left" },
    { C_ENUM(GDK_SCROLL_RIGHT), "GDK_SCROLL_RIGHT", "right" },
    { C_ENUM(GDK_SCROLL_SMOOTH), "GDK_SCROLL_SMOOTH", "smooth" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_enum_register_static (g_intern_static_string ("GdkScrollDirection"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

GType
gdk_scroll_unit_get_type (void)
{
static gsize gtype_id = 0;
static const GEnumValue values[] = {
    { C_ENUM(GDK_SCROLL_UNIT_WHEEL), "GDK_SCROLL_UNIT_WHEEL", "wheel" },
    { C_ENUM(GDK_SCROLL_UNIT_SURFACE), "GDK_SCROLL_UNIT_SURFACE", "surface" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_enum_register_static (g_intern_static_string ("GdkScrollUnit"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

GType
gdk_notify_type_get_type (void)
{
static gsize gtype_id = 0;
static const GEnumValue values[] = {
    { C_ENUM(GDK_NOTIFY_ANCESTOR), "GDK_NOTIFY_ANCESTOR", "ancestor" },
    { C_ENUM(GDK_NOTIFY_VIRTUAL), "GDK_NOTIFY_VIRTUAL", "virtual" },
    { C_ENUM(GDK_NOTIFY_INFERIOR), "GDK_NOTIFY_INFERIOR", "inferior" },
    { C_ENUM(GDK_NOTIFY_NONLINEAR), "GDK_NOTIFY_NONLINEAR", "nonlinear" },
    { C_ENUM(GDK_NOTIFY_NONLINEAR_VIRTUAL), "GDK_NOTIFY_NONLINEAR_VIRTUAL", "nonlinear-virtual" },
    { C_ENUM(GDK_NOTIFY_UNKNOWN), "GDK_NOTIFY_UNKNOWN", "unknown" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_enum_register_static (g_intern_static_string ("GdkNotifyType"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

GType
gdk_crossing_mode_get_type (void)
{
static gsize gtype_id = 0;
static const GEnumValue values[] = {
    { C_ENUM(GDK_CROSSING_NORMAL), "GDK_CROSSING_NORMAL", "normal" },
    { C_ENUM(GDK_CROSSING_GRAB), "GDK_CROSSING_GRAB", "grab" },
    { C_ENUM(GDK_CROSSING_UNGRAB), "GDK_CROSSING_UNGRAB", "ungrab" },
    { C_ENUM(GDK_CROSSING_GTK_GRAB), "GDK_CROSSING_GTK_GRAB", "gtk-grab" },
    { C_ENUM(GDK_CROSSING_GTK_UNGRAB), "GDK_CROSSING_GTK_UNGRAB", "gtk-ungrab" },
    { C_ENUM(GDK_CROSSING_STATE_CHANGED), "GDK_CROSSING_STATE_CHANGED", "state-changed" },
    { C_ENUM(GDK_CROSSING_TOUCH_BEGIN), "GDK_CROSSING_TOUCH_BEGIN", "touch-begin" },
    { C_ENUM(GDK_CROSSING_TOUCH_END), "GDK_CROSSING_TOUCH_END", "touch-end" },
    { C_ENUM(GDK_CROSSING_DEVICE_SWITCH), "GDK_CROSSING_DEVICE_SWITCH", "device-switch" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_enum_register_static (g_intern_static_string ("GdkCrossingMode"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

GType
gdk_key_match_get_type (void)
{
static gsize gtype_id = 0;
static const GEnumValue values[] = {
    { C_ENUM(GDK_KEY_MATCH_NONE), "GDK_KEY_MATCH_NONE", "none" },
    { C_ENUM(GDK_KEY_MATCH_PARTIAL), "GDK_KEY_MATCH_PARTIAL", "partial" },
    { C_ENUM(GDK_KEY_MATCH_EXACT), "GDK_KEY_MATCH_EXACT", "exact" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_enum_register_static (g_intern_static_string ("GdkKeyMatch"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

/* enumerations from "gdkframeclock.h" */

GType
gdk_frame_clock_phase_get_type (void)
{
static gsize gtype_id = 0;
static const GFlagsValue values[] = {
    { C_FLAGS(GDK_FRAME_CLOCK_PHASE_NONE), "GDK_FRAME_CLOCK_PHASE_NONE", "none" },
    { C_FLAGS(GDK_FRAME_CLOCK_PHASE_FLUSH_EVENTS), "GDK_FRAME_CLOCK_PHASE_FLUSH_EVENTS", "flush-events" },
    { C_FLAGS(GDK_FRAME_CLOCK_PHASE_BEFORE_PAINT), "GDK_FRAME_CLOCK_PHASE_BEFORE_PAINT", "before-paint" },
    { C_FLAGS(GDK_FRAME_CLOCK_PHASE_UPDATE), "GDK_FRAME_CLOCK_PHASE_UPDATE", "update" },
    { C_FLAGS(GDK_FRAME_CLOCK_PHASE_LAYOUT), "GDK_FRAME_CLOCK_PHASE_LAYOUT", "layout" },
    { C_FLAGS(GDK_FRAME_CLOCK_PHASE_PAINT), "GDK_FRAME_CLOCK_PHASE_PAINT", "paint" },
    { C_FLAGS(GDK_FRAME_CLOCK_PHASE_RESUME_EVENTS), "GDK_FRAME_CLOCK_PHASE_RESUME_EVENTS", "resume-events" },
    { C_FLAGS(GDK_FRAME_CLOCK_PHASE_AFTER_PAINT), "GDK_FRAME_CLOCK_PHASE_AFTER_PAINT", "after-paint" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_flags_register_static (g_intern_static_string ("GdkFrameClockPhase"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

/* enumerations from "gdkmonitor.h" */

GType
gdk_subpixel_layout_get_type (void)
{
static gsize gtype_id = 0;
static const GEnumValue values[] = {
    { C_ENUM(GDK_SUBPIXEL_LAYOUT_UNKNOWN), "GDK_SUBPIXEL_LAYOUT_UNKNOWN", "unknown" },
    { C_ENUM(GDK_SUBPIXEL_LAYOUT_NONE), "GDK_SUBPIXEL_LAYOUT_NONE", "none" },
    { C_ENUM(GDK_SUBPIXEL_LAYOUT_HORIZONTAL_RGB), "GDK_SUBPIXEL_LAYOUT_HORIZONTAL_RGB", "horizontal-rgb" },
    { C_ENUM(GDK_SUBPIXEL_LAYOUT_HORIZONTAL_BGR), "GDK_SUBPIXEL_LAYOUT_HORIZONTAL_BGR", "horizontal-bgr" },
    { C_ENUM(GDK_SUBPIXEL_LAYOUT_VERTICAL_RGB), "GDK_SUBPIXEL_LAYOUT_VERTICAL_RGB", "vertical-rgb" },
    { C_ENUM(GDK_SUBPIXEL_LAYOUT_VERTICAL_BGR), "GDK_SUBPIXEL_LAYOUT_VERTICAL_BGR", "vertical-bgr" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_enum_register_static (g_intern_static_string ("GdkSubpixelLayout"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

/* enumerations from "gdkpaintable.h" */

GType
gdk_paintable_flags_get_type (void)
{
static gsize gtype_id = 0;
static const GFlagsValue values[] = {
    { C_FLAGS(GDK_PAINTABLE_STATIC_SIZE), "GDK_PAINTABLE_STATIC_SIZE", "size" },
    { C_FLAGS(GDK_PAINTABLE_STATIC_CONTENTS), "GDK_PAINTABLE_STATIC_CONTENTS", "contents" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_flags_register_static (g_intern_static_string ("GdkPaintableFlags"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

/* enumerations from "gdkpopuplayout.h" */

GType
gdk_anchor_hints_get_type (void)
{
static gsize gtype_id = 0;
static const GFlagsValue values[] = {
    { C_FLAGS(GDK_ANCHOR_FLIP_X), "GDK_ANCHOR_FLIP_X", "flip-x" },
    { C_FLAGS(GDK_ANCHOR_FLIP_Y), "GDK_ANCHOR_FLIP_Y", "flip-y" },
    { C_FLAGS(GDK_ANCHOR_SLIDE_X), "GDK_ANCHOR_SLIDE_X", "slide-x" },
    { C_FLAGS(GDK_ANCHOR_SLIDE_Y), "GDK_ANCHOR_SLIDE_Y", "slide-y" },
    { C_FLAGS(GDK_ANCHOR_RESIZE_X), "GDK_ANCHOR_RESIZE_X", "resize-x" },
    { C_FLAGS(GDK_ANCHOR_RESIZE_Y), "GDK_ANCHOR_RESIZE_Y", "resize-y" },
    { C_FLAGS(GDK_ANCHOR_FLIP), "GDK_ANCHOR_FLIP", "flip" },
    { C_FLAGS(GDK_ANCHOR_SLIDE), "GDK_ANCHOR_SLIDE", "slide" },
    { C_FLAGS(GDK_ANCHOR_RESIZE), "GDK_ANCHOR_RESIZE", "resize" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_flags_register_static (g_intern_static_string ("GdkAnchorHints"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

/* enumerations from "gdkseat.h" */

GType
gdk_seat_capabilities_get_type (void)
{
static gsize gtype_id = 0;
static const GFlagsValue values[] = {
    { C_FLAGS(GDK_SEAT_CAPABILITY_NONE), "GDK_SEAT_CAPABILITY_NONE", "none" },
    { C_FLAGS(GDK_SEAT_CAPABILITY_POINTER), "GDK_SEAT_CAPABILITY_POINTER", "pointer" },
    { C_FLAGS(GDK_SEAT_CAPABILITY_TOUCH), "GDK_SEAT_CAPABILITY_TOUCH", "touch" },
    { C_FLAGS(GDK_SEAT_CAPABILITY_TABLET_STYLUS), "GDK_SEAT_CAPABILITY_TABLET_STYLUS", "tablet-stylus" },
    { C_FLAGS(GDK_SEAT_CAPABILITY_KEYBOARD), "GDK_SEAT_CAPABILITY_KEYBOARD", "keyboard" },
    { C_FLAGS(GDK_SEAT_CAPABILITY_TABLET_PAD), "GDK_SEAT_CAPABILITY_TABLET_PAD", "tablet-pad" },
    { C_FLAGS(GDK_SEAT_CAPABILITY_ALL_POINTING), "GDK_SEAT_CAPABILITY_ALL_POINTING", "all-pointing" },
    { C_FLAGS(GDK_SEAT_CAPABILITY_ALL), "GDK_SEAT_CAPABILITY_ALL", "all" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_flags_register_static (g_intern_static_string ("GdkSeatCapabilities"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

/* enumerations from "gdktexture.h" */

GType
gdk_texture_error_get_type (void)
{
static gsize gtype_id = 0;
static const GEnumValue values[] = {
    { C_ENUM(GDK_TEXTURE_ERROR_TOO_LARGE), "GDK_TEXTURE_ERROR_TOO_LARGE", "too-large" },
    { C_ENUM(GDK_TEXTURE_ERROR_CORRUPT_IMAGE), "GDK_TEXTURE_ERROR_CORRUPT_IMAGE", "corrupt-image" },
    { C_ENUM(GDK_TEXTURE_ERROR_UNSUPPORTED_CONTENT), "GDK_TEXTURE_ERROR_UNSUPPORTED_CONTENT", "unsupported-content" },
    { C_ENUM(GDK_TEXTURE_ERROR_UNSUPPORTED_FORMAT), "GDK_TEXTURE_ERROR_UNSUPPORTED_FORMAT", "unsupported-format" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_enum_register_static (g_intern_static_string ("GdkTextureError"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

/* enumerations from "gdktoplevel.h" */

GType
gdk_surface_edge_get_type (void)
{
static gsize gtype_id = 0;
static const GEnumValue values[] = {
    { C_ENUM(GDK_SURFACE_EDGE_NORTH_WEST), "GDK_SURFACE_EDGE_NORTH_WEST", "north-west" },
    { C_ENUM(GDK_SURFACE_EDGE_NORTH), "GDK_SURFACE_EDGE_NORTH", "north" },
    { C_ENUM(GDK_SURFACE_EDGE_NORTH_EAST), "GDK_SURFACE_EDGE_NORTH_EAST", "north-east" },
    { C_ENUM(GDK_SURFACE_EDGE_WEST), "GDK_SURFACE_EDGE_WEST", "west" },
    { C_ENUM(GDK_SURFACE_EDGE_EAST), "GDK_SURFACE_EDGE_EAST", "east" },
    { C_ENUM(GDK_SURFACE_EDGE_SOUTH_WEST), "GDK_SURFACE_EDGE_SOUTH_WEST", "south-west" },
    { C_ENUM(GDK_SURFACE_EDGE_SOUTH), "GDK_SURFACE_EDGE_SOUTH", "south" },
    { C_ENUM(GDK_SURFACE_EDGE_SOUTH_EAST), "GDK_SURFACE_EDGE_SOUTH_EAST", "south-east" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_enum_register_static (g_intern_static_string ("GdkSurfaceEdge"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

GType
gdk_fullscreen_mode_get_type (void)
{
static gsize gtype_id = 0;
static const GEnumValue values[] = {
    { C_ENUM(GDK_FULLSCREEN_ON_CURRENT_MONITOR), "GDK_FULLSCREEN_ON_CURRENT_MONITOR", "current-monitor" },
    { C_ENUM(GDK_FULLSCREEN_ON_ALL_MONITORS), "GDK_FULLSCREEN_ON_ALL_MONITORS", "all-monitors" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_enum_register_static (g_intern_static_string ("GdkFullscreenMode"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

GType
gdk_toplevel_state_get_type (void)
{
static gsize gtype_id = 0;
static const GFlagsValue values[] = {
    { C_FLAGS(GDK_TOPLEVEL_STATE_MINIMIZED), "GDK_TOPLEVEL_STATE_MINIMIZED", "minimized" },
    { C_FLAGS(GDK_TOPLEVEL_STATE_MAXIMIZED), "GDK_TOPLEVEL_STATE_MAXIMIZED", "maximized" },
    { C_FLAGS(GDK_TOPLEVEL_STATE_STICKY), "GDK_TOPLEVEL_STATE_STICKY", "sticky" },
    { C_FLAGS(GDK_TOPLEVEL_STATE_FULLSCREEN), "GDK_TOPLEVEL_STATE_FULLSCREEN", "fullscreen" },
    { C_FLAGS(GDK_TOPLEVEL_STATE_ABOVE), "GDK_TOPLEVEL_STATE_ABOVE", "above" },
    { C_FLAGS(GDK_TOPLEVEL_STATE_BELOW), "GDK_TOPLEVEL_STATE_BELOW", "below" },
    { C_FLAGS(GDK_TOPLEVEL_STATE_FOCUSED), "GDK_TOPLEVEL_STATE_FOCUSED", "focused" },
    { C_FLAGS(GDK_TOPLEVEL_STATE_TILED), "GDK_TOPLEVEL_STATE_TILED", "tiled" },
    { C_FLAGS(GDK_TOPLEVEL_STATE_TOP_TILED), "GDK_TOPLEVEL_STATE_TOP_TILED", "top-tiled" },
    { C_FLAGS(GDK_TOPLEVEL_STATE_TOP_RESIZABLE), "GDK_TOPLEVEL_STATE_TOP_RESIZABLE", "top-resizable" },
    { C_FLAGS(GDK_TOPLEVEL_STATE_RIGHT_TILED), "GDK_TOPLEVEL_STATE_RIGHT_TILED", "right-tiled" },
    { C_FLAGS(GDK_TOPLEVEL_STATE_RIGHT_RESIZABLE), "GDK_TOPLEVEL_STATE_RIGHT_RESIZABLE", "right-resizable" },
    { C_FLAGS(GDK_TOPLEVEL_STATE_BOTTOM_TILED), "GDK_TOPLEVEL_STATE_BOTTOM_TILED", "bottom-tiled" },
    { C_FLAGS(GDK_TOPLEVEL_STATE_BOTTOM_RESIZABLE), "GDK_TOPLEVEL_STATE_BOTTOM_RESIZABLE", "bottom-resizable" },
    { C_FLAGS(GDK_TOPLEVEL_STATE_LEFT_TILED), "GDK_TOPLEVEL_STATE_LEFT_TILED", "left-tiled" },
    { C_FLAGS(GDK_TOPLEVEL_STATE_LEFT_RESIZABLE), "GDK_TOPLEVEL_STATE_LEFT_RESIZABLE", "left-resizable" },
    { C_FLAGS(GDK_TOPLEVEL_STATE_SUSPENDED), "GDK_TOPLEVEL_STATE_SUSPENDED", "suspended" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_flags_register_static (g_intern_static_string ("GdkToplevelState"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

GType
gdk_titlebar_gesture_get_type (void)
{
static gsize gtype_id = 0;
static const GEnumValue values[] = {
    { C_ENUM(GDK_TITLEBAR_GESTURE_DOUBLE_CLICK), "GDK_TITLEBAR_GESTURE_DOUBLE_CLICK", "double-click" },
    { C_ENUM(GDK_TITLEBAR_GESTURE_RIGHT_CLICK), "GDK_TITLEBAR_GESTURE_RIGHT_CLICK", "right-click" },
    { C_ENUM(GDK_TITLEBAR_GESTURE_MIDDLE_CLICK), "GDK_TITLEBAR_GESTURE_MIDDLE_CLICK", "middle-click" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_enum_register_static (g_intern_static_string ("GdkTitlebarGesture"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

/* Generated data ends here */

