/*
 * This file is generated by gdbus-codegen, do not modify it.
 *
 * The license of this code is the same as for the D-Bus interface description
 * it was derived from. Note that it links to GLib, so must comply with the
 * LGPL linking clauses.
 */

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include "atspi-value.h"

#include <string.h>

/* ------------------------------------------------------------------------ */
/* Definitions for org.a11y.atspi.Value */

const GDBusPropertyInfo atspi_value_interface__minimum_value_property =
{
  -1,  /* ref count */
  (gchar *) "MinimumValue",
  (gchar *) "d",
  G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
  NULL,  /* no annotations */
};

const GDBusPropertyInfo atspi_value_interface__maximum_value_property =
{
  -1,  /* ref count */
  (gchar *) "MaximumValue",
  (gchar *) "d",
  G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
  NULL,  /* no annotations */
};

const GDBusPropertyInfo atspi_value_interface__minimum_increment_property =
{
  -1,  /* ref count */
  (gchar *) "MinimumIncrement",
  (gchar *) "d",
  G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
  NULL,  /* no annotations */
};

const GDBusPropertyInfo atspi_value_interface__current_value_property =
{
  -1,  /* ref count */
  (gchar *) "CurrentValue",
  (gchar *) "d",
  G_DBUS_PROPERTY_INFO_FLAGS_READABLE | G_DBUS_PROPERTY_INFO_FLAGS_WRITABLE,
  NULL,  /* no annotations */
};

const GDBusPropertyInfo atspi_value_interface__text_property =
{
  -1,  /* ref count */
  (gchar *) "Text",
  (gchar *) "s",
  G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
  NULL,  /* no annotations */
};

const GDBusPropertyInfo * const atspi_value_interface_properties[] =
{
  &atspi_value_interface__minimum_value_property,
  &atspi_value_interface__maximum_value_property,
  &atspi_value_interface__minimum_increment_property,
  &atspi_value_interface__current_value_property,
  &atspi_value_interface__text_property,
  NULL,
};

const GDBusInterfaceInfo atspi_value_interface =
{
  -1,  /* ref count */
  (gchar *) "org.a11y.atspi.Value",
  NULL,  /* no methods */
  NULL,  /* no signals */
  (GDBusPropertyInfo **) atspi_value_interface_properties,
  NULL,  /* no annotations */
};

