/*
 * This file is generated by gdbus-codegen, do not modify it.
 *
 * The license of this code is the same as for the D-Bus interface description
 * it was derived from. Note that it links to GLib, so must comply with the
 * LGPL linking clauses.
 */

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include "atspi-text.h"

#include <string.h>

/* ------------------------------------------------------------------------ */
/* Definitions for org.a11y.atspi.Text */

const GDBusArgInfo atspi_text_interface__get_string_at_offset_method_in_args__offset_arg =
{
  -1,  /* ref count */
  (gchar *) "offset",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_text_interface__get_string_at_offset_method_in_args__granularity_arg =
{
  -1,  /* ref count */
  (gchar *) "granularity",
  (gchar *) "u",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_text_interface__get_string_at_offset_method_in_args[] =
{
  &atspi_text_interface__get_string_at_offset_method_in_args__offset_arg,
  &atspi_text_interface__get_string_at_offset_method_in_args__granularity_arg,
  NULL,
};

const GDBusArgInfo atspi_text_interface__get_string_at_offset_method_out_args__unnamed_arg2_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg2",
  (gchar *) "s",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_text_interface__get_string_at_offset_method_out_args__startOffset_arg =
{
  -1,  /* ref count */
  (gchar *) "startOffset",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_text_interface__get_string_at_offset_method_out_args__endOffset_arg =
{
  -1,  /* ref count */
  (gchar *) "endOffset",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_text_interface__get_string_at_offset_method_out_args[] =
{
  &atspi_text_interface__get_string_at_offset_method_out_args__unnamed_arg2_arg,
  &atspi_text_interface__get_string_at_offset_method_out_args__startOffset_arg,
  &atspi_text_interface__get_string_at_offset_method_out_args__endOffset_arg,
  NULL,
};

const GDBusMethodInfo atspi_text_interface__get_string_at_offset_method =
{
  -1,  /* ref count */
  (gchar *) "GetStringAtOffset",
  (GDBusArgInfo **) atspi_text_interface__get_string_at_offset_method_in_args,
  (GDBusArgInfo **) atspi_text_interface__get_string_at_offset_method_out_args,
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_text_interface__get_text_method_in_args__startOffset_arg =
{
  -1,  /* ref count */
  (gchar *) "startOffset",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_text_interface__get_text_method_in_args__endOffset_arg =
{
  -1,  /* ref count */
  (gchar *) "endOffset",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_text_interface__get_text_method_in_args[] =
{
  &atspi_text_interface__get_text_method_in_args__startOffset_arg,
  &atspi_text_interface__get_text_method_in_args__endOffset_arg,
  NULL,
};

const GDBusArgInfo atspi_text_interface__get_text_method_out_args__unnamed_arg2_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg2",
  (gchar *) "s",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_text_interface__get_text_method_out_args[] =
{
  &atspi_text_interface__get_text_method_out_args__unnamed_arg2_arg,
  NULL,
};

const GDBusMethodInfo atspi_text_interface__get_text_method =
{
  -1,  /* ref count */
  (gchar *) "GetText",
  (GDBusArgInfo **) atspi_text_interface__get_text_method_in_args,
  (GDBusArgInfo **) atspi_text_interface__get_text_method_out_args,
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_text_interface__set_caret_offset_method_in_args__offset_arg =
{
  -1,  /* ref count */
  (gchar *) "offset",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_text_interface__set_caret_offset_method_in_args[] =
{
  &atspi_text_interface__set_caret_offset_method_in_args__offset_arg,
  NULL,
};

const GDBusArgInfo atspi_text_interface__set_caret_offset_method_out_args__unnamed_arg1_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg1",
  (gchar *) "b",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_text_interface__set_caret_offset_method_out_args[] =
{
  &atspi_text_interface__set_caret_offset_method_out_args__unnamed_arg1_arg,
  NULL,
};

const GDBusMethodInfo atspi_text_interface__set_caret_offset_method =
{
  -1,  /* ref count */
  (gchar *) "SetCaretOffset",
  (GDBusArgInfo **) atspi_text_interface__set_caret_offset_method_in_args,
  (GDBusArgInfo **) atspi_text_interface__set_caret_offset_method_out_args,
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_text_interface__get_text_before_offset_method_in_args__offset_arg =
{
  -1,  /* ref count */
  (gchar *) "offset",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_text_interface__get_text_before_offset_method_in_args__type_arg =
{
  -1,  /* ref count */
  (gchar *) "type",
  (gchar *) "u",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_text_interface__get_text_before_offset_method_in_args[] =
{
  &atspi_text_interface__get_text_before_offset_method_in_args__offset_arg,
  &atspi_text_interface__get_text_before_offset_method_in_args__type_arg,
  NULL,
};

const GDBusArgInfo atspi_text_interface__get_text_before_offset_method_out_args__unnamed_arg2_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg2",
  (gchar *) "s",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_text_interface__get_text_before_offset_method_out_args__startOffset_arg =
{
  -1,  /* ref count */
  (gchar *) "startOffset",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_text_interface__get_text_before_offset_method_out_args__endOffset_arg =
{
  -1,  /* ref count */
  (gchar *) "endOffset",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_text_interface__get_text_before_offset_method_out_args[] =
{
  &atspi_text_interface__get_text_before_offset_method_out_args__unnamed_arg2_arg,
  &atspi_text_interface__get_text_before_offset_method_out_args__startOffset_arg,
  &atspi_text_interface__get_text_before_offset_method_out_args__endOffset_arg,
  NULL,
};

const GDBusMethodInfo atspi_text_interface__get_text_before_offset_method =
{
  -1,  /* ref count */
  (gchar *) "GetTextBeforeOffset",
  (GDBusArgInfo **) atspi_text_interface__get_text_before_offset_method_in_args,
  (GDBusArgInfo **) atspi_text_interface__get_text_before_offset_method_out_args,
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_text_interface__get_text_at_offset_method_in_args__offset_arg =
{
  -1,  /* ref count */
  (gchar *) "offset",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_text_interface__get_text_at_offset_method_in_args__type_arg =
{
  -1,  /* ref count */
  (gchar *) "type",
  (gchar *) "u",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_text_interface__get_text_at_offset_method_in_args[] =
{
  &atspi_text_interface__get_text_at_offset_method_in_args__offset_arg,
  &atspi_text_interface__get_text_at_offset_method_in_args__type_arg,
  NULL,
};

const GDBusArgInfo atspi_text_interface__get_text_at_offset_method_out_args__unnamed_arg2_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg2",
  (gchar *) "s",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_text_interface__get_text_at_offset_method_out_args__startOffset_arg =
{
  -1,  /* ref count */
  (gchar *) "startOffset",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_text_interface__get_text_at_offset_method_out_args__endOffset_arg =
{
  -1,  /* ref count */
  (gchar *) "endOffset",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_text_interface__get_text_at_offset_method_out_args[] =
{
  &atspi_text_interface__get_text_at_offset_method_out_args__unnamed_arg2_arg,
  &atspi_text_interface__get_text_at_offset_method_out_args__startOffset_arg,
  &atspi_text_interface__get_text_at_offset_method_out_args__endOffset_arg,
  NULL,
};

const GDBusMethodInfo atspi_text_interface__get_text_at_offset_method =
{
  -1,  /* ref count */
  (gchar *) "GetTextAtOffset",
  (GDBusArgInfo **) atspi_text_interface__get_text_at_offset_method_in_args,
  (GDBusArgInfo **) atspi_text_interface__get_text_at_offset_method_out_args,
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_text_interface__get_text_after_offset_method_in_args__offset_arg =
{
  -1,  /* ref count */
  (gchar *) "offset",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_text_interface__get_text_after_offset_method_in_args__type_arg =
{
  -1,  /* ref count */
  (gchar *) "type",
  (gchar *) "u",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_text_interface__get_text_after_offset_method_in_args[] =
{
  &atspi_text_interface__get_text_after_offset_method_in_args__offset_arg,
  &atspi_text_interface__get_text_after_offset_method_in_args__type_arg,
  NULL,
};

const GDBusArgInfo atspi_text_interface__get_text_after_offset_method_out_args__unnamed_arg2_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg2",
  (gchar *) "s",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_text_interface__get_text_after_offset_method_out_args__startOffset_arg =
{
  -1,  /* ref count */
  (gchar *) "startOffset",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_text_interface__get_text_after_offset_method_out_args__endOffset_arg =
{
  -1,  /* ref count */
  (gchar *) "endOffset",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_text_interface__get_text_after_offset_method_out_args[] =
{
  &atspi_text_interface__get_text_after_offset_method_out_args__unnamed_arg2_arg,
  &atspi_text_interface__get_text_after_offset_method_out_args__startOffset_arg,
  &atspi_text_interface__get_text_after_offset_method_out_args__endOffset_arg,
  NULL,
};

const GDBusMethodInfo atspi_text_interface__get_text_after_offset_method =
{
  -1,  /* ref count */
  (gchar *) "GetTextAfterOffset",
  (GDBusArgInfo **) atspi_text_interface__get_text_after_offset_method_in_args,
  (GDBusArgInfo **) atspi_text_interface__get_text_after_offset_method_out_args,
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_text_interface__get_character_at_offset_method_in_args__offset_arg =
{
  -1,  /* ref count */
  (gchar *) "offset",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_text_interface__get_character_at_offset_method_in_args[] =
{
  &atspi_text_interface__get_character_at_offset_method_in_args__offset_arg,
  NULL,
};

const GDBusArgInfo atspi_text_interface__get_character_at_offset_method_out_args__unnamed_arg1_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg1",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_text_interface__get_character_at_offset_method_out_args[] =
{
  &atspi_text_interface__get_character_at_offset_method_out_args__unnamed_arg1_arg,
  NULL,
};

const GDBusMethodInfo atspi_text_interface__get_character_at_offset_method =
{
  -1,  /* ref count */
  (gchar *) "GetCharacterAtOffset",
  (GDBusArgInfo **) atspi_text_interface__get_character_at_offset_method_in_args,
  (GDBusArgInfo **) atspi_text_interface__get_character_at_offset_method_out_args,
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_text_interface__get_attribute_value_method_in_args__offset_arg =
{
  -1,  /* ref count */
  (gchar *) "offset",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_text_interface__get_attribute_value_method_in_args__attributeName_arg =
{
  -1,  /* ref count */
  (gchar *) "attributeName",
  (gchar *) "s",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_text_interface__get_attribute_value_method_in_args[] =
{
  &atspi_text_interface__get_attribute_value_method_in_args__offset_arg,
  &atspi_text_interface__get_attribute_value_method_in_args__attributeName_arg,
  NULL,
};

const GDBusArgInfo atspi_text_interface__get_attribute_value_method_out_args__unnamed_arg2_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg2",
  (gchar *) "s",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_text_interface__get_attribute_value_method_out_args[] =
{
  &atspi_text_interface__get_attribute_value_method_out_args__unnamed_arg2_arg,
  NULL,
};

const GDBusMethodInfo atspi_text_interface__get_attribute_value_method =
{
  -1,  /* ref count */
  (gchar *) "GetAttributeValue",
  (GDBusArgInfo **) atspi_text_interface__get_attribute_value_method_in_args,
  (GDBusArgInfo **) atspi_text_interface__get_attribute_value_method_out_args,
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_text_interface__get_attributes_method_in_args__offset_arg =
{
  -1,  /* ref count */
  (gchar *) "offset",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_text_interface__get_attributes_method_in_args[] =
{
  &atspi_text_interface__get_attributes_method_in_args__offset_arg,
  NULL,
};

const GDBusArgInfo atspi_text_interface__get_attributes_method_out_args__unnamed_arg1_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg1",
  (gchar *) "a{ss}",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_text_interface__get_attributes_method_out_args__startOffset_arg =
{
  -1,  /* ref count */
  (gchar *) "startOffset",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_text_interface__get_attributes_method_out_args__endOffset_arg =
{
  -1,  /* ref count */
  (gchar *) "endOffset",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_text_interface__get_attributes_method_out_args[] =
{
  &atspi_text_interface__get_attributes_method_out_args__unnamed_arg1_arg,
  &atspi_text_interface__get_attributes_method_out_args__startOffset_arg,
  &atspi_text_interface__get_attributes_method_out_args__endOffset_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_text_interface__get_attributes_method_annotations__org_qtproject__qt_dbus__qt_type_name__out0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.Out0",
  (gchar *) "QSpiAttributeSet",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_text_interface__get_attributes_method_annotations[] =
{
  &atspi_text_interface__get_attributes_method_annotations__org_qtproject__qt_dbus__qt_type_name__out0_annotation,
  NULL,
};

const GDBusMethodInfo atspi_text_interface__get_attributes_method =
{
  -1,  /* ref count */
  (gchar *) "GetAttributes",
  (GDBusArgInfo **) atspi_text_interface__get_attributes_method_in_args,
  (GDBusArgInfo **) atspi_text_interface__get_attributes_method_out_args,
  (GDBusAnnotationInfo **) atspi_text_interface__get_attributes_method_annotations,
};

const GDBusArgInfo atspi_text_interface__get_default_attributes_method_out_args__unnamed_arg0_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg0",
  (gchar *) "a{ss}",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_text_interface__get_default_attributes_method_out_args[] =
{
  &atspi_text_interface__get_default_attributes_method_out_args__unnamed_arg0_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_text_interface__get_default_attributes_method_annotations__org_qtproject__qt_dbus__qt_type_name__out0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.Out0",
  (gchar *) "QSpiAttributeSet",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_text_interface__get_default_attributes_method_annotations[] =
{
  &atspi_text_interface__get_default_attributes_method_annotations__org_qtproject__qt_dbus__qt_type_name__out0_annotation,
  NULL,
};

const GDBusMethodInfo atspi_text_interface__get_default_attributes_method =
{
  -1,  /* ref count */
  (gchar *) "GetDefaultAttributes",
  NULL,  /* no in args */
  (GDBusArgInfo **) atspi_text_interface__get_default_attributes_method_out_args,
  (GDBusAnnotationInfo **) atspi_text_interface__get_default_attributes_method_annotations,
};

const GDBusArgInfo atspi_text_interface__get_character_extents_method_in_args__offset_arg =
{
  -1,  /* ref count */
  (gchar *) "offset",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_text_interface__get_character_extents_method_in_args__coordType_arg =
{
  -1,  /* ref count */
  (gchar *) "coordType",
  (gchar *) "u",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_text_interface__get_character_extents_method_in_args[] =
{
  &atspi_text_interface__get_character_extents_method_in_args__offset_arg,
  &atspi_text_interface__get_character_extents_method_in_args__coordType_arg,
  NULL,
};

const GDBusArgInfo atspi_text_interface__get_character_extents_method_out_args__x_arg =
{
  -1,  /* ref count */
  (gchar *) "x",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_text_interface__get_character_extents_method_out_args__y_arg =
{
  -1,  /* ref count */
  (gchar *) "y",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_text_interface__get_character_extents_method_out_args__width_arg =
{
  -1,  /* ref count */
  (gchar *) "width",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_text_interface__get_character_extents_method_out_args__height_arg =
{
  -1,  /* ref count */
  (gchar *) "height",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_text_interface__get_character_extents_method_out_args[] =
{
  &atspi_text_interface__get_character_extents_method_out_args__x_arg,
  &atspi_text_interface__get_character_extents_method_out_args__y_arg,
  &atspi_text_interface__get_character_extents_method_out_args__width_arg,
  &atspi_text_interface__get_character_extents_method_out_args__height_arg,
  NULL,
};

const GDBusMethodInfo atspi_text_interface__get_character_extents_method =
{
  -1,  /* ref count */
  (gchar *) "GetCharacterExtents",
  (GDBusArgInfo **) atspi_text_interface__get_character_extents_method_in_args,
  (GDBusArgInfo **) atspi_text_interface__get_character_extents_method_out_args,
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_text_interface__get_offset_at_point_method_in_args__x_arg =
{
  -1,  /* ref count */
  (gchar *) "x",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_text_interface__get_offset_at_point_method_in_args__y_arg =
{
  -1,  /* ref count */
  (gchar *) "y",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_text_interface__get_offset_at_point_method_in_args__coordType_arg =
{
  -1,  /* ref count */
  (gchar *) "coordType",
  (gchar *) "u",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_text_interface__get_offset_at_point_method_in_args[] =
{
  &atspi_text_interface__get_offset_at_point_method_in_args__x_arg,
  &atspi_text_interface__get_offset_at_point_method_in_args__y_arg,
  &atspi_text_interface__get_offset_at_point_method_in_args__coordType_arg,
  NULL,
};

const GDBusArgInfo atspi_text_interface__get_offset_at_point_method_out_args__unnamed_arg3_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg3",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_text_interface__get_offset_at_point_method_out_args[] =
{
  &atspi_text_interface__get_offset_at_point_method_out_args__unnamed_arg3_arg,
  NULL,
};

const GDBusMethodInfo atspi_text_interface__get_offset_at_point_method =
{
  -1,  /* ref count */
  (gchar *) "GetOffsetAtPoint",
  (GDBusArgInfo **) atspi_text_interface__get_offset_at_point_method_in_args,
  (GDBusArgInfo **) atspi_text_interface__get_offset_at_point_method_out_args,
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_text_interface__get_nselections_method_out_args__unnamed_arg0_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg0",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_text_interface__get_nselections_method_out_args[] =
{
  &atspi_text_interface__get_nselections_method_out_args__unnamed_arg0_arg,
  NULL,
};

const GDBusMethodInfo atspi_text_interface__get_nselections_method =
{
  -1,  /* ref count */
  (gchar *) "GetNSelections",
  NULL,  /* no in args */
  (GDBusArgInfo **) atspi_text_interface__get_nselections_method_out_args,
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_text_interface__get_selection_method_in_args__selectionNum_arg =
{
  -1,  /* ref count */
  (gchar *) "selectionNum",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_text_interface__get_selection_method_in_args[] =
{
  &atspi_text_interface__get_selection_method_in_args__selectionNum_arg,
  NULL,
};

const GDBusArgInfo atspi_text_interface__get_selection_method_out_args__startOffset_arg =
{
  -1,  /* ref count */
  (gchar *) "startOffset",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_text_interface__get_selection_method_out_args__endOffset_arg =
{
  -1,  /* ref count */
  (gchar *) "endOffset",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_text_interface__get_selection_method_out_args[] =
{
  &atspi_text_interface__get_selection_method_out_args__startOffset_arg,
  &atspi_text_interface__get_selection_method_out_args__endOffset_arg,
  NULL,
};

const GDBusMethodInfo atspi_text_interface__get_selection_method =
{
  -1,  /* ref count */
  (gchar *) "GetSelection",
  (GDBusArgInfo **) atspi_text_interface__get_selection_method_in_args,
  (GDBusArgInfo **) atspi_text_interface__get_selection_method_out_args,
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_text_interface__add_selection_method_in_args__startOffset_arg =
{
  -1,  /* ref count */
  (gchar *) "startOffset",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_text_interface__add_selection_method_in_args__endOffset_arg =
{
  -1,  /* ref count */
  (gchar *) "endOffset",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_text_interface__add_selection_method_in_args[] =
{
  &atspi_text_interface__add_selection_method_in_args__startOffset_arg,
  &atspi_text_interface__add_selection_method_in_args__endOffset_arg,
  NULL,
};

const GDBusArgInfo atspi_text_interface__add_selection_method_out_args__unnamed_arg2_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg2",
  (gchar *) "b",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_text_interface__add_selection_method_out_args[] =
{
  &atspi_text_interface__add_selection_method_out_args__unnamed_arg2_arg,
  NULL,
};

const GDBusMethodInfo atspi_text_interface__add_selection_method =
{
  -1,  /* ref count */
  (gchar *) "AddSelection",
  (GDBusArgInfo **) atspi_text_interface__add_selection_method_in_args,
  (GDBusArgInfo **) atspi_text_interface__add_selection_method_out_args,
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_text_interface__remove_selection_method_in_args__selectionNum_arg =
{
  -1,  /* ref count */
  (gchar *) "selectionNum",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_text_interface__remove_selection_method_in_args[] =
{
  &atspi_text_interface__remove_selection_method_in_args__selectionNum_arg,
  NULL,
};

const GDBusArgInfo atspi_text_interface__remove_selection_method_out_args__unnamed_arg1_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg1",
  (gchar *) "b",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_text_interface__remove_selection_method_out_args[] =
{
  &atspi_text_interface__remove_selection_method_out_args__unnamed_arg1_arg,
  NULL,
};

const GDBusMethodInfo atspi_text_interface__remove_selection_method =
{
  -1,  /* ref count */
  (gchar *) "RemoveSelection",
  (GDBusArgInfo **) atspi_text_interface__remove_selection_method_in_args,
  (GDBusArgInfo **) atspi_text_interface__remove_selection_method_out_args,
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_text_interface__set_selection_method_in_args__selectionNum_arg =
{
  -1,  /* ref count */
  (gchar *) "selectionNum",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_text_interface__set_selection_method_in_args__startOffset_arg =
{
  -1,  /* ref count */
  (gchar *) "startOffset",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_text_interface__set_selection_method_in_args__endOffset_arg =
{
  -1,  /* ref count */
  (gchar *) "endOffset",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_text_interface__set_selection_method_in_args[] =
{
  &atspi_text_interface__set_selection_method_in_args__selectionNum_arg,
  &atspi_text_interface__set_selection_method_in_args__startOffset_arg,
  &atspi_text_interface__set_selection_method_in_args__endOffset_arg,
  NULL,
};

const GDBusArgInfo atspi_text_interface__set_selection_method_out_args__unnamed_arg3_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg3",
  (gchar *) "b",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_text_interface__set_selection_method_out_args[] =
{
  &atspi_text_interface__set_selection_method_out_args__unnamed_arg3_arg,
  NULL,
};

const GDBusMethodInfo atspi_text_interface__set_selection_method =
{
  -1,  /* ref count */
  (gchar *) "SetSelection",
  (GDBusArgInfo **) atspi_text_interface__set_selection_method_in_args,
  (GDBusArgInfo **) atspi_text_interface__set_selection_method_out_args,
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_text_interface__get_range_extents_method_in_args__startOffset_arg =
{
  -1,  /* ref count */
  (gchar *) "startOffset",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_text_interface__get_range_extents_method_in_args__endOffset_arg =
{
  -1,  /* ref count */
  (gchar *) "endOffset",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_text_interface__get_range_extents_method_in_args__coordType_arg =
{
  -1,  /* ref count */
  (gchar *) "coordType",
  (gchar *) "u",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_text_interface__get_range_extents_method_in_args[] =
{
  &atspi_text_interface__get_range_extents_method_in_args__startOffset_arg,
  &atspi_text_interface__get_range_extents_method_in_args__endOffset_arg,
  &atspi_text_interface__get_range_extents_method_in_args__coordType_arg,
  NULL,
};

const GDBusArgInfo atspi_text_interface__get_range_extents_method_out_args__x_arg =
{
  -1,  /* ref count */
  (gchar *) "x",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_text_interface__get_range_extents_method_out_args__y_arg =
{
  -1,  /* ref count */
  (gchar *) "y",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_text_interface__get_range_extents_method_out_args__width_arg =
{
  -1,  /* ref count */
  (gchar *) "width",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_text_interface__get_range_extents_method_out_args__height_arg =
{
  -1,  /* ref count */
  (gchar *) "height",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_text_interface__get_range_extents_method_out_args[] =
{
  &atspi_text_interface__get_range_extents_method_out_args__x_arg,
  &atspi_text_interface__get_range_extents_method_out_args__y_arg,
  &atspi_text_interface__get_range_extents_method_out_args__width_arg,
  &atspi_text_interface__get_range_extents_method_out_args__height_arg,
  NULL,
};

const GDBusMethodInfo atspi_text_interface__get_range_extents_method =
{
  -1,  /* ref count */
  (gchar *) "GetRangeExtents",
  (GDBusArgInfo **) atspi_text_interface__get_range_extents_method_in_args,
  (GDBusArgInfo **) atspi_text_interface__get_range_extents_method_out_args,
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_text_interface__get_bounded_ranges_method_in_args__x_arg =
{
  -1,  /* ref count */
  (gchar *) "x",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_text_interface__get_bounded_ranges_method_in_args__y_arg =
{
  -1,  /* ref count */
  (gchar *) "y",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_text_interface__get_bounded_ranges_method_in_args__width_arg =
{
  -1,  /* ref count */
  (gchar *) "width",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_text_interface__get_bounded_ranges_method_in_args__height_arg =
{
  -1,  /* ref count */
  (gchar *) "height",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_text_interface__get_bounded_ranges_method_in_args__coordType_arg =
{
  -1,  /* ref count */
  (gchar *) "coordType",
  (gchar *) "u",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_text_interface__get_bounded_ranges_method_in_args__xClipType_arg =
{
  -1,  /* ref count */
  (gchar *) "xClipType",
  (gchar *) "u",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_text_interface__get_bounded_ranges_method_in_args__yClipType_arg =
{
  -1,  /* ref count */
  (gchar *) "yClipType",
  (gchar *) "u",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_text_interface__get_bounded_ranges_method_in_args[] =
{
  &atspi_text_interface__get_bounded_ranges_method_in_args__x_arg,
  &atspi_text_interface__get_bounded_ranges_method_in_args__y_arg,
  &atspi_text_interface__get_bounded_ranges_method_in_args__width_arg,
  &atspi_text_interface__get_bounded_ranges_method_in_args__height_arg,
  &atspi_text_interface__get_bounded_ranges_method_in_args__coordType_arg,
  &atspi_text_interface__get_bounded_ranges_method_in_args__xClipType_arg,
  &atspi_text_interface__get_bounded_ranges_method_in_args__yClipType_arg,
  NULL,
};

const GDBusArgInfo atspi_text_interface__get_bounded_ranges_method_out_args__unnamed_arg7_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg7",
  (gchar *) "a(iisv)",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_text_interface__get_bounded_ranges_method_out_args[] =
{
  &atspi_text_interface__get_bounded_ranges_method_out_args__unnamed_arg7_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_text_interface__get_bounded_ranges_method_annotations__org_qtproject__qt_dbus__qt_type_name__out0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.Out0",
  (gchar *) "QSpiRangeList",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_text_interface__get_bounded_ranges_method_annotations[] =
{
  &atspi_text_interface__get_bounded_ranges_method_annotations__org_qtproject__qt_dbus__qt_type_name__out0_annotation,
  NULL,
};

const GDBusMethodInfo atspi_text_interface__get_bounded_ranges_method =
{
  -1,  /* ref count */
  (gchar *) "GetBoundedRanges",
  (GDBusArgInfo **) atspi_text_interface__get_bounded_ranges_method_in_args,
  (GDBusArgInfo **) atspi_text_interface__get_bounded_ranges_method_out_args,
  (GDBusAnnotationInfo **) atspi_text_interface__get_bounded_ranges_method_annotations,
};

const GDBusArgInfo atspi_text_interface__get_attribute_run_method_in_args__offset_arg =
{
  -1,  /* ref count */
  (gchar *) "offset",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_text_interface__get_attribute_run_method_in_args__includeDefaults_arg =
{
  -1,  /* ref count */
  (gchar *) "includeDefaults",
  (gchar *) "b",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_text_interface__get_attribute_run_method_in_args[] =
{
  &atspi_text_interface__get_attribute_run_method_in_args__offset_arg,
  &atspi_text_interface__get_attribute_run_method_in_args__includeDefaults_arg,
  NULL,
};

const GDBusArgInfo atspi_text_interface__get_attribute_run_method_out_args__unnamed_arg2_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg2",
  (gchar *) "a{ss}",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_text_interface__get_attribute_run_method_out_args__startOffset_arg =
{
  -1,  /* ref count */
  (gchar *) "startOffset",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_text_interface__get_attribute_run_method_out_args__endOffset_arg =
{
  -1,  /* ref count */
  (gchar *) "endOffset",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_text_interface__get_attribute_run_method_out_args[] =
{
  &atspi_text_interface__get_attribute_run_method_out_args__unnamed_arg2_arg,
  &atspi_text_interface__get_attribute_run_method_out_args__startOffset_arg,
  &atspi_text_interface__get_attribute_run_method_out_args__endOffset_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_text_interface__get_attribute_run_method_annotations__org_qtproject__qt_dbus__qt_type_name__out0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.Out0",
  (gchar *) "QSpiAttributeSet",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_text_interface__get_attribute_run_method_annotations[] =
{
  &atspi_text_interface__get_attribute_run_method_annotations__org_qtproject__qt_dbus__qt_type_name__out0_annotation,
  NULL,
};

const GDBusMethodInfo atspi_text_interface__get_attribute_run_method =
{
  -1,  /* ref count */
  (gchar *) "GetAttributeRun",
  (GDBusArgInfo **) atspi_text_interface__get_attribute_run_method_in_args,
  (GDBusArgInfo **) atspi_text_interface__get_attribute_run_method_out_args,
  (GDBusAnnotationInfo **) atspi_text_interface__get_attribute_run_method_annotations,
};

const GDBusArgInfo atspi_text_interface__get_default_attribute_set_method_out_args__unnamed_arg0_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg0",
  (gchar *) "a{ss}",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_text_interface__get_default_attribute_set_method_out_args[] =
{
  &atspi_text_interface__get_default_attribute_set_method_out_args__unnamed_arg0_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_text_interface__get_default_attribute_set_method_annotations__org_qtproject__qt_dbus__qt_type_name__out0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.Out0",
  (gchar *) "QSpiAttributeSet",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_text_interface__get_default_attribute_set_method_annotations[] =
{
  &atspi_text_interface__get_default_attribute_set_method_annotations__org_qtproject__qt_dbus__qt_type_name__out0_annotation,
  NULL,
};

const GDBusMethodInfo atspi_text_interface__get_default_attribute_set_method =
{
  -1,  /* ref count */
  (gchar *) "GetDefaultAttributeSet",
  NULL,  /* no in args */
  (GDBusArgInfo **) atspi_text_interface__get_default_attribute_set_method_out_args,
  (GDBusAnnotationInfo **) atspi_text_interface__get_default_attribute_set_method_annotations,
};

const GDBusArgInfo atspi_text_interface__scroll_substring_to_method_in_args__startOffset_arg =
{
  -1,  /* ref count */
  (gchar *) "startOffset",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_text_interface__scroll_substring_to_method_in_args__endOffset_arg =
{
  -1,  /* ref count */
  (gchar *) "endOffset",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_text_interface__scroll_substring_to_method_in_args__type_arg =
{
  -1,  /* ref count */
  (gchar *) "type",
  (gchar *) "u",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_text_interface__scroll_substring_to_method_in_args[] =
{
  &atspi_text_interface__scroll_substring_to_method_in_args__startOffset_arg,
  &atspi_text_interface__scroll_substring_to_method_in_args__endOffset_arg,
  &atspi_text_interface__scroll_substring_to_method_in_args__type_arg,
  NULL,
};

const GDBusMethodInfo atspi_text_interface__scroll_substring_to_method =
{
  -1,  /* ref count */
  (gchar *) "ScrollSubstringTo",
  (GDBusArgInfo **) atspi_text_interface__scroll_substring_to_method_in_args,
  NULL,  /* no out args */
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_text_interface__scroll_substring_to_point_method_in_args__startOffset_arg =
{
  -1,  /* ref count */
  (gchar *) "startOffset",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_text_interface__scroll_substring_to_point_method_in_args__endOffset_arg =
{
  -1,  /* ref count */
  (gchar *) "endOffset",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_text_interface__scroll_substring_to_point_method_in_args__type_arg =
{
  -1,  /* ref count */
  (gchar *) "type",
  (gchar *) "u",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_text_interface__scroll_substring_to_point_method_in_args__x_arg =
{
  -1,  /* ref count */
  (gchar *) "x",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_text_interface__scroll_substring_to_point_method_in_args__y_arg =
{
  -1,  /* ref count */
  (gchar *) "y",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_text_interface__scroll_substring_to_point_method_in_args[] =
{
  &atspi_text_interface__scroll_substring_to_point_method_in_args__startOffset_arg,
  &atspi_text_interface__scroll_substring_to_point_method_in_args__endOffset_arg,
  &atspi_text_interface__scroll_substring_to_point_method_in_args__type_arg,
  &atspi_text_interface__scroll_substring_to_point_method_in_args__x_arg,
  &atspi_text_interface__scroll_substring_to_point_method_in_args__y_arg,
  NULL,
};

const GDBusMethodInfo atspi_text_interface__scroll_substring_to_point_method =
{
  -1,  /* ref count */
  (gchar *) "ScrollSubstringToPoint",
  (GDBusArgInfo **) atspi_text_interface__scroll_substring_to_point_method_in_args,
  NULL,  /* no out args */
  NULL,  /* no annotations */
};

const GDBusMethodInfo * const atspi_text_interface_methods[] =
{
  &atspi_text_interface__get_string_at_offset_method,
  &atspi_text_interface__get_text_method,
  &atspi_text_interface__set_caret_offset_method,
  &atspi_text_interface__get_text_before_offset_method,
  &atspi_text_interface__get_text_at_offset_method,
  &atspi_text_interface__get_text_after_offset_method,
  &atspi_text_interface__get_character_at_offset_method,
  &atspi_text_interface__get_attribute_value_method,
  &atspi_text_interface__get_attributes_method,
  &atspi_text_interface__get_default_attributes_method,
  &atspi_text_interface__get_character_extents_method,
  &atspi_text_interface__get_offset_at_point_method,
  &atspi_text_interface__get_nselections_method,
  &atspi_text_interface__get_selection_method,
  &atspi_text_interface__add_selection_method,
  &atspi_text_interface__remove_selection_method,
  &atspi_text_interface__set_selection_method,
  &atspi_text_interface__get_range_extents_method,
  &atspi_text_interface__get_bounded_ranges_method,
  &atspi_text_interface__get_attribute_run_method,
  &atspi_text_interface__get_default_attribute_set_method,
  &atspi_text_interface__scroll_substring_to_method,
  &atspi_text_interface__scroll_substring_to_point_method,
  NULL,
};

const GDBusPropertyInfo atspi_text_interface__character_count_property =
{
  -1,  /* ref count */
  (gchar *) "CharacterCount",
  (gchar *) "i",
  G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
  NULL,  /* no annotations */
};

const GDBusPropertyInfo atspi_text_interface__caret_offset_property =
{
  -1,  /* ref count */
  (gchar *) "CaretOffset",
  (gchar *) "i",
  G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
  NULL,  /* no annotations */
};

const GDBusPropertyInfo * const atspi_text_interface_properties[] =
{
  &atspi_text_interface__character_count_property,
  &atspi_text_interface__caret_offset_property,
  NULL,
};

const GDBusInterfaceInfo atspi_text_interface =
{
  -1,  /* ref count */
  (gchar *) "org.a11y.atspi.Text",
  (GDBusMethodInfo **) atspi_text_interface_methods,
  NULL,  /* no signals */
  (GDBusPropertyInfo **) atspi_text_interface_properties,
  NULL,  /* no annotations */
};

