/*
 * This file is generated by gdbus-codegen, do not modify it.
 *
 * The license of this code is the same as for the D-Bus interface description
 * it was derived from. Note that it links to GLib, so must comply with the
 * LGPL linking clauses.
 */

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include "atspi-deviceeventcontroller.h"

#include <string.h>

/* ------------------------------------------------------------------------ */
/* Definitions for org.a11y.atspi.DeviceEventController */

const GDBusArgInfo atspi_device_event_controller_interface__register_keystroke_listener_method_in_args__listener_arg =
{
  -1,  /* ref count */
  (gchar *) "listener",
  (gchar *) "o",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_device_event_controller_interface__register_keystroke_listener_method_in_args__keys_arg =
{
  -1,  /* ref count */
  (gchar *) "keys",
  (gchar *) "a(iisi)",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_device_event_controller_interface__register_keystroke_listener_method_in_args__mask_arg =
{
  -1,  /* ref count */
  (gchar *) "mask",
  (gchar *) "u",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_device_event_controller_interface__register_keystroke_listener_method_in_args__type_arg =
{
  -1,  /* ref count */
  (gchar *) "type",
  (gchar *) "au",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_device_event_controller_interface__register_keystroke_listener_method_in_args__mode_arg =
{
  -1,  /* ref count */
  (gchar *) "mode",
  (gchar *) "(bbb)",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_device_event_controller_interface__register_keystroke_listener_method_in_args[] =
{
  &atspi_device_event_controller_interface__register_keystroke_listener_method_in_args__listener_arg,
  &atspi_device_event_controller_interface__register_keystroke_listener_method_in_args__keys_arg,
  &atspi_device_event_controller_interface__register_keystroke_listener_method_in_args__mask_arg,
  &atspi_device_event_controller_interface__register_keystroke_listener_method_in_args__type_arg,
  &atspi_device_event_controller_interface__register_keystroke_listener_method_in_args__mode_arg,
  NULL,
};

const GDBusArgInfo atspi_device_event_controller_interface__register_keystroke_listener_method_out_args__unnamed_arg5_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg5",
  (gchar *) "b",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_device_event_controller_interface__register_keystroke_listener_method_out_args[] =
{
  &atspi_device_event_controller_interface__register_keystroke_listener_method_out_args__unnamed_arg5_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_device_event_controller_interface__register_keystroke_listener_method_annotations__org_qtproject__qt_dbus__qt_type_name__in1_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.In1",
  (gchar *) "QSpiKeyTypeArray",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo atspi_device_event_controller_interface__register_keystroke_listener_method_annotations__org_qtproject__qt_dbus__qt_type_name__in3_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.In3",
  (gchar *) "QSpiEventTypeArray",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo atspi_device_event_controller_interface__register_keystroke_listener_method_annotations__org_qtproject__qt_dbus__qt_type_name__in4_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.In4",
  (gchar *) "QSpiEventMode",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_device_event_controller_interface__register_keystroke_listener_method_annotations[] =
{
  &atspi_device_event_controller_interface__register_keystroke_listener_method_annotations__org_qtproject__qt_dbus__qt_type_name__in1_annotation,
  &atspi_device_event_controller_interface__register_keystroke_listener_method_annotations__org_qtproject__qt_dbus__qt_type_name__in3_annotation,
  &atspi_device_event_controller_interface__register_keystroke_listener_method_annotations__org_qtproject__qt_dbus__qt_type_name__in4_annotation,
  NULL,
};

const GDBusMethodInfo atspi_device_event_controller_interface__register_keystroke_listener_method =
{
  -1,  /* ref count */
  (gchar *) "RegisterKeystrokeListener",
  (GDBusArgInfo **) atspi_device_event_controller_interface__register_keystroke_listener_method_in_args,
  (GDBusArgInfo **) atspi_device_event_controller_interface__register_keystroke_listener_method_out_args,
  (GDBusAnnotationInfo **) atspi_device_event_controller_interface__register_keystroke_listener_method_annotations,
};

const GDBusArgInfo atspi_device_event_controller_interface__deregister_keystroke_listener_method_in_args__listener_arg =
{
  -1,  /* ref count */
  (gchar *) "listener",
  (gchar *) "o",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_device_event_controller_interface__deregister_keystroke_listener_method_in_args__keys_arg =
{
  -1,  /* ref count */
  (gchar *) "keys",
  (gchar *) "a(iisi)",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_device_event_controller_interface__deregister_keystroke_listener_method_in_args__mask_arg =
{
  -1,  /* ref count */
  (gchar *) "mask",
  (gchar *) "u",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_device_event_controller_interface__deregister_keystroke_listener_method_in_args__type_arg =
{
  -1,  /* ref count */
  (gchar *) "type",
  (gchar *) "u",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_device_event_controller_interface__deregister_keystroke_listener_method_in_args[] =
{
  &atspi_device_event_controller_interface__deregister_keystroke_listener_method_in_args__listener_arg,
  &atspi_device_event_controller_interface__deregister_keystroke_listener_method_in_args__keys_arg,
  &atspi_device_event_controller_interface__deregister_keystroke_listener_method_in_args__mask_arg,
  &atspi_device_event_controller_interface__deregister_keystroke_listener_method_in_args__type_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_device_event_controller_interface__deregister_keystroke_listener_method_annotations__org_qtproject__qt_dbus__qt_type_name__in1_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.In1",
  (gchar *) "QSpiKeyTypeArray",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_device_event_controller_interface__deregister_keystroke_listener_method_annotations[] =
{
  &atspi_device_event_controller_interface__deregister_keystroke_listener_method_annotations__org_qtproject__qt_dbus__qt_type_name__in1_annotation,
  NULL,
};

const GDBusMethodInfo atspi_device_event_controller_interface__deregister_keystroke_listener_method =
{
  -1,  /* ref count */
  (gchar *) "DeregisterKeystrokeListener",
  (GDBusArgInfo **) atspi_device_event_controller_interface__deregister_keystroke_listener_method_in_args,
  NULL,  /* no out args */
  (GDBusAnnotationInfo **) atspi_device_event_controller_interface__deregister_keystroke_listener_method_annotations,
};

const GDBusArgInfo atspi_device_event_controller_interface__register_device_event_listener_method_in_args__listener_arg =
{
  -1,  /* ref count */
  (gchar *) "listener",
  (gchar *) "o",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_device_event_controller_interface__register_device_event_listener_method_in_args__types_arg =
{
  -1,  /* ref count */
  (gchar *) "types",
  (gchar *) "u",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_device_event_controller_interface__register_device_event_listener_method_in_args[] =
{
  &atspi_device_event_controller_interface__register_device_event_listener_method_in_args__listener_arg,
  &atspi_device_event_controller_interface__register_device_event_listener_method_in_args__types_arg,
  NULL,
};

const GDBusArgInfo atspi_device_event_controller_interface__register_device_event_listener_method_out_args__unnamed_arg2_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg2",
  (gchar *) "b",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_device_event_controller_interface__register_device_event_listener_method_out_args[] =
{
  &atspi_device_event_controller_interface__register_device_event_listener_method_out_args__unnamed_arg2_arg,
  NULL,
};

const GDBusMethodInfo atspi_device_event_controller_interface__register_device_event_listener_method =
{
  -1,  /* ref count */
  (gchar *) "RegisterDeviceEventListener",
  (GDBusArgInfo **) atspi_device_event_controller_interface__register_device_event_listener_method_in_args,
  (GDBusArgInfo **) atspi_device_event_controller_interface__register_device_event_listener_method_out_args,
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_device_event_controller_interface__deregister_device_event_listener_method_in_args__listener_arg =
{
  -1,  /* ref count */
  (gchar *) "listener",
  (gchar *) "o",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_device_event_controller_interface__deregister_device_event_listener_method_in_args__types_arg =
{
  -1,  /* ref count */
  (gchar *) "types",
  (gchar *) "u",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_device_event_controller_interface__deregister_device_event_listener_method_in_args[] =
{
  &atspi_device_event_controller_interface__deregister_device_event_listener_method_in_args__listener_arg,
  &atspi_device_event_controller_interface__deregister_device_event_listener_method_in_args__types_arg,
  NULL,
};

const GDBusMethodInfo atspi_device_event_controller_interface__deregister_device_event_listener_method =
{
  -1,  /* ref count */
  (gchar *) "DeregisterDeviceEventListener",
  (GDBusArgInfo **) atspi_device_event_controller_interface__deregister_device_event_listener_method_in_args,
  NULL,  /* no out args */
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_device_event_controller_interface__generate_keyboard_event_method_in_args__keycode_arg =
{
  -1,  /* ref count */
  (gchar *) "keycode",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_device_event_controller_interface__generate_keyboard_event_method_in_args__keystring_arg =
{
  -1,  /* ref count */
  (gchar *) "keystring",
  (gchar *) "s",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_device_event_controller_interface__generate_keyboard_event_method_in_args__type_arg =
{
  -1,  /* ref count */
  (gchar *) "type",
  (gchar *) "u",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_device_event_controller_interface__generate_keyboard_event_method_in_args[] =
{
  &atspi_device_event_controller_interface__generate_keyboard_event_method_in_args__keycode_arg,
  &atspi_device_event_controller_interface__generate_keyboard_event_method_in_args__keystring_arg,
  &atspi_device_event_controller_interface__generate_keyboard_event_method_in_args__type_arg,
  NULL,
};

const GDBusMethodInfo atspi_device_event_controller_interface__generate_keyboard_event_method =
{
  -1,  /* ref count */
  (gchar *) "GenerateKeyboardEvent",
  (GDBusArgInfo **) atspi_device_event_controller_interface__generate_keyboard_event_method_in_args,
  NULL,  /* no out args */
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_device_event_controller_interface__generate_mouse_event_method_in_args__x_arg =
{
  -1,  /* ref count */
  (gchar *) "x",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_device_event_controller_interface__generate_mouse_event_method_in_args__y_arg =
{
  -1,  /* ref count */
  (gchar *) "y",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_device_event_controller_interface__generate_mouse_event_method_in_args__eventName_arg =
{
  -1,  /* ref count */
  (gchar *) "eventName",
  (gchar *) "s",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_device_event_controller_interface__generate_mouse_event_method_in_args[] =
{
  &atspi_device_event_controller_interface__generate_mouse_event_method_in_args__x_arg,
  &atspi_device_event_controller_interface__generate_mouse_event_method_in_args__y_arg,
  &atspi_device_event_controller_interface__generate_mouse_event_method_in_args__eventName_arg,
  NULL,
};

const GDBusMethodInfo atspi_device_event_controller_interface__generate_mouse_event_method =
{
  -1,  /* ref count */
  (gchar *) "GenerateMouseEvent",
  (GDBusArgInfo **) atspi_device_event_controller_interface__generate_mouse_event_method_in_args,
  NULL,  /* no out args */
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_device_event_controller_interface__notify_listeners_sync_method_in_args__event_arg =
{
  -1,  /* ref count */
  (gchar *) "event",
  (gchar *) "(uiuuisb)",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_device_event_controller_interface__notify_listeners_sync_method_in_args[] =
{
  &atspi_device_event_controller_interface__notify_listeners_sync_method_in_args__event_arg,
  NULL,
};

const GDBusArgInfo atspi_device_event_controller_interface__notify_listeners_sync_method_out_args__unnamed_arg1_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg1",
  (gchar *) "b",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_device_event_controller_interface__notify_listeners_sync_method_out_args[] =
{
  &atspi_device_event_controller_interface__notify_listeners_sync_method_out_args__unnamed_arg1_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_device_event_controller_interface__notify_listeners_sync_method_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.In0",
  (gchar *) "QSpiDeviceEvent",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_device_event_controller_interface__notify_listeners_sync_method_annotations[] =
{
  &atspi_device_event_controller_interface__notify_listeners_sync_method_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation,
  NULL,
};

const GDBusMethodInfo atspi_device_event_controller_interface__notify_listeners_sync_method =
{
  -1,  /* ref count */
  (gchar *) "NotifyListenersSync",
  (GDBusArgInfo **) atspi_device_event_controller_interface__notify_listeners_sync_method_in_args,
  (GDBusArgInfo **) atspi_device_event_controller_interface__notify_listeners_sync_method_out_args,
  (GDBusAnnotationInfo **) atspi_device_event_controller_interface__notify_listeners_sync_method_annotations,
};

const GDBusArgInfo atspi_device_event_controller_interface__notify_listeners_async_method_in_args__event_arg =
{
  -1,  /* ref count */
  (gchar *) "event",
  (gchar *) "(uiuuisb)",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_device_event_controller_interface__notify_listeners_async_method_in_args[] =
{
  &atspi_device_event_controller_interface__notify_listeners_async_method_in_args__event_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_device_event_controller_interface__notify_listeners_async_method_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.In0",
  (gchar *) "QSpiDeviceEvent",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_device_event_controller_interface__notify_listeners_async_method_annotations[] =
{
  &atspi_device_event_controller_interface__notify_listeners_async_method_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation,
  NULL,
};

const GDBusMethodInfo atspi_device_event_controller_interface__notify_listeners_async_method =
{
  -1,  /* ref count */
  (gchar *) "NotifyListenersAsync",
  (GDBusArgInfo **) atspi_device_event_controller_interface__notify_listeners_async_method_in_args,
  NULL,  /* no out args */
  (GDBusAnnotationInfo **) atspi_device_event_controller_interface__notify_listeners_async_method_annotations,
};

const GDBusMethodInfo * const atspi_device_event_controller_interface_methods[] =
{
  &atspi_device_event_controller_interface__register_keystroke_listener_method,
  &atspi_device_event_controller_interface__deregister_keystroke_listener_method,
  &atspi_device_event_controller_interface__register_device_event_listener_method,
  &atspi_device_event_controller_interface__deregister_device_event_listener_method,
  &atspi_device_event_controller_interface__generate_keyboard_event_method,
  &atspi_device_event_controller_interface__generate_mouse_event_method,
  &atspi_device_event_controller_interface__notify_listeners_sync_method,
  &atspi_device_event_controller_interface__notify_listeners_async_method,
  NULL,
};

const GDBusInterfaceInfo atspi_device_event_controller_interface =
{
  -1,  /* ref count */
  (gchar *) "org.a11y.atspi.DeviceEventController",
  (GDBusMethodInfo **) atspi_device_event_controller_interface_methods,
  NULL,  /* no signals */
NULL,  /* no properties */
  NULL,  /* no annotations */
};

