
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#include "config.h"
#include "gskenumtypes.h"
#include <gsk.h>
/* enumerations from "gskenums.h" */
GType
gsk_render_node_type_get_type (void)
{
  static gsize g_define_type_id__volatile = 0;

  if (g_once_init_enter (&g_define_type_id__volatile))
    {
      static const GEnumValue values[] = {
        { GSK_NOT_A_RENDER_NODE, "GSK_NOT_A_RENDER_NODE", "not-a-render-node" },
        { GSK_CONTAINER_NODE, "GSK_CONTAINER_NODE", "container-node" },
        { GSK_CAIRO_NODE, "GSK_CAIRO_NODE", "cairo-node" },
        { GSK_COLOR_NODE, "GSK_COLOR_NODE", "color-node" },
        { GSK_LINEAR_GRADIENT_NODE, "GSK_LINEAR_GRADIENT_NODE", "linear-gradient-node" },
        { GSK_REPEATING_LINEAR_GRADIENT_NODE, "GSK_REPEATING_LINEAR_GRADIENT_NODE", "repeating-linear-gradient-node" },
        { GSK_RADIAL_GRADIENT_NODE, "GSK_RADIAL_GRADIENT_NODE", "radial-gradient-node" },
        { GSK_REPEATING_RADIAL_GRADIENT_NODE, "GSK_REPEATING_RADIAL_GRADIENT_NODE", "repeating-radial-gradient-node" },
        { GSK_CONIC_GRADIENT_NODE, "GSK_CONIC_GRADIENT_NODE", "conic-gradient-node" },
        { GSK_BORDER_NODE, "GSK_BORDER_NODE", "border-node" },
        { GSK_TEXTURE_NODE, "GSK_TEXTURE_NODE", "texture-node" },
        { GSK_INSET_SHADOW_NODE, "GSK_INSET_SHADOW_NODE", "inset-shadow-node" },
        { GSK_OUTSET_SHADOW_NODE, "GSK_OUTSET_SHADOW_NODE", "outset-shadow-node" },
        { GSK_TRANSFORM_NODE, "GSK_TRANSFORM_NODE", "transform-node" },
        { GSK_OPACITY_NODE, "GSK_OPACITY_NODE", "opacity-node" },
        { GSK_COLOR_MATRIX_NODE, "GSK_COLOR_MATRIX_NODE", "color-matrix-node" },
        { GSK_REPEAT_NODE, "GSK_REPEAT_NODE", "repeat-node" },
        { GSK_CLIP_NODE, "GSK_CLIP_NODE", "clip-node" },
        { GSK_ROUNDED_CLIP_NODE, "GSK_ROUNDED_CLIP_NODE", "rounded-clip-node" },
        { GSK_SHADOW_NODE, "GSK_SHADOW_NODE", "shadow-node" },
        { GSK_BLEND_NODE, "GSK_BLEND_NODE", "blend-node" },
        { GSK_CROSS_FADE_NODE, "GSK_CROSS_FADE_NODE", "cross-fade-node" },
        { GSK_TEXT_NODE, "GSK_TEXT_NODE", "text-node" },
        { GSK_BLUR_NODE, "GSK_BLUR_NODE", "blur-node" },
        { GSK_DEBUG_NODE, "GSK_DEBUG_NODE", "debug-node" },
        { GSK_GL_SHADER_NODE, "GSK_GL_SHADER_NODE", "gl-shader-node" },
        { GSK_TEXTURE_SCALE_NODE, "GSK_TEXTURE_SCALE_NODE", "texture-scale-node" },
        { GSK_MASK_NODE, "GSK_MASK_NODE", "mask-node" },
        { GSK_FILL_NODE, "GSK_FILL_NODE", "fill-node" },
        { GSK_STROKE_NODE, "GSK_STROKE_NODE", "stroke-node" },
        { GSK_SUBSURFACE_NODE, "GSK_SUBSURFACE_NODE", "subsurface-node" },
        { 0, NULL, NULL }
      };
      GType g_define_type_id =
        g_enum_register_static (g_intern_static_string ("GskRenderNodeType"), values);
      g_once_init_leave (&g_define_type_id__volatile, g_define_type_id);
    }

  return g_define_type_id__volatile;
}
GType
gsk_scaling_filter_get_type (void)
{
  static gsize g_define_type_id__volatile = 0;

  if (g_once_init_enter (&g_define_type_id__volatile))
    {
      static const GEnumValue values[] = {
        { GSK_SCALING_FILTER_LINEAR, "GSK_SCALING_FILTER_LINEAR", "linear" },
        { GSK_SCALING_FILTER_NEAREST, "GSK_SCALING_FILTER_NEAREST", "nearest" },
        { GSK_SCALING_FILTER_TRILINEAR, "GSK_SCALING_FILTER_TRILINEAR", "trilinear" },
        { 0, NULL, NULL }
      };
      GType g_define_type_id =
        g_enum_register_static (g_intern_static_string ("GskScalingFilter"), values);
      g_once_init_leave (&g_define_type_id__volatile, g_define_type_id);
    }

  return g_define_type_id__volatile;
}
GType
gsk_blend_mode_get_type (void)
{
  static gsize g_define_type_id__volatile = 0;

  if (g_once_init_enter (&g_define_type_id__volatile))
    {
      static const GEnumValue values[] = {
        { GSK_BLEND_MODE_DEFAULT, "GSK_BLEND_MODE_DEFAULT", "default" },
        { GSK_BLEND_MODE_MULTIPLY, "GSK_BLEND_MODE_MULTIPLY", "multiply" },
        { GSK_BLEND_MODE_SCREEN, "GSK_BLEND_MODE_SCREEN", "screen" },
        { GSK_BLEND_MODE_OVERLAY, "GSK_BLEND_MODE_OVERLAY", "overlay" },
        { GSK_BLEND_MODE_DARKEN, "GSK_BLEND_MODE_DARKEN", "darken" },
        { GSK_BLEND_MODE_LIGHTEN, "GSK_BLEND_MODE_LIGHTEN", "lighten" },
        { GSK_BLEND_MODE_COLOR_DODGE, "GSK_BLEND_MODE_COLOR_DODGE", "color-dodge" },
        { GSK_BLEND_MODE_COLOR_BURN, "GSK_BLEND_MODE_COLOR_BURN", "color-burn" },
        { GSK_BLEND_MODE_HARD_LIGHT, "GSK_BLEND_MODE_HARD_LIGHT", "hard-light" },
        { GSK_BLEND_MODE_SOFT_LIGHT, "GSK_BLEND_MODE_SOFT_LIGHT", "soft-light" },
        { GSK_BLEND_MODE_DIFFERENCE, "GSK_BLEND_MODE_DIFFERENCE", "difference" },
        { GSK_BLEND_MODE_EXCLUSION, "GSK_BLEND_MODE_EXCLUSION", "exclusion" },
        { GSK_BLEND_MODE_COLOR, "GSK_BLEND_MODE_COLOR", "color" },
        { GSK_BLEND_MODE_HUE, "GSK_BLEND_MODE_HUE", "hue" },
        { GSK_BLEND_MODE_SATURATION, "GSK_BLEND_MODE_SATURATION", "saturation" },
        { GSK_BLEND_MODE_LUMINOSITY, "GSK_BLEND_MODE_LUMINOSITY", "luminosity" },
        { 0, NULL, NULL }
      };
      GType g_define_type_id =
        g_enum_register_static (g_intern_static_string ("GskBlendMode"), values);
      g_once_init_leave (&g_define_type_id__volatile, g_define_type_id);
    }

  return g_define_type_id__volatile;
}
GType
gsk_corner_get_type (void)
{
  static gsize g_define_type_id__volatile = 0;

  if (g_once_init_enter (&g_define_type_id__volatile))
    {
      static const GEnumValue values[] = {
        { GSK_CORNER_TOP_LEFT, "GSK_CORNER_TOP_LEFT", "top-left" },
        { GSK_CORNER_TOP_RIGHT, "GSK_CORNER_TOP_RIGHT", "top-right" },
        { GSK_CORNER_BOTTOM_RIGHT, "GSK_CORNER_BOTTOM_RIGHT", "bottom-right" },
        { GSK_CORNER_BOTTOM_LEFT, "GSK_CORNER_BOTTOM_LEFT", "bottom-left" },
        { 0, NULL, NULL }
      };
      GType g_define_type_id =
        g_enum_register_static (g_intern_static_string ("GskCorner"), values);
      g_once_init_leave (&g_define_type_id__volatile, g_define_type_id);
    }

  return g_define_type_id__volatile;
}
GType
gsk_fill_rule_get_type (void)
{
  static gsize g_define_type_id__volatile = 0;

  if (g_once_init_enter (&g_define_type_id__volatile))
    {
      static const GEnumValue values[] = {
        { GSK_FILL_RULE_WINDING, "GSK_FILL_RULE_WINDING", "winding" },
        { GSK_FILL_RULE_EVEN_ODD, "GSK_FILL_RULE_EVEN_ODD", "even-odd" },
        { 0, NULL, NULL }
      };
      GType g_define_type_id =
        g_enum_register_static (g_intern_static_string ("GskFillRule"), values);
      g_once_init_leave (&g_define_type_id__volatile, g_define_type_id);
    }

  return g_define_type_id__volatile;
}
GType
gsk_line_cap_get_type (void)
{
  static gsize g_define_type_id__volatile = 0;

  if (g_once_init_enter (&g_define_type_id__volatile))
    {
      static const GEnumValue values[] = {
        { GSK_LINE_CAP_BUTT, "GSK_LINE_CAP_BUTT", "butt" },
        { GSK_LINE_CAP_ROUND, "GSK_LINE_CAP_ROUND", "round" },
        { GSK_LINE_CAP_SQUARE, "GSK_LINE_CAP_SQUARE", "square" },
        { 0, NULL, NULL }
      };
      GType g_define_type_id =
        g_enum_register_static (g_intern_static_string ("GskLineCap"), values);
      g_once_init_leave (&g_define_type_id__volatile, g_define_type_id);
    }

  return g_define_type_id__volatile;
}
GType
gsk_line_join_get_type (void)
{
  static gsize g_define_type_id__volatile = 0;

  if (g_once_init_enter (&g_define_type_id__volatile))
    {
      static const GEnumValue values[] = {
        { GSK_LINE_JOIN_MITER, "GSK_LINE_JOIN_MITER", "miter" },
        { GSK_LINE_JOIN_ROUND, "GSK_LINE_JOIN_ROUND", "round" },
        { GSK_LINE_JOIN_BEVEL, "GSK_LINE_JOIN_BEVEL", "bevel" },
        { 0, NULL, NULL }
      };
      GType g_define_type_id =
        g_enum_register_static (g_intern_static_string ("GskLineJoin"), values);
      g_once_init_leave (&g_define_type_id__volatile, g_define_type_id);
    }

  return g_define_type_id__volatile;
}
GType
gsk_path_operation_get_type (void)
{
  static gsize g_define_type_id__volatile = 0;

  if (g_once_init_enter (&g_define_type_id__volatile))
    {
      static const GEnumValue values[] = {
        { GSK_PATH_MOVE, "GSK_PATH_MOVE", "move" },
        { GSK_PATH_CLOSE, "GSK_PATH_CLOSE", "close" },
        { GSK_PATH_LINE, "GSK_PATH_LINE", "line" },
        { GSK_PATH_QUAD, "GSK_PATH_QUAD", "quad" },
        { GSK_PATH_CUBIC, "GSK_PATH_CUBIC", "cubic" },
        { GSK_PATH_CONIC, "GSK_PATH_CONIC", "conic" },
        { 0, NULL, NULL }
      };
      GType g_define_type_id =
        g_enum_register_static (g_intern_static_string ("GskPathOperation"), values);
      g_once_init_leave (&g_define_type_id__volatile, g_define_type_id);
    }

  return g_define_type_id__volatile;
}
GType
gsk_path_direction_get_type (void)
{
  static gsize g_define_type_id__volatile = 0;

  if (g_once_init_enter (&g_define_type_id__volatile))
    {
      static const GEnumValue values[] = {
        { GSK_PATH_FROM_START, "GSK_PATH_FROM_START", "from-start" },
        { GSK_PATH_TO_START, "GSK_PATH_TO_START", "to-start" },
        { GSK_PATH_TO_END, "GSK_PATH_TO_END", "to-end" },
        { GSK_PATH_FROM_END, "GSK_PATH_FROM_END", "from-end" },
        { 0, NULL, NULL }
      };
      GType g_define_type_id =
        g_enum_register_static (g_intern_static_string ("GskPathDirection"), values);
      g_once_init_leave (&g_define_type_id__volatile, g_define_type_id);
    }

  return g_define_type_id__volatile;
}
GType
gsk_serialization_error_get_type (void)
{
  static gsize g_define_type_id__volatile = 0;

  if (g_once_init_enter (&g_define_type_id__volatile))
    {
      static const GEnumValue values[] = {
        { GSK_SERIALIZATION_UNSUPPORTED_FORMAT, "GSK_SERIALIZATION_UNSUPPORTED_FORMAT", "unsupported-format" },
        { GSK_SERIALIZATION_UNSUPPORTED_VERSION, "GSK_SERIALIZATION_UNSUPPORTED_VERSION", "unsupported-version" },
        { GSK_SERIALIZATION_INVALID_DATA, "GSK_SERIALIZATION_INVALID_DATA", "invalid-data" },
        { 0, NULL, NULL }
      };
      GType g_define_type_id =
        g_enum_register_static (g_intern_static_string ("GskSerializationError"), values);
      g_once_init_leave (&g_define_type_id__volatile, g_define_type_id);
    }

  return g_define_type_id__volatile;
}
GType
gsk_transform_category_get_type (void)
{
  static gsize g_define_type_id__volatile = 0;

  if (g_once_init_enter (&g_define_type_id__volatile))
    {
      static const GEnumValue values[] = {
        { GSK_TRANSFORM_CATEGORY_UNKNOWN, "GSK_TRANSFORM_CATEGORY_UNKNOWN", "unknown" },
        { GSK_TRANSFORM_CATEGORY_ANY, "GSK_TRANSFORM_CATEGORY_ANY", "any" },
        { GSK_TRANSFORM_CATEGORY_3D, "GSK_TRANSFORM_CATEGORY_3D", "3d" },
        { GSK_TRANSFORM_CATEGORY_2D, "GSK_TRANSFORM_CATEGORY_2D", "2d" },
        { GSK_TRANSFORM_CATEGORY_2D_AFFINE, "GSK_TRANSFORM_CATEGORY_2D_AFFINE", "2d-affine" },
        { GSK_TRANSFORM_CATEGORY_2D_TRANSLATE, "GSK_TRANSFORM_CATEGORY_2D_TRANSLATE", "2d-translate" },
        { GSK_TRANSFORM_CATEGORY_IDENTITY, "GSK_TRANSFORM_CATEGORY_IDENTITY", "identity" },
        { 0, NULL, NULL }
      };
      GType g_define_type_id =
        g_enum_register_static (g_intern_static_string ("GskTransformCategory"), values);
      g_once_init_leave (&g_define_type_id__volatile, g_define_type_id);
    }

  return g_define_type_id__volatile;
}
GType
gsk_gl_uniform_type_get_type (void)
{
  static gsize g_define_type_id__volatile = 0;

  if (g_once_init_enter (&g_define_type_id__volatile))
    {
      static const GEnumValue values[] = {
        { GSK_GL_UNIFORM_TYPE_NONE, "GSK_GL_UNIFORM_TYPE_NONE", "none" },
        { GSK_GL_UNIFORM_TYPE_FLOAT, "GSK_GL_UNIFORM_TYPE_FLOAT", "float" },
        { GSK_GL_UNIFORM_TYPE_INT, "GSK_GL_UNIFORM_TYPE_INT", "int" },
        { GSK_GL_UNIFORM_TYPE_UINT, "GSK_GL_UNIFORM_TYPE_UINT", "uint" },
        { GSK_GL_UNIFORM_TYPE_BOOL, "GSK_GL_UNIFORM_TYPE_BOOL", "bool" },
        { GSK_GL_UNIFORM_TYPE_VEC2, "GSK_GL_UNIFORM_TYPE_VEC2", "vec2" },
        { GSK_GL_UNIFORM_TYPE_VEC3, "GSK_GL_UNIFORM_TYPE_VEC3", "vec3" },
        { GSK_GL_UNIFORM_TYPE_VEC4, "GSK_GL_UNIFORM_TYPE_VEC4", "vec4" },
        { 0, NULL, NULL }
      };
      GType g_define_type_id =
        g_enum_register_static (g_intern_static_string ("GskGLUniformType"), values);
      g_once_init_leave (&g_define_type_id__volatile, g_define_type_id);
    }

  return g_define_type_id__volatile;
}
GType
gsk_mask_mode_get_type (void)
{
  static gsize g_define_type_id__volatile = 0;

  if (g_once_init_enter (&g_define_type_id__volatile))
    {
      static const GEnumValue values[] = {
        { GSK_MASK_MODE_ALPHA, "GSK_MASK_MODE_ALPHA", "alpha" },
        { GSK_MASK_MODE_INVERTED_ALPHA, "GSK_MASK_MODE_INVERTED_ALPHA", "inverted-alpha" },
        { GSK_MASK_MODE_LUMINANCE, "GSK_MASK_MODE_LUMINANCE", "luminance" },
        { GSK_MASK_MODE_INVERTED_LUMINANCE, "GSK_MASK_MODE_INVERTED_LUMINANCE", "inverted-luminance" },
        { 0, NULL, NULL }
      };
      GType g_define_type_id =
        g_enum_register_static (g_intern_static_string ("GskMaskMode"), values);
      g_once_init_leave (&g_define_type_id__volatile, g_define_type_id);
    }

  return g_define_type_id__volatile;
}
/* enumerations from "gskpath.h" */
GType
gsk_path_foreach_flags_get_type (void)
{
  static gsize g_define_type_id__volatile = 0;

  if (g_once_init_enter (&g_define_type_id__volatile))
    {
      static const GFlagsValue values[] = {
        { GSK_PATH_FOREACH_ALLOW_ONLY_LINES, "GSK_PATH_FOREACH_ALLOW_ONLY_LINES", "only-lines" },
        { GSK_PATH_FOREACH_ALLOW_QUAD, "GSK_PATH_FOREACH_ALLOW_QUAD", "quad" },
        { GSK_PATH_FOREACH_ALLOW_CUBIC, "GSK_PATH_FOREACH_ALLOW_CUBIC", "cubic" },
        { GSK_PATH_FOREACH_ALLOW_CONIC, "GSK_PATH_FOREACH_ALLOW_CONIC", "conic" },
        { 0, NULL, NULL }
      };
      GType g_define_type_id =
        g_flags_register_static (g_intern_static_string ("GskPathForeachFlags"), values);
      g_once_init_leave (&g_define_type_id__volatile, g_define_type_id);
    }

  return g_define_type_id__volatile;
}

/* Generated data ends here */

