/*
 * This file is generated by gdbus-codegen, do not modify it.
 *
 * The license of this code is the same as for the D-Bus interface description
 * it was derived from. Note that it links to GLib, so must comply with the
 * LGPL linking clauses.
 */

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include "atspi-registry.h"

#include <string.h>

/* ------------------------------------------------------------------------ */
/* Definitions for org.a11y.atspi.Registry */

const GDBusArgInfo atspi_registry_interface__register_event_method_in_args__event_arg =
{
  -1,  /* ref count */
  (gchar *) "event",
  (gchar *) "s",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_registry_interface__register_event_method_in_args[] =
{
  &atspi_registry_interface__register_event_method_in_args__event_arg,
  NULL,
};

const GDBusMethodInfo atspi_registry_interface__register_event_method =
{
  -1,  /* ref count */
  (gchar *) "RegisterEvent",
  (GDBusArgInfo **) atspi_registry_interface__register_event_method_in_args,
  NULL,  /* no out args */
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_registry_interface__deregister_event_method_in_args__event_arg =
{
  -1,  /* ref count */
  (gchar *) "event",
  (gchar *) "s",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_registry_interface__deregister_event_method_in_args[] =
{
  &atspi_registry_interface__deregister_event_method_in_args__event_arg,
  NULL,
};

const GDBusMethodInfo atspi_registry_interface__deregister_event_method =
{
  -1,  /* ref count */
  (gchar *) "DeregisterEvent",
  (GDBusArgInfo **) atspi_registry_interface__deregister_event_method_in_args,
  NULL,  /* no out args */
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_registry_interface__get_registered_events_method_out_args__events_arg =
{
  -1,  /* ref count */
  (gchar *) "events",
  (gchar *) "a(ss)",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_registry_interface__get_registered_events_method_out_args[] =
{
  &atspi_registry_interface__get_registered_events_method_out_args__events_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_registry_interface__get_registered_events_method_annotations__org_qtproject__qt_dbus__qt_type_name__out0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.Out0",
  (gchar *) "QSpiEventListenerArray",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_registry_interface__get_registered_events_method_annotations[] =
{
  &atspi_registry_interface__get_registered_events_method_annotations__org_qtproject__qt_dbus__qt_type_name__out0_annotation,
  NULL,
};

const GDBusMethodInfo atspi_registry_interface__get_registered_events_method =
{
  -1,  /* ref count */
  (gchar *) "GetRegisteredEvents",
  NULL,  /* no in args */
  (GDBusArgInfo **) atspi_registry_interface__get_registered_events_method_out_args,
  (GDBusAnnotationInfo **) atspi_registry_interface__get_registered_events_method_annotations,
};

const GDBusMethodInfo * const atspi_registry_interface_methods[] =
{
  &atspi_registry_interface__register_event_method,
  &atspi_registry_interface__deregister_event_method,
  &atspi_registry_interface__get_registered_events_method,
  NULL,
};

const GDBusArgInfo atspi_registry_interface__event_listener_registered_signal_args__bus_arg =
{
  -1,  /* ref count */
  (gchar *) "bus",
  (gchar *) "s",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_registry_interface__event_listener_registered_signal_args__path_arg =
{
  -1,  /* ref count */
  (gchar *) "path",
  (gchar *) "s",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_registry_interface__event_listener_registered_signal_args[] =
{
  &atspi_registry_interface__event_listener_registered_signal_args__bus_arg,
  &atspi_registry_interface__event_listener_registered_signal_args__path_arg,
  NULL,
};

const GDBusSignalInfo atspi_registry_interface__event_listener_registered_signal =
{
  -1,  /* ref count */
  (gchar *) "EventListenerRegistered",
  (GDBusArgInfo **) atspi_registry_interface__event_listener_registered_signal_args,
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_registry_interface__event_listener_deregistered_signal_args__bus_arg =
{
  -1,  /* ref count */
  (gchar *) "bus",
  (gchar *) "s",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_registry_interface__event_listener_deregistered_signal_args__path_arg =
{
  -1,  /* ref count */
  (gchar *) "path",
  (gchar *) "s",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_registry_interface__event_listener_deregistered_signal_args[] =
{
  &atspi_registry_interface__event_listener_deregistered_signal_args__bus_arg,
  &atspi_registry_interface__event_listener_deregistered_signal_args__path_arg,
  NULL,
};

const GDBusSignalInfo atspi_registry_interface__event_listener_deregistered_signal =
{
  -1,  /* ref count */
  (gchar *) "EventListenerDeregistered",
  (GDBusArgInfo **) atspi_registry_interface__event_listener_deregistered_signal_args,
  NULL,  /* no annotations */
};

const GDBusSignalInfo * const atspi_registry_interface_signals[] =
{
  &atspi_registry_interface__event_listener_registered_signal,
  &atspi_registry_interface__event_listener_deregistered_signal,
  NULL,
};

const GDBusInterfaceInfo atspi_registry_interface =
{
  -1,  /* ref count */
  (gchar *) "org.a11y.atspi.Registry",
  (GDBusMethodInfo **) atspi_registry_interface_methods,
  (GDBusSignalInfo **) atspi_registry_interface_signals,
NULL,  /* no properties */
  NULL,  /* no annotations */
};

