/*
 * This file is generated by gdbus-codegen, do not modify it.
 *
 * The license of this code is the same as for the D-Bus interface description
 * it was derived from. Note that it links to GLib, so must comply with the
 * LGPL linking clauses.
 */

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include "atspi-editabletext.h"

#include <string.h>

/* ------------------------------------------------------------------------ */
/* Definitions for org.a11y.atspi.EditableText */

const GDBusArgInfo atspi_editable_text_interface__set_text_contents_method_in_args__newContents_arg =
{
  -1,  /* ref count */
  (gchar *) "newContents",
  (gchar *) "s",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_editable_text_interface__set_text_contents_method_in_args[] =
{
  &atspi_editable_text_interface__set_text_contents_method_in_args__newContents_arg,
  NULL,
};

const GDBusArgInfo atspi_editable_text_interface__set_text_contents_method_out_args__unnamed_arg1_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg1",
  (gchar *) "b",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_editable_text_interface__set_text_contents_method_out_args[] =
{
  &atspi_editable_text_interface__set_text_contents_method_out_args__unnamed_arg1_arg,
  NULL,
};

const GDBusMethodInfo atspi_editable_text_interface__set_text_contents_method =
{
  -1,  /* ref count */
  (gchar *) "SetTextContents",
  (GDBusArgInfo **) atspi_editable_text_interface__set_text_contents_method_in_args,
  (GDBusArgInfo **) atspi_editable_text_interface__set_text_contents_method_out_args,
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_editable_text_interface__insert_text_method_in_args__position_arg =
{
  -1,  /* ref count */
  (gchar *) "position",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_editable_text_interface__insert_text_method_in_args__text_arg =
{
  -1,  /* ref count */
  (gchar *) "text",
  (gchar *) "s",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_editable_text_interface__insert_text_method_in_args__length_arg =
{
  -1,  /* ref count */
  (gchar *) "length",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_editable_text_interface__insert_text_method_in_args[] =
{
  &atspi_editable_text_interface__insert_text_method_in_args__position_arg,
  &atspi_editable_text_interface__insert_text_method_in_args__text_arg,
  &atspi_editable_text_interface__insert_text_method_in_args__length_arg,
  NULL,
};

const GDBusArgInfo atspi_editable_text_interface__insert_text_method_out_args__unnamed_arg3_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg3",
  (gchar *) "b",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_editable_text_interface__insert_text_method_out_args[] =
{
  &atspi_editable_text_interface__insert_text_method_out_args__unnamed_arg3_arg,
  NULL,
};

const GDBusMethodInfo atspi_editable_text_interface__insert_text_method =
{
  -1,  /* ref count */
  (gchar *) "InsertText",
  (GDBusArgInfo **) atspi_editable_text_interface__insert_text_method_in_args,
  (GDBusArgInfo **) atspi_editable_text_interface__insert_text_method_out_args,
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_editable_text_interface__copy_text_method_in_args__startPos_arg =
{
  -1,  /* ref count */
  (gchar *) "startPos",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_editable_text_interface__copy_text_method_in_args__endPos_arg =
{
  -1,  /* ref count */
  (gchar *) "endPos",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_editable_text_interface__copy_text_method_in_args[] =
{
  &atspi_editable_text_interface__copy_text_method_in_args__startPos_arg,
  &atspi_editable_text_interface__copy_text_method_in_args__endPos_arg,
  NULL,
};

const GDBusMethodInfo atspi_editable_text_interface__copy_text_method =
{
  -1,  /* ref count */
  (gchar *) "CopyText",
  (GDBusArgInfo **) atspi_editable_text_interface__copy_text_method_in_args,
  NULL,  /* no out args */
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_editable_text_interface__cut_text_method_in_args__startPos_arg =
{
  -1,  /* ref count */
  (gchar *) "startPos",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_editable_text_interface__cut_text_method_in_args__endPos_arg =
{
  -1,  /* ref count */
  (gchar *) "endPos",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_editable_text_interface__cut_text_method_in_args[] =
{
  &atspi_editable_text_interface__cut_text_method_in_args__startPos_arg,
  &atspi_editable_text_interface__cut_text_method_in_args__endPos_arg,
  NULL,
};

const GDBusArgInfo atspi_editable_text_interface__cut_text_method_out_args__unnamed_arg2_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg2",
  (gchar *) "b",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_editable_text_interface__cut_text_method_out_args[] =
{
  &atspi_editable_text_interface__cut_text_method_out_args__unnamed_arg2_arg,
  NULL,
};

const GDBusMethodInfo atspi_editable_text_interface__cut_text_method =
{
  -1,  /* ref count */
  (gchar *) "CutText",
  (GDBusArgInfo **) atspi_editable_text_interface__cut_text_method_in_args,
  (GDBusArgInfo **) atspi_editable_text_interface__cut_text_method_out_args,
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_editable_text_interface__delete_text_method_in_args__startPos_arg =
{
  -1,  /* ref count */
  (gchar *) "startPos",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_editable_text_interface__delete_text_method_in_args__endPos_arg =
{
  -1,  /* ref count */
  (gchar *) "endPos",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_editable_text_interface__delete_text_method_in_args[] =
{
  &atspi_editable_text_interface__delete_text_method_in_args__startPos_arg,
  &atspi_editable_text_interface__delete_text_method_in_args__endPos_arg,
  NULL,
};

const GDBusArgInfo atspi_editable_text_interface__delete_text_method_out_args__unnamed_arg2_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg2",
  (gchar *) "b",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_editable_text_interface__delete_text_method_out_args[] =
{
  &atspi_editable_text_interface__delete_text_method_out_args__unnamed_arg2_arg,
  NULL,
};

const GDBusMethodInfo atspi_editable_text_interface__delete_text_method =
{
  -1,  /* ref count */
  (gchar *) "DeleteText",
  (GDBusArgInfo **) atspi_editable_text_interface__delete_text_method_in_args,
  (GDBusArgInfo **) atspi_editable_text_interface__delete_text_method_out_args,
  NULL,  /* no annotations */
};

const GDBusArgInfo atspi_editable_text_interface__paste_text_method_in_args__position_arg =
{
  -1,  /* ref count */
  (gchar *) "position",
  (gchar *) "i",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_editable_text_interface__paste_text_method_in_args[] =
{
  &atspi_editable_text_interface__paste_text_method_in_args__position_arg,
  NULL,
};

const GDBusArgInfo atspi_editable_text_interface__paste_text_method_out_args__unnamed_arg1_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg1",
  (gchar *) "b",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_editable_text_interface__paste_text_method_out_args[] =
{
  &atspi_editable_text_interface__paste_text_method_out_args__unnamed_arg1_arg,
  NULL,
};

const GDBusMethodInfo atspi_editable_text_interface__paste_text_method =
{
  -1,  /* ref count */
  (gchar *) "PasteText",
  (GDBusArgInfo **) atspi_editable_text_interface__paste_text_method_in_args,
  (GDBusArgInfo **) atspi_editable_text_interface__paste_text_method_out_args,
  NULL,  /* no annotations */
};

const GDBusMethodInfo * const atspi_editable_text_interface_methods[] =
{
  &atspi_editable_text_interface__set_text_contents_method,
  &atspi_editable_text_interface__insert_text_method,
  &atspi_editable_text_interface__copy_text_method,
  &atspi_editable_text_interface__cut_text_method,
  &atspi_editable_text_interface__delete_text_method,
  &atspi_editable_text_interface__paste_text_method,
  NULL,
};

const GDBusInterfaceInfo atspi_editable_text_interface =
{
  -1,  /* ref count */
  (gchar *) "org.a11y.atspi.EditableText",
  (GDBusMethodInfo **) atspi_editable_text_interface_methods,
  NULL,  /* no signals */
NULL,  /* no properties */
  NULL,  /* no annotations */
};

