/*
 * This file is generated by gdbus-codegen, do not modify it.
 *
 * The license of this code is the same as for the D-Bus interface description
 * it was derived from. Note that it links to GLib, so must comply with the
 * LGPL linking clauses.
 */

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include "atspi-deviceeventlistener.h"

#include <string.h>

/* ------------------------------------------------------------------------ */
/* Definitions for org.a11y.atspi.DeviceEventListener */

const GDBusArgInfo atspi_device_event_listener_interface__notify_event_method_in_args__event_arg =
{
  -1,  /* ref count */
  (gchar *) "event",
  (gchar *) "(uiuuisb)",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_device_event_listener_interface__notify_event_method_in_args[] =
{
  &atspi_device_event_listener_interface__notify_event_method_in_args__event_arg,
  NULL,
};

const GDBusArgInfo atspi_device_event_listener_interface__notify_event_method_out_args__unnamed_arg1_arg =
{
  -1,  /* ref count */
  (gchar *) "unnamed_arg1",
  (gchar *) "b",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const atspi_device_event_listener_interface__notify_event_method_out_args[] =
{
  &atspi_device_event_listener_interface__notify_event_method_out_args__unnamed_arg1_arg,
  NULL,
};

const GDBusAnnotationInfo atspi_device_event_listener_interface__notify_event_method_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation =
{
  -1,  /* ref count */
  (gchar *) "org.qtproject.QtDBus.QtTypeName.In0",
  (gchar *) "QSpiDeviceEvent",
  NULL,  /* no annotations */
};

const GDBusAnnotationInfo * const atspi_device_event_listener_interface__notify_event_method_annotations[] =
{
  &atspi_device_event_listener_interface__notify_event_method_annotations__org_qtproject__qt_dbus__qt_type_name__in0_annotation,
  NULL,
};

const GDBusMethodInfo atspi_device_event_listener_interface__notify_event_method =
{
  -1,  /* ref count */
  (gchar *) "NotifyEvent",
  (GDBusArgInfo **) atspi_device_event_listener_interface__notify_event_method_in_args,
  (GDBusArgInfo **) atspi_device_event_listener_interface__notify_event_method_out_args,
  (GDBusAnnotationInfo **) atspi_device_event_listener_interface__notify_event_method_annotations,
};

const GDBusMethodInfo * const atspi_device_event_listener_interface_methods[] =
{
  &atspi_device_event_listener_interface__notify_event_method,
  NULL,
};

const GDBusInterfaceInfo atspi_device_event_listener_interface =
{
  -1,  /* ref count */
  (gchar *) "org.a11y.atspi.DeviceEventListener",
  (GDBusMethodInfo **) atspi_device_event_listener_interface_methods,
  NULL,  /* no signals */
NULL,  /* no properties */
  NULL,  /* no annotations */
};

