/*
 * This file is generated by gdbus-codegen, do not modify it.
 *
 * The license of this code is the same as for the D-Bus interface description
 * it was derived from. Note that it links to GLib, so must comply with the
 * LGPL linking clauses.
 */

#ifndef __XDP_DBUS_H__
#define __XDP_DBUS_H__

#include <gio/gio.h>

G_BEGIN_DECLS


/* ------------------------------------------------------------------------ */
/* Declarations for org.freedesktop.portal.OpenURI */

#define GTK_XDP_TYPE_OPEN_URI (gtk_xdp_open_uri_get_type ())
#define GTK_XDP_OPEN_URI(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), GTK_XDP_TYPE_OPEN_URI, GtkXdpOpenURI))
#define GTK_XDP_IS_OPEN_URI(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), GTK_XDP_TYPE_OPEN_URI))
#define GTK_XDP_OPEN_URI_GET_IFACE(o) (G_TYPE_INSTANCE_GET_INTERFACE ((o), GTK_XDP_TYPE_OPEN_URI, GtkXdpOpenURIIface))

struct _GtkXdpOpenURI;
typedef struct _GtkXdpOpenURI GtkXdpOpenURI;
typedef struct _GtkXdpOpenURIIface GtkXdpOpenURIIface;

struct _GtkXdpOpenURIIface
{
  GTypeInterface parent_iface;


  gboolean (*handle_open_directory) (
    GtkXdpOpenURI *object,
    GDBusMethodInvocation *invocation,
    GUnixFDList *fd_list,
    const gchar *arg_parent_window,
    GVariant *arg_fd,
    GVariant *arg_options);

  gboolean (*handle_open_file) (
    GtkXdpOpenURI *object,
    GDBusMethodInvocation *invocation,
    GUnixFDList *fd_list,
    const gchar *arg_parent_window,
    GVariant *arg_fd,
    GVariant *arg_options);

  gboolean (*handle_open_uri) (
    GtkXdpOpenURI *object,
    GDBusMethodInvocation *invocation,
    const gchar *arg_parent_window,
    const gchar *arg_uri,
    GVariant *arg_options);

  guint  (*get_version) (GtkXdpOpenURI *object);

};

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (GtkXdpOpenURI, g_object_unref)
#endif

GType gtk_xdp_open_uri_get_type (void) G_GNUC_CONST;

GDBusInterfaceInfo *gtk_xdp_open_uri_interface_info (void);
guint gtk_xdp_open_uri_override_properties (GObjectClass *klass, guint property_id_begin);


/* D-Bus method call completion functions: */
void gtk_xdp_open_uri_complete_open_uri (
    GtkXdpOpenURI *object,
    GDBusMethodInvocation *invocation,
    const gchar *handle);

void gtk_xdp_open_uri_complete_open_file (
    GtkXdpOpenURI *object,
    GDBusMethodInvocation *invocation,
    GUnixFDList *fd_list,
    const gchar *handle);

void gtk_xdp_open_uri_complete_open_directory (
    GtkXdpOpenURI *object,
    GDBusMethodInvocation *invocation,
    GUnixFDList *fd_list,
    const gchar *handle);



/* D-Bus method calls: */
void gtk_xdp_open_uri_call_open_uri (
    GtkXdpOpenURI *proxy,
    const gchar *arg_parent_window,
    const gchar *arg_uri,
    GVariant *arg_options,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean gtk_xdp_open_uri_call_open_uri_finish (
    GtkXdpOpenURI *proxy,
    gchar **out_handle,
    GAsyncResult *res,
    GError **error);

gboolean gtk_xdp_open_uri_call_open_uri_sync (
    GtkXdpOpenURI *proxy,
    const gchar *arg_parent_window,
    const gchar *arg_uri,
    GVariant *arg_options,
    gchar **out_handle,
    GCancellable *cancellable,
    GError **error);

void gtk_xdp_open_uri_call_open_file (
    GtkXdpOpenURI *proxy,
    const gchar *arg_parent_window,
    GVariant *arg_fd,
    GVariant *arg_options,
    GUnixFDList *fd_list,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean gtk_xdp_open_uri_call_open_file_finish (
    GtkXdpOpenURI *proxy,
    gchar **out_handle,
    GUnixFDList **out_fd_list,
    GAsyncResult *res,
    GError **error);

gboolean gtk_xdp_open_uri_call_open_file_sync (
    GtkXdpOpenURI *proxy,
    const gchar *arg_parent_window,
    GVariant *arg_fd,
    GVariant *arg_options,
    GUnixFDList  *fd_list,
    gchar **out_handle,
    GUnixFDList **out_fd_list,
    GCancellable *cancellable,
    GError **error);

void gtk_xdp_open_uri_call_open_directory (
    GtkXdpOpenURI *proxy,
    const gchar *arg_parent_window,
    GVariant *arg_fd,
    GVariant *arg_options,
    GUnixFDList *fd_list,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean gtk_xdp_open_uri_call_open_directory_finish (
    GtkXdpOpenURI *proxy,
    gchar **out_handle,
    GUnixFDList **out_fd_list,
    GAsyncResult *res,
    GError **error);

gboolean gtk_xdp_open_uri_call_open_directory_sync (
    GtkXdpOpenURI *proxy,
    const gchar *arg_parent_window,
    GVariant *arg_fd,
    GVariant *arg_options,
    GUnixFDList  *fd_list,
    gchar **out_handle,
    GUnixFDList **out_fd_list,
    GCancellable *cancellable,
    GError **error);



/* D-Bus property accessors: */
guint gtk_xdp_open_uri_get_version (GtkXdpOpenURI *object);
void gtk_xdp_open_uri_set_version (GtkXdpOpenURI *object, guint value);


/* ---- */

#define GTK_XDP_TYPE_OPEN_URI_PROXY (gtk_xdp_open_uri_proxy_get_type ())
#define GTK_XDP_OPEN_URI_PROXY(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), GTK_XDP_TYPE_OPEN_URI_PROXY, GtkXdpOpenURIProxy))
#define GTK_XDP_OPEN_URI_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), GTK_XDP_TYPE_OPEN_URI_PROXY, GtkXdpOpenURIProxyClass))
#define GTK_XDP_OPEN_URI_PROXY_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), GTK_XDP_TYPE_OPEN_URI_PROXY, GtkXdpOpenURIProxyClass))
#define GTK_XDP_IS_OPEN_URI_PROXY(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), GTK_XDP_TYPE_OPEN_URI_PROXY))
#define GTK_XDP_IS_OPEN_URI_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), GTK_XDP_TYPE_OPEN_URI_PROXY))

typedef struct _GtkXdpOpenURIProxy GtkXdpOpenURIProxy;
typedef struct _GtkXdpOpenURIProxyClass GtkXdpOpenURIProxyClass;
typedef struct _GtkXdpOpenURIProxyPrivate GtkXdpOpenURIProxyPrivate;

struct _GtkXdpOpenURIProxy
{
  /*< private >*/
  GDBusProxy parent_instance;
  GtkXdpOpenURIProxyPrivate *priv;
};

struct _GtkXdpOpenURIProxyClass
{
  GDBusProxyClass parent_class;
};

GType gtk_xdp_open_uri_proxy_get_type (void) G_GNUC_CONST;

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (GtkXdpOpenURIProxy, g_object_unref)
#endif

void gtk_xdp_open_uri_proxy_new (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
GtkXdpOpenURI *gtk_xdp_open_uri_proxy_new_finish (
    GAsyncResult        *res,
    GError             **error);
GtkXdpOpenURI *gtk_xdp_open_uri_proxy_new_sync (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);

void gtk_xdp_open_uri_proxy_new_for_bus (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
GtkXdpOpenURI *gtk_xdp_open_uri_proxy_new_for_bus_finish (
    GAsyncResult        *res,
    GError             **error);
GtkXdpOpenURI *gtk_xdp_open_uri_proxy_new_for_bus_sync (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);


/* ---- */

#define GTK_XDP_TYPE_OPEN_URI_SKELETON (gtk_xdp_open_uri_skeleton_get_type ())
#define GTK_XDP_OPEN_URI_SKELETON(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), GTK_XDP_TYPE_OPEN_URI_SKELETON, GtkXdpOpenURISkeleton))
#define GTK_XDP_OPEN_URI_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), GTK_XDP_TYPE_OPEN_URI_SKELETON, GtkXdpOpenURISkeletonClass))
#define GTK_XDP_OPEN_URI_SKELETON_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), GTK_XDP_TYPE_OPEN_URI_SKELETON, GtkXdpOpenURISkeletonClass))
#define GTK_XDP_IS_OPEN_URI_SKELETON(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), GTK_XDP_TYPE_OPEN_URI_SKELETON))
#define GTK_XDP_IS_OPEN_URI_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), GTK_XDP_TYPE_OPEN_URI_SKELETON))

typedef struct _GtkXdpOpenURISkeleton GtkXdpOpenURISkeleton;
typedef struct _GtkXdpOpenURISkeletonClass GtkXdpOpenURISkeletonClass;
typedef struct _GtkXdpOpenURISkeletonPrivate GtkXdpOpenURISkeletonPrivate;

struct _GtkXdpOpenURISkeleton
{
  /*< private >*/
  GDBusInterfaceSkeleton parent_instance;
  GtkXdpOpenURISkeletonPrivate *priv;
};

struct _GtkXdpOpenURISkeletonClass
{
  GDBusInterfaceSkeletonClass parent_class;
};

GType gtk_xdp_open_uri_skeleton_get_type (void) G_GNUC_CONST;

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (GtkXdpOpenURISkeleton, g_object_unref)
#endif

GtkXdpOpenURI *gtk_xdp_open_uri_skeleton_new (void);


G_END_DECLS

#endif /* __XDP_DBUS_H__ */
