/* This file is generated by glib-genmarshal, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */
#pragma once

#include <glib-object.h>

G_BEGIN_DECLS

/* BOOLEAN:BOXED (../gtk-4.14.1/gdk/gdkmarshalers.list:1) */
extern
void _gdk_marshal_BOOLEAN__BOXED (GClosure     *closure,
                                  GValue       *return_value,
                                  guint         n_param_values,
                                  const GValue *param_values,
                                  gpointer      invocation_hint,
                                  gpointer      marshal_data);
extern
void _gdk_marshal_BOOLEAN__BOXEDv (GClosure *closure,
                                   GValue   *return_value,
                                   gpointer  instance,
                                   va_list   args,
                                   gpointer  marshal_data,
                                   int       n_params,
                                   GType    *param_types);

/* BOOLEAN:OBJECT (../gtk-4.14.1/gdk/gdkmarshalers.list:2) */
extern
void _gdk_marshal_BOOLEAN__OBJECT (GClosure     *closure,
                                   GValue       *return_value,
                                   guint         n_param_values,
                                   const GValue *param_values,
                                   gpointer      invocation_hint,
                                   gpointer      marshal_data);
extern
void _gdk_marshal_BOOLEAN__OBJECTv (GClosure *closure,
                                    GValue   *return_value,
                                    gpointer  instance,
                                    va_list   args,
                                    gpointer  marshal_data,
                                    int       n_params,
                                    GType    *param_types);

/* BOOLEAN:POINTER (../gtk-4.14.1/gdk/gdkmarshalers.list:3) */
extern
void _gdk_marshal_BOOLEAN__POINTER (GClosure     *closure,
                                    GValue       *return_value,
                                    guint         n_param_values,
                                    const GValue *param_values,
                                    gpointer      invocation_hint,
                                    gpointer      marshal_data);
extern
void _gdk_marshal_BOOLEAN__POINTERv (GClosure *closure,
                                     GValue   *return_value,
                                     gpointer  instance,
                                     va_list   args,
                                     gpointer  marshal_data,
                                     int       n_params,
                                     GType    *param_types);

/* VOID:POINTER,POINTER,BOOLEAN,BOOLEAN (../gtk-4.14.1/gdk/gdkmarshalers.list:4) */
extern
void _gdk_marshal_VOID__POINTER_POINTER_BOOLEAN_BOOLEAN (GClosure     *closure,
                                                         GValue       *return_value,
                                                         guint         n_param_values,
                                                         const GValue *param_values,
                                                         gpointer      invocation_hint,
                                                         gpointer      marshal_data);
extern
void _gdk_marshal_VOID__POINTER_POINTER_BOOLEAN_BOOLEANv (GClosure *closure,
                                                          GValue   *return_value,
                                                          gpointer  instance,
                                                          va_list   args,
                                                          gpointer  marshal_data,
                                                          int       n_params,
                                                          GType    *param_types);

/* VOID:INT,INT (../gtk-4.14.1/gdk/gdkmarshalers.list:5) */
extern
void _gdk_marshal_VOID__INT_INT (GClosure     *closure,
                                 GValue       *return_value,
                                 guint         n_param_values,
                                 const GValue *param_values,
                                 gpointer      invocation_hint,
                                 gpointer      marshal_data);
extern
void _gdk_marshal_VOID__INT_INTv (GClosure *closure,
                                  GValue   *return_value,
                                  gpointer  instance,
                                  va_list   args,
                                  gpointer  marshal_data,
                                  int       n_params,
                                  GType    *param_types);


G_END_DECLS
