#pragma once

#include "gskgpuopprivate.h"

G_BEGIN_DECLS

void                    gsk_gpu_mipmap_op                               (GskGpuFrame                    *frame,
                                                                         GskGpuImage                    *image);

G_END_DECLS

