#pragma once

#include "gskgputypesprivate.h"

G_BEGIN_DECLS

void                    gsk_gpu_blend_op                                (GskGpuFrame                    *frame,
                                                                         GskGpuBlend                     blend);


G_END_DECLS

